# Memri iOS App Architecture

The Memri application is designed with an ambitious goal in mind: to provide a way to show **any** of your data the way you want it. This is achieved through two key components:

- The memri pod maintains a **dynamic schema** which enables data types to be defined
- The memri **CVU language** describes ways to display and interact with these data types within the app



# App Controller hierarchy

1) App Controller

- This class is responsible for holding different controllers that are shared throughout the app - eg. database controller, cvu controller

2) Scene Controller

- This class holds things specific to a particular `scene` which on MacOS or iPadOS is a window (currently on iOS there is only a single scene per app)
- The sceneController maintains a hierarchy of ViewContext structs, this allows for persistence of the hierarchy to the database, and for backward navigation

3) ViewContext Controller

- This controller is created for every data screen shown within the app. It maintains a `viewContext` describing the following
  - The database query used to fetch the data
  - The CVU renderer used to display the data
  - Any `viewArguments` passed down from a parent view



# UI Hierarchy

## View hierarchy

1) MainView - determines whether to display Setup, Auth, or Browser

2) SceneView - displays the 'chrome' (eg. navigation bar, navigation panel), and contains a 'UINavigationController' that displays the active view context

3) SceneContentView - selects the correct renderer view and provides the BottomBar/SearchView

4) RendererViews (multiple) - the renderer is responsible for displaying the ViewContext. Each renderer can query CVU for renderer options, how to display an item



## Other views

- **Navigation panel**
  - displays a list of `NavigationItem` items (from the database)
  - If one is pressed, it tells the SceneController to navigate to this item
- **Filter panel** - provides three tabs
  - 1) Renderer selection - if multiple renderers are defined in the CVU it allows the user to select which one to use
  - 2) Renderer options - if the renderer provides a 'RendererSettingsView' this is displayed here
  - 3) Filter options - currently allows filtering by dateModified/dateCreated

# CVU

## CVU Expressions

Within CVU you can provide an expression for a property. This is dynamically resolved when the UI requires that property.

Expressions have two forms in CVU (note the different start/finish brackets):

- `{{.height * 100}}` 
  - A standard expression, the UI could request to resolve this to a number, or a string depending on its purpose
- `"{.firstName " " .lastName}" `
  - An expression that is intended to output a string only. It will merge the results of each subexpression to form a string.

The CVU language doesn't have fixed types for properties. It is up to the renderer or UI element to request the type of result it requires (eg. string, number, bool, etc) and the expression resolver will take care of the rest.



## CVU Components

When the app needs to display a data item, it asks the CVUController for the appropriate `CVUUINode` . This represents the CVU definition of how to display the item. 

The `CVUElementView` type translate this node into SwiftUI views.



**Adding support for new elements to CVU**

1) Add the new element type to `CVUUIElementFamily`

2) Create a SwiftUI view that displays the node using a `nodeResolver` to fetch any CVU properties in the desired type. See `CVU_Text` or `CVU_Toggle` for different examples of this

3) Add the element to the CVUElementView switch statement (which resolves to the correct SwiftUI view for each type of element)



# Database

The Memri iOS app uses a SQLite database to store data. In the future this will be encrypted with SQLCipher.

### Schema

The dynamic schema of the pod is not to be confused with the database schema (which provides a way to store the data that is resilient to changes in the dynamic schema)



# Networking

The Memri App works alongside the pod. The data is synced between the app and the pod when changes are made.

- **SyncController** 
  - This controller is responsible for scheduling and executing sync events with the pod
  - (To be implemented: ) The sync controller will sync on these occasions:
    - First connection to pod (sync ALL items)
    - **Opening a new screen** (check if any changes relevant to the current database query)
    - **Pull to refresh** on a screen (check if any changes relevant to the current database query)
    - **Automatic schedule** - Check every <u>x</u> minutes for ANY changes from the pod
- **PodAPIRequests**
  - **PodAPIStandardRequest**
    - used to construct and execute a standard API request.
    - There are convenience functions for known API endpoints (eg. `PodAPIStandardRequest.getItemWithEdges(...)`)
  - **PodAPIUploadRequest**
    - used to upload a file from the pod
  - **PodAPIDownloadRequest**
    - used to download a file from the pod

