//
//  CVUController.swift
//  MemriDatabase
//
//  Created by T Brennan on 7/12/20.
//

import Foundation

class CVUController {
    var definitions: [CVUParsedDefinition]
    
    init() {
        do {
            definitions = Constants.isRunningTests ? [] : try CVUController.parseCVU()
        } catch {
            print(error)
            definitions = []
        }
    }
    
    static func parseCVU() throws -> [CVUParsedDefinition] {
        let string = CVUController.readCVUString()
        return try parseCVU(string: string)
    }
    
    static func parseCVU(string: String) throws -> [CVUParsedDefinition] {
        let lexer = CVULexer(input: string)
        let tokens = try lexer.tokenize()
        let parser = CVUParser(tokens)
        return try parser.parse()
    }
    
    static private func readCVUString() -> String {
        let resourceKeys: [URLResourceKey] = [.isDirectoryKey]
        guard
            let directory = Bundle.main.resourceURL?.appendingPathComponent(
                "defaultCVU",
                isDirectory: true
            ),
            let enumerator = FileManager.default.enumerator(at: directory,
                                                            includingPropertiesForKeys: resourceKeys,
                                                            options: [.skipsHiddenFiles],
                                                            errorHandler: { (_, _) -> Bool in
                                                                true
                                                            })
        else { return "" }
        return enumerator.compactMap { $0 as? URL }.compactMap {
            guard $0.pathExtension == "cvu" else { return nil }
            return try? String(contentsOf: $0)
        }.joined(separator: "\n")
    }
    
    private func definitionFor(type: CVUDefinitionType, selector: String? = nil, exactSelector: Bool = false, viewName: String? = nil, rendererName: String? = nil) -> CVUParsedDefinition? {
        CVUController.definitionFrom(definitions: definitions, type: type, selector: selector, exactSelector: exactSelector, viewName: viewName, rendererName: rendererName)
    }
    
    func nodeDefinitionFor(item: ItemRecord, selector: String? = nil, renderer: String?) -> CVUDefinitionContent? {
        return definitionFor(type: .uiNode, selector: item.type, rendererName: renderer)?.parsed
    }
    
    func viewDefinitionFor(viewName: String, customDefinition: CVUDefinitionContent? = nil) -> CVUDefinitionContent? {
        let definition = definitionFor(type: .view, viewName: viewName)?.parsed
        return definition?.merge(with: customDefinition) ?? customDefinition
    }
    
    func viewDefinitionFor(itemRecord: ItemRecord?) -> CVUDefinitionContent? {
        let definition = definitionFor(type: .view, selector:itemRecord?.type)?.parsed
        return definition
    }
    
    func edgeDefinitionFor(itemRecord: ItemRecord) -> CVUDefinitionContent? {
        let definition = definitionFor(type: .view, selector:"\(itemRecord.type)[]")?.parsed
        return definition
    }
    
    func rendererDefinitionFor(_ context: CVUContext) -> CVUParsedDefinition? {
        guard let specificDefinition = CVUController.definitionFrom(definitions: context.viewDefinition.definitions, type: .renderer, viewName: context.rendererName)
        else { return nil }
        if let globalDefinition = definitionFor(type: .renderer, viewName: context.rendererName) {
            return globalDefinition.merge(with: specificDefinition)
        } else {
            return specificDefinition
        }
    }
    
    func rendererDefinitionFor(selector: String? = nil, viewName: String? = nil) -> CVUDefinitionContent? {
        let definition = definitionFor(type: .renderer, selector:selector, viewName: viewName)?.parsed
        return definition
    }
    
    func nodeDefinitionFor(_ context: CVUContext) -> CVUDefinitionContent? {
        guard let currentItem = context.currentItem
        else { return nil }
        
        let globalDefinition = definitionFor(type: .uiNode, selector: currentItem.type, rendererName: context.rendererName)?.parsed
        let specificDefinition = CVUController.definitionFrom(definitions: context.viewDefinition.definitions, type: .uiNode, selector: currentItem.type, rendererName: context.rendererName)?.parsed
        
        return globalDefinition?.merge(with: specificDefinition) ?? specificDefinition
    }
    
    func defaultViewDefinitionFor(_ context: CVUContext) -> CVUDefinitionContent? {
        guard let currentItem = context.currentItem
        else { return nil }
        
        for selector in ["\(currentItem.type)[]", "*[]"] {
            if let globalDefinition = definitionFor(type: .view, selector: selector)?.parsed {
                if globalDefinition.children.count > 0 {
                    return globalDefinition
                }
                
                if let rendererDefinition = globalDefinition.definitions.filter({$0.name == context.rendererName && $0.parsed.children.count > 0}).first {
                    return rendererDefinition.parsed
                }
            }
        }
        return nil
    }
    
    private static func definitionFrom(definitions: [CVUParsedDefinition], type: CVUDefinitionType, selector: String? = nil, exactSelector: Bool = false, viewName: String? = nil, rendererName: String? = nil) -> CVUParsedDefinition? {
        let relevantDefinitions = definitions.filter { (def) -> Bool in
            // Find a definition of the requested type
            guard def.type == type else { return false }
            
            // with matching name
            if let viewName = viewName {
                guard def.name?.lowercased() == viewName.lowercased() else {
                    return false
                }
            }
            
            // with matching rendererName
            if let rendererName = rendererName {
                guard def.renderer?.lowercased() == rendererName.lowercased() else {
                    return false
                }
            }
            
            // with matching selector or wildcard selector
            if let selector = selector {
                guard (def.selector?.lowercased() == selector.lowercased() || (!exactSelector && (def.selector == "*" || def.selector == nil))) else {
                    return false
                }
            }
            return true
        }.sorted { (lhs, rhs) -> Bool in
            if lhs.selector == "*" || lhs.selector == "*[]" || lhs.selector == nil {
                return true
            } else if rhs.selector == "*" || rhs.selector == "*[]" || rhs.selector == nil {
                return false
            } else {
                // TODO: Improve this VERY crude way of determining selector specificity
                return (lhs.selector?.count ?? 0) <= (rhs.selector?.count ?? 0)
            }
        }
        guard let firstDefinition = relevantDefinitions.first else { return nil }
        
        let mergedDefinition = relevantDefinitions.dropFirst().reduce(firstDefinition) { (old, latest) in
            old.merge(with: latest)
        }
        return mergedDefinition
    }
    
}

import SwiftUI
extension CVUController {
    
    @ViewBuilder
    func render(cvuContext: CVUContext, nodeDefinition: CVUDefinitionContent? = nil, lookup: CVULookupController, db: DatabaseController, blankIfNoDefinition: Bool) -> some View {
        if let node = nodeDefinition?.children.first {
            CVUElementView(nodeResolver: .init(context: cvuContext, lookup: lookup, node: node, db: db))
        } else if let node = nodeDefinitionFor(cvuContext)?.children.first{
            CVUElementView(nodeResolver: .init(context: cvuContext, lookup: lookup, node: node, db: db))
        } else if let node = defaultViewDefinitionFor(cvuContext)?.children.first{
            CVUElementView(nodeResolver: .init(context: cvuContext, lookup: lookup, node: node, db: db))
        } else if !blankIfNoDefinition, let type = cvuContext.currentItem?.type {
            Text("No definition for displaying a `\(type)` in this context").font(.caption)
        }
    }
}
