//
//  CVUAction.swift
//  MemriDatabase
//
//  Created by T Brennan on 8/1/21.
//

import Foundation
import UIKit
import SwiftUI

protocol CVUAction {
    init(vars: [String: CVUValue])
    var id: String { get }
    var defaultVars: [String: CVUValue] { get }
    var vars: [String: CVUValue] { get }
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?)
    func getString(_ key: String, _ context: CVUContext, _ viewContext: ViewContextController?) -> String?
}

extension CVUAction {
    var id: String {
        return UUID().uuidString
    }

    var defaultVars: [String: CVUValue] {
        return [:]
    }
    
    func getString(_ key: String, _ context: CVUContext, _ viewContext: ViewContextController?) -> String? {
        guard case let .constant(.string(value)) = (vars[key] ?? defaultVars[key]) else {
            return nil
        }
        
        return value
    }
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        // Default implementation does nothing
        completion?(nil)
    }
}

/// Used to look up the concrete type matching a CVU action name
func cvuAction(named: String) -> CVUAction.Type? {
    switch named.lowercased() {
    case "back":
        return CVUAction_NavigateBack.self
    case "additem":
        return CVUAction_AddItem.self
    case "openview":
        return CVUAction_OpenView.self
    case "opengroup":
        return CVUAction_OpenGroup.self
    case "openviewbyname":
        return CVUAction_OpenViewByName.self
    case "toggleeditmode":
        return CVUAction_ToggleEditMode.self
    case "togglefilterpanel":
        return CVUAction_ToggleFilterPanel.self
    case "star":
        return CVUAction_Star.self
    case "showstarred":
        return CVUAction_ShowStarred.self
    case "showcontextpane":
        return CVUAction_ShowContextPane.self
    case "shownavigation":
        return CVUAction_ShowNavigation.self
    case "duplicate":
        return CVUAction_Duplicate.self
    case "schedule":
        return CVUAction_Schedule.self
    case "delete":
        return CVUAction_Delete.self
    case "showsessionswitcher":
        return CVUAction_showSessionSwitcher.self
    case "forward":
        return CVUAction_NavigateBack.self
    case "forwardtofront":
        return CVUAction_ForwardToFront.self
    case "backassession":
        return CVUAction_BackAsSession.self
    case "opensession":
        return CVUAction_OpenSession.self
    case "opensessionbyname":
        return CVUAction_OpenSessionByName.self
    case "closepopup":
        return CVUAction_ClosePopup.self
    case "link":
        return CVUAction_Link.self
    case "unlink":
        return CVUAction_Unlink.self
    case "multiaction":
        return CVUAction_MultiAction.self
    case "requestcontacts":
        return CVUAction_RequestContactsPermission.self
    case "requestlocation":
        return CVUAction_RequestLocationPermission.self
    case "requestcamera":
        return CVUAction_RequestCameraPermission.self
    case "requestphotolibrary":
        return CVUAction_RequestPhotoLibraryPermission.self
    case "requeststorage":
        return CVUAction_RequestStoragePermission.self
    case "openpopup":
        return CVUAction_OpenPopup.self
    case "runindexer":
        return CVUAction_RunIndexer.self
    case "startplugin":
        return CVUAction_StartPlugin.self
    case "setproperty":
        return CVUAction_SetProperty.self
    case "setsetting":
        return CVUAction_SetSetting.self
    case "copytoclipboard":
        return CVUAction_CopyToClipboard.self
    case "noop":
        return CVUAction_Noop.self
    case "togglenavigation":
        return CVUAction_ToggleNavigation.self
    case "togglefullscreen":
        return CVUAction_ToggleFullScreen.self
    case "tonextitem":
        return CVUAction_ToNextItem.self
    case "topreviousitem":
        return CVUAction_ToPreviousItem.self
    case "selectall":
        return CVUAction_SelectAll.self
    case "sync":
        return CVUAction_Sync.self
    case "deselectall":
        return CVUAction_DeselectAll.self
    default:
        return CVUAction_Noop.self
    }
}

struct CVUAction_ShowStarred: CVUAction {
    var vars: [String: CVUValue]
    var defaultVars: [String: CVUValue] { [
        "icon": .constant(.string("star")),
    ]}
    
    func getString(_ key: String, _ context: CVUContext, _ viewContext: ViewContextController?) -> String? {
        guard key == "icon",
              let validViewContext = viewContext else {
            return nil
        }
        
        for condition in validViewContext.config.query.conditions {
            if case let .propertyEquals(property) = condition,
               property.name == "starred" {
                return "star.fill"
            }
        }
        
        return "star"
    }
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        guard let viewContext = sceneController.topMostContext else {
            return
        }
        
        var active = false
        var filteredConditions = viewContext.config.query.conditions.filter{
            if case let .propertyEquals(property) = $0,
               property.name == "starred" {
                active = true
                return false
            }
            
            return true
        }
        
        if (!active) {
            filteredConditions.append(.propertyEquals(DatabaseQueryCondition.PropertyEquals(name: "starred", value: .bool(true))))
        }
        
        viewContext.config.query.conditions = filteredConditions
        completion?(nil)
    }
}

struct CVUAction_OpenView: CVUAction {
    var vars: [String: CVUValue] = [:]
    var viewName: String?
    var renderer: String?
    var uids: Set<Int64>?
    var dateRange: DateInterval?
    var viewDefinition: CVUDefinitionContent?
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        var customDefinition: CVUDefinitionContent? = viewDefinition
        if (customDefinition == nil) {
            if case let .subdefinition(viewDef) = vars["view"] {
                customDefinition = viewDef
            }
        }
        let viewArguments: CVUViewArguments
        if case let .subdefinition(viewArgs) = vars["viewArguments"] {
            viewArguments = CVUViewArguments(args: viewArgs.properties, argumentItem: context.currentItem, parentArguments: context.viewArguments)
        } else {
            viewArguments = CVUViewArguments()
        }
        let db = sceneController.appController.databaseController
        let resolver = CVUPropertyResolver(context: context, lookup: CVULookupController(), db: db, properties: vars)
        sceneController.contextPaneVisible = false
        sceneController.navigateToNewContext(viewName: viewName ?? resolver.string("viewName") ?? "customView",
                                             inheritDatasource: resolver.bool("inheritDatasource", defaultValue: true),
                                             overrideRenderer: renderer ?? resolver.string("renderer"),
                                             defaultRenderer: "singleItem",
                                             targetItem: context.currentItem,
                                             overrideUIDs: uids,
                                             dateRange: dateRange,
                                             customDefinition: customDefinition,
                                             viewArguments: viewArguments)
        completion?(nil)
    }
}


extension CVUAction_OpenView {
    init(vars: [String: CVUValue]) {
        self.vars = vars
    }
}


struct CVUAction_OpenViewByName: CVUAction {
    var vars: [String: CVUValue] = [:]

    var viewName: String?
    var renderer: String?
    var itemType: String?
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        var rendererName: String? = renderer
        if case let .constant(.string(cvuRenderer)) = vars["renderer"] {
            rendererName = cvuRenderer
        }
        
        CVUAction_OpenView(vars: vars, viewName: viewName, renderer: rendererName).execute(sceneController: sceneController, context: context, completion: completion)
    }
    
    func getViewContext(context: CVUContext) -> ViewContextController? {
        
        var customDefinition: CVUDefinitionContent?
        if case let .subdefinition(viewDef) = vars["view"] {
            customDefinition = viewDef
        }
        let viewArguments: CVUViewArguments
        if case let .subdefinition(viewArgs) = vars["viewArguments"] {
            viewArguments = CVUViewArguments(args: viewArgs.properties, argumentItem: context.currentItem, parentArguments: context.viewArguments)
        } else {
            viewArguments = CVUViewArguments()
        }
        
        let appController: AppController = .shared
        let db = appController.databaseController
        let resolver = CVUPropertyResolver(context: context, lookup: CVULookupController(), db: db, properties: vars)
        let viewName = self.viewName ?? resolver.string("viewName") ?? "customView"
        let viewDefinition = appController.cvuController.viewDefinitionFor(viewName: viewName, customDefinition: customDefinition) ?? CVUDefinitionContent()
        let newContext = CVUContext(currentItem: nil, selector: nil, viewName: viewName, viewDefinition: viewDefinition, viewArguments: viewArguments)
        
        let defaultRenderer = "list"
        let rendererName = { () -> String in
            return viewDefinition.propertyResolver(context: newContext, lookup: CVULookupController(), db: appController.databaseController).string("defaultRenderer") ?? defaultRenderer
        }()

        var queryConfig = DatabaseQueryConfig()
        if let itemType = self.itemType{
            queryConfig.itemTypes = [itemType]
        } else if case let .constant(.string(query)) = viewArguments.args["query"] {
            queryConfig.itemTypes = [query]
        }
        
        let config = ViewContext(viewName: viewName, rendererName: rendererName, viewDefinition: viewDefinition, query: queryConfig, viewArguments: viewArguments, focusedItem: context.currentItem)
        let holder = ViewContextHolder(config)
        let newViewContext = ViewContextController(config: holder, databaseController: appController.databaseController, cvuController: appController.cvuController)
        
        return newViewContext
    }
}


extension CVUAction_OpenViewByName {
    init(vars: [String: CVUValue]) {
        self.vars = vars
    }
}

struct CVUAction_NavigateBack: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        sceneController.navigateBack()
        completion?(nil)
    }
}

struct CVUAction_CopyToClipboard: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        let db = sceneController.appController.databaseController
        let resolver = CVUPropertyResolver(context: context, lookup: CVULookupController(), db: db, properties: vars)
        if let value = resolver.string("value") {
            UIPasteboard.general.string = value
        }
        completion?(nil)
    }
}

struct CVUAction_AddItem: CVUAction {
    var vars: [String: CVUValue]
    var defaultVars: [String: CVUValue] { ["icon":.constant(.string("plus"))] }

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        let db = sceneController.appController.databaseController
        let resolver = CVUPropertyResolver(context: context, lookup: CVULookupController(), db: db, properties: vars)
        if let template = resolver.subdefinition("template"), let type = template.string("_type") {
            let item = ItemRecord(type: type)
            do {
                try item.save()
            } catch let error {
                print("ERROR Adding item: ", error)
            }
          
            guard let itemRowId = item.rowId else {
                return
            }
            
            /// Take all the properties defined in the template definition (in CVU) and map them against the schema. Resolve the CVU based on the type expected by the schema.
            let propertiesAndEdges = template.properties.keys.compactMap { key -> BaseRecord? in
                guard let valueType = db.schema.expectedType(forItemType: type, propertyOrEdgeName: key) else { return nil }
                switch valueType {
                case .property(let propertyType):
                    switch propertyType {
                    case .string:
                        guard let stringValue = template.string(key) else {
                            return nil
                        }
                        return StringRecord(item: itemRowId, name: key, value: stringValue)
                    case .bool:
                        guard let boolValue = template.bool(key) else {
                            return nil
                        }
                        return IntegerRecord(item: itemRowId, name: key, value: boolValue ? 1 : 0)
                    case .int:
                        guard let intValue = template.int(key) else {
                            return nil
                        }
                        return IntegerRecord(item: itemRowId, name: key, value: Int64(intValue))
                    case .double:
                        guard let doubleValue = template.double(key) else {
                            return nil
                        }
                        return RealRecord(item: itemRowId, name: key, value: doubleValue)
                    case .datetime:
                        guard let datetimeValue = template.dateTime(key) else {
                            return nil
                        }
                        return RealRecord(item: itemRowId, name: key, value: Double(DatabaseHelperFunctions.encode(datetimeValue)))
                    default:
                        return nil
                    }
                case .edge:
                    // todomigrate: Make sure this passes
                    guard let target = template.item(key),
                          let targetRowId = target.rowId else { return nil }
                    return try? EdgeRecord.createFor(source: itemRowId, type: key, target: targetRowId)
                }
            }
            try? db.write { (db) in
                try propertiesAndEdges.forEach { try $0.save(db) }
            }
            
            var renderer = "generalEditor"
            if let viewDefinition = AppController.shared.cvuController.viewDefinitionFor(itemRecord: item),
                case let .constant(.argument(defaultRenderer)) = viewDefinition.properties["defaultRenderer"] {
                renderer = defaultRenderer
            }
            
            var newVars = vars
            newVars["viewArguments"] = .subdefinition(CVUDefinitionContent(properties: ["readOnly": .constant(.bool(false)), "isEditing": .constant(.bool(true))]))

            CVUAction_OpenView(vars: newVars, viewName: type, renderer: renderer).execute(sceneController: sceneController, context: context.replacingItem(item), completion: completion)
            
            sceneController.isInEditMode = true
            
            try? AppController.shared.syncController.sync()
        }
    }
}

struct CVUAction_StartPlugin: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        
        let lookup = CVULookupController()
        let db = sceneController.appController.databaseController

        guard let plugin = context.currentItem,
            let targetItemIdValue = vars["targetItemId"],
           let targetItemId: String = lookup.resolve(value: targetItemIdValue, context: context, db: db),
           let containerValue = vars["container"],
           let container: String = lookup.resolve(value: containerValue, context: context, db: db) else {
            return
        }

        do {
            let startPluginItem = ItemRecord(type: "StartPlugin")
            try startPluginItem.save()
            try startPluginItem.setPropertyValue(name: "targetItemId", value: .string(targetItemId))
            try startPluginItem.setPropertyValue(name: "container", value: .string(container))
            try startPluginItem.setPropertyValue(name: "state", value: .string("idle"))

            // This is test code to render cvu for a plugin locally without connecting to pod
            // Ideally plugins will contain more detailed container name
            // We can remove this once plugins are able to send CVUs in stored definition
            if (container == "cvu") {
                PluginHandler.presentCVUforPlugin(plugin: plugin, runner: startPluginItem, sceneController: sceneController, context: context)
            } else {
                PluginHandler.start(plugin: plugin, runner: startPluginItem, sceneController: sceneController, context: context)
            }
        } catch let error {
            print("Error starting plugin: ", error)
        }
    }
}

struct CVUAction_Sync: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        do {
            var pendingItems = [BaseRecord]()
            
            if let item = context.currentItem {
                if (item.syncState == .skip) {
                    item.syncState = .create
                    pendingItems.append(item)
                }
                
                let edges = item.edges() + item.reverseEdgeObjects()
                for edge in edges {
                    if (edge.syncState == .skip) {
                        edge.syncState = .create
                        pendingItems.append(edge)
                    }
                }
                
                let edgeItems = item.edgeItems() + item.reverseEdgeItems()
                for edgeItem in edgeItems {
                    if (edgeItem.syncState == .skip) {
                        edgeItem.syncState = .create
                        pendingItems.append(edgeItem)
                    }
                }
            }
            
            try AppController.shared.databaseController.writeSync { (db) in
                for item in pendingItems {
                    try item.save(db)
                }
            }
            
            try AppController.shared.syncController.sync()
            if (sceneController.isInEditMode) {
                sceneController.toggleEditMode()
            }
        } catch let error {
            print("Error starting sync: ", error)
        }
        
        completion?(nil)
    }
}

struct CVUAction_ToggleEditMode: CVUAction {
    var vars: [String: CVUValue]
    
    var defaultVars: [String: CVUValue] { [
        "icon": .constant(.string("pencil"))
    ]}
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        sceneController.toggleEditMode()
        completion?(nil)
    }
}


struct CVUAction_ToggleNavigation: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        withAnimation {
            sceneController.navigationIsVisible.toggle()
        }
        completion?(nil)
    }
}

struct CVUAction_ToggleFullScreen: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_ToNextItem: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {

        defer {
            sceneController.scheduleUIUpdate()
        }
        
        guard let viewContext = sceneController.topMostContext else {
            return
        }
        
        guard viewContext.focusedIndex < viewContext.items.count - 1 else {
            viewContext.focusedIndex = 0
            return
        }
        
        viewContext.focusedIndex = viewContext.focusedIndex + 1
        completion?(nil)
    }
}

struct CVUAction_ToPreviousItem: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {

        defer {
            sceneController.scheduleUIUpdate()
        }
        
        guard let viewContext = sceneController.topMostContext else {
            return
        }
        
        guard viewContext.focusedIndex > 0 else {
            viewContext.focusedIndex = viewContext.items.count - 1
            return
        }
        
        viewContext.focusedIndex = viewContext.focusedIndex - 1
        completion?(nil)
    }
}

struct CVUAction_Delete: CVUAction {
    var vars: [String: CVUValue] = [:]

    var defaultVars: [String: CVUValue] { [
        "icon": .constant(.string("trash")),
    ]}
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        if let item = context.currentItem {
            var isShowingItem = false
            if let context = sceneController.topMostContext,
               context.config.focusedItem == item {
                isShowingItem = true
            }
            
            let deleted = try? item.markDeleted()
                
            if isShowingItem == true, deleted == true {
                if (sceneController.isInEditMode) {
                    sceneController.toggleEditMode()
                }
                sceneController.navigateBack()
            }
        }
        
        if let context = sceneController.topMostContext {
            var selectedItems = context.selectedItems
            for itemId in context.selectedItems {
                if let item = ItemRecord.fetchWithID(itemId, db: AppController.shared.databaseController) {
                    let itemId = item.id
                    do {
                        let deleted = try item.markDeleted()
                        if deleted == true {
                            selectedItems.remove(itemId)
                        }
                    } catch let error {
                        print("Error deleting item: ", error)
                    }
                }
            }
            
            context.selectedItems = selectedItems
        }
        
        completion?(nil)
    }
}


struct CVUAction_SelectAll: CVUAction {
    var vars: [String: CVUValue]
    
    var defaultVars: [String: CVUValue] { [
        "title": .constant(.string("Select All")),
    ]}
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        guard let context = sceneController.topMostContext else {
            return
        }
        
        context.selectAll()
        completion?(nil)
    }
}

struct CVUAction_DeselectAll: CVUAction {
    var vars: [String: CVUValue]
    
    var defaultVars: [String: CVUValue] { [
        "title": .constant(.string("Deselect All")),
    ]}
    
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        guard let context = sceneController.topMostContext else {
            return
        }
        
        context.deselectAll()
        completion?(nil)
    }
}

struct CVUAction_Link: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        let lookup = CVULookupController()
        let db = sceneController.appController.databaseController

        guard let currentItem = context.currentItem,
              let currentItemRowId = currentItem.rowId,
           let subjectVal = vars["subject"],
           let subjectItem: ItemRecord = lookup.resolve(value: subjectVal, context: context, db: db),
           let subjectItemRowId = subjectItem.rowId,
           let edgeTypeVal = vars["edgeType"],
           let edgeType: String = lookup.resolve(value: edgeTypeVal, context: context, db: db),
           let distinctVal = vars["distinct"],
           let unique: Bool = lookup.resolve(value: distinctVal, context: context, db: db) else {
            return
        }
        
        if unique {
            for currentEdge in subjectItem.edges(edgeType) {
                if currentEdge.name == edgeType {
                    guard let result = try? currentEdge.markDeleted(), result == true else {
                        print("ERROR CVUAction_Unlink: item: \(subjectItem.type) with id: \(subjectItem.id) edge item: \(currentEdge.item)")
                        return
                    }
                }
            }
        }
        
        let _ = try? EdgeRecord.createFor(source: subjectItemRowId, type: edgeType, target: currentItemRowId)
        
        sceneController.scheduleUIUpdate()
        completion?(nil)
    }
}

struct CVUAction_Unlink: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        let lookup = CVULookupController()
        let db = sceneController.appController.databaseController

        guard let currentItem = context.currentItem,
           let subjectVal = vars["subject"],
           let subjectItem: ItemRecord = lookup.resolve(value: subjectVal, context: context, db: db),
           let edgeTypeVal = vars["edgeType"],
           let edgeType: String = lookup.resolve(value: edgeTypeVal, context: context, db: db) else {
            return
        }
        
        // todomigrate: Make sure this passes
        var edge: EdgeRecord? = nil
        for currentEdge in subjectItem.edges(edgeType) {
            if currentEdge.source == subjectItem.rowId && currentEdge.target == currentItem.rowId && currentEdge.name == edgeType {
                edge = currentEdge
                break
            }
        }
        
        
        guard let edgeToUnlink = edge else {
            return
        }
        
        guard let result = try? edgeToUnlink.markDeleted(), result == true else {
            print("ERROR CVUAction_Unlink: item: \(subjectItem.type) with id: \(subjectItem.id) edge item: \(edgeToUnlink.item)")
            return
        }
        
        sceneController.scheduleUIUpdate()
        completion?(nil)
    }
}

struct CVUAction_Star: CVUAction {
    var vars: [String: CVUValue]
    var defaultVars: [String: CVUValue] { [
        "icon": .constant(.string("star")),
    ]}

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        guard let currentItem = context.currentItem else {
            return
        }
        
        let prop = "starred"
        let currentVal = currentItem.propertyValue(prop)?.asBool() ?? false
        do {
            try currentItem.setPropertyValue(name: prop, value: .bool(!currentVal))
            sceneController.scheduleUIUpdate()
        } catch let error {
            print("ERROR CVUAction_Star: item: \(currentItem.type) with id: \(currentItem.id) error: \(error)")
        }
        
        completion?(nil)
    }
    
    func getString(_ key: String, _ context: CVUContext, _ viewContext: ViewContextController?) -> String? {
        guard key == "icon",
              let currentItem = context.currentItem else {
            return nil
        }
               
        let starred = currentItem.propertyValue("starred")?.asBool() ?? false
        return starred ? "star.fill" : "star"
    }
}

struct CVUAction_ClosePopup: CVUAction {
    @EnvironmentObject var sceneController: SceneController
    var vars: [String: CVUValue]

    init(vars: [String : CVUValue]) {
        self.vars = vars
    }

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        sceneController.closeLastInStack()
        completion?(nil)
    }
}

struct CVUAction_ToggleFilterPanel: CVUAction {
    @EnvironmentObject var sceneController: SceneController
    init(vars: [String : CVUValue]) {
        self.vars = vars
    }
    
    var vars: [String: CVUValue]
    var defaultVars: [String: CVUValue] { [
        "icon": .constant(.string("rhombus")),
    ]}

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        sceneController.filterPanelIsVisible = !sceneController.filterPanelIsVisible
        completion?(nil)
    }
}

struct CVUAction_ShowContextPane: CVUAction {
    var vars: [String: CVUValue]
    var defaultVars: [String: CVUValue] { [
        "icon": .constant(.string("ellipsis")),
    ]}

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        guard let item = context.currentItem,
              let _ = CVUController().viewDefinitionFor(itemRecord: item) else {
            return
        }
        
        sceneController.contextPaneVisible = !sceneController.contextPaneVisible
        completion?(nil)
    }
}


struct CVUAction_SetProperty: CVUAction {
    var vars: [String: CVUValue]
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        let lookup = CVULookupController()
        let db = sceneController.appController.databaseController
        let schema = AppController.shared.databaseController.schema

        guard let subjectVal = vars["subject"],
           let subjectItem: ItemRecord = lookup.resolve(value: subjectVal, context: context, db: db),
           case let .constant(.string(property)) = vars["property"],
           let expectedType = schema.expectedType(forItemType: subjectItem.type, propertyName: property),
           let value = vars["value"],
           let databaseValue = PropertyDatabaseValue(cvuValue: value, propertyType: expectedType) else {
            return
        }
        
        try? subjectItem.setPropertyValue(name: property, value: databaseValue)
        completion?(nil)
    }
}

struct CVUAction_RequestContactsPermission: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        AppController.shared.permissionController.requestContacts { result in
            completion?(result)
        }
    }
}

struct CVUAction_RequestLocationPermission: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        AppController.shared.permissionController.requestLocation { result in
            completion?(result)
        }
    }
}

struct CVUAction_RequestCameraPermission: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        AppController.shared.permissionController.requestCamera { result in
            completion?(result)
        }
    }
}

struct CVUAction_RequestPhotoLibraryPermission: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        AppController.shared.permissionController.requestPhotos { result in
            completion?(result)
        }
    }
}


struct CVUAction_RequestStoragePermission: CVUAction {
    var vars: [String: CVUValue]

    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        // Noop since iOS doesn't have any permission for storage (Apps are sandboxed so always have storage access)
        completion?(nil)
    }
}

struct CVUAction_OpenPopup: CVUAction {
    var vars: [String: CVUValue]

    func setPopupSettings(sceneController: SceneController, context: CVUContext) -> [String: Any]? {
        let lookupController = CVULookupController()
        let db = sceneController.appController.databaseController
        guard let titleProp = vars["title"],
              let textProp = vars["text"],
              let title: String = lookupController.resolve(value: titleProp, context: context, db: db),
              let text: String = lookupController.resolve(value: textProp, context: context, db: db) else {
            return nil
        }

        let resolver = CVUPropertyResolver(context: context, lookup: lookupController, db: db, properties: vars)
        let resolvedActions = resolver.actions(key: "actions") ?? []
        return ["title": title, "text": text, "actions": resolvedActions]
    }
    
    func execute(sceneController: SceneController, context: CVUContext, completion: ((Any?) -> ())?) {
        completion?(nil)
    }
}

struct CVUAction_Noop: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_Schedule: CVUAction {
    var vars: [String: CVUValue]
    var defaultVars: [String: CVUValue] { [
        "icon": .constant(.string("alarm")),
    ]}
}

// TODO:

struct CVUAction_OpenGroup: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_ShowNavigation: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_Duplicate: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_showSessionSwitcher: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_ForwardToFront: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_BackAsSession: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_OpenSession: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_OpenSessionByName: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_MultiAction: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_RunIndexer: CVUAction {
    var vars: [String: CVUValue]
}

struct CVUAction_SetSetting: CVUAction {
    var vars: [String: CVUValue]
}
