//
// CVUParser.swift
// Copyright © 2020 memri. All rights reserved.

import Foundation
import SwiftUI

class CVUParser {
    let tokens: [CVUToken]
    var index = 0
    var lastToken: CVUToken?

    init(
        _ tokens: [CVUToken]
    ) {
        self.tokens = tokens
    }

    func peekCurrentToken() -> CVUToken {
        index >= tokens.count
            ? CVUToken.EOF
            : tokens[index]
    }

    func popCurrentToken() -> CVUToken {
        if index >= tokens.count {
            lastToken = CVUToken.EOF
            return CVUToken.EOF
        }

        lastToken = tokens[index]
        index += 1
        return lastToken! // Check for out of bound?
    }

    func parse() throws -> [CVUParsedDefinition] {
        index = 0
        var result = [CVUParsedDefinition]()

        while true {
            if case CVUToken.EOF = peekCurrentToken() { return result }
            if case CVUToken.Newline = peekCurrentToken() {
                _ = popCurrentToken()
                continue
            }

            var dsl = try parseViewDSL()
            if dsl["sessions"] != nil {
                dsl.type = .sessions
            }
            else if dsl["views"] != nil {
                dsl.type = .views
            }

            result.append(dsl)
        }

        return result
    }

    func parseViewDSL() throws -> CVUParsedDefinition {
        let node = try parsePrimary()

        if case CVUToken.Colon = peekCurrentToken() {
            _ = popCurrentToken()
        }

        return try parseDefinition(node)
    }

    func parsePrimary(_: Bool = false) throws -> CVUParsedDefinition {
        switch peekCurrentToken() {
        case .Identifier:
            return try parseIdentifierSelector()
        case .NamedIdentifier:
            return try parseNamedIdentifierSelector()
        case .BracketOpen:
            return try parseBracketsSelector()
        case .String:
            return try parseStringSelector()
        default:
            throw CVUParseErrors.ExpectedDefinition(popCurrentToken())
        }
    }

    func parseIdentifierSelector() throws -> CVUParsedDefinition {
        // Example: Person {
        guard case var CVUToken.Identifier(typeIdentifier, _, _) = popCurrentToken() else {
            throw CVUParseErrors.ExpectedIdentifier(lastToken!)
        }

        // Example: Person[name = 'john']
        if case CVUToken.BracketOpen = peekCurrentToken() {
            _ = popCurrentToken()
            if case CVUToken.BracketClose = peekCurrentToken() {
                _ = popCurrentToken()
                typeIdentifier += "[]"
            }
            else {
                // TODO:
            }
        }
        
        if case CVUToken.Caret(_, _) = peekCurrentToken() {
            _ = popCurrentToken()
            guard case let CVUToken.Identifier(renderer, _, _) = popCurrentToken() else {
                throw CVUParseErrors.ExpectedIdentifier(lastToken!)
            }
            return CVUParsedDefinition(type: .uiNode, selector: typeIdentifier, renderer: renderer)
        }
        
        return CVUParsedDefinition(type: .view, selector: typeIdentifier.trimmingCharacters(in: ["."]))
    }

    func parseNamedIdentifierSelector() throws -> CVUParsedDefinition {
        // Example: "Some Name" {
        guard case let CVUToken.NamedIdentifier(name, _, _) = popCurrentToken() else {
            throw CVUParseErrors.UnexpectedToken(lastToken!)
        }

        return CVUParsedDefinition(type: .view, selector: ".\(name)", name: name)
    }

    // For JSON support
    func parseStringSelector() throws -> CVUParsedDefinition {
        guard case let CVUToken.String(value, _, _) = popCurrentToken() else {
            throw CVUParseErrors.UnexpectedToken(lastToken!)
        }

        if value.first == "." {
            return CVUParsedDefinition(type: .view,
                                       selector: value,
                                       name: String(value.suffix(from: value.index(value.startIndex, offsetBy: 1))))
        }
        else if value.first == "[" {
            throw StringError(description: "Not supported yet") // TODO:
        }
        else {
            return CVUParsedDefinition(type: .view, selector: value)
        }
    }

    func parseBracketsSelector(_ token: CVUToken? = nil) throws -> CVUParsedDefinition {
        guard case CVUToken.BracketOpen = (token ?? popCurrentToken()) else {
            throw CVUParseErrors.ExpectedCharacter("[", lastToken!)
        }
        let typeToken = token ?? lastToken!

        guard case let CVUToken.Identifier(type, _, _) = popCurrentToken() else {
            throw CVUParseErrors.ExpectedIdentifier(lastToken!)
        }

        // TODO: Only allow inside other definition
        if ["session", "view"].contains(type), case CVUToken.BracketClose = peekCurrentToken() {
            _ = popCurrentToken()
            switch type {
            case "session": return CVUParsedDefinition(selector: "[session]")
            case "view": return CVUParsedDefinition(selector: "[view]")
            default: _ = 1 // Can never get here
            }
        }

        guard case let CVUToken.Operator(op, _, _) = popCurrentToken() else {
            throw CVUParseErrors.ExpectedCharacter("=", lastToken!)
        }

        if case CVUOperator.ConditionEquals = op {
            var name: String

            if case let CVUToken.String(nm, _, _) = popCurrentToken() { name = nm }
            else if case let CVUToken.Identifier(nm, _, _) = lastToken! { name = nm }
            else {
                throw CVUParseErrors.ExpectedString(lastToken!)
            }

            guard case CVUToken.BracketClose = popCurrentToken() else {
                throw CVUParseErrors.ExpectedCharacter("]", lastToken!)
            }

            switch type {
            case "sessions": return CVUParsedDefinition(type: .sessions, selector: "[sessions = \(name)]", name: name)
            case "session": return CVUParsedDefinition(type: .views, selector: "[session = \(name)]", name: name)
            case "view": return CVUParsedDefinition(type: .view, selector: "[view = \(name)]", name: name)
            case "datasource": return CVUParsedDefinition(type: .datasource, selector: "[datasource = \(name)]",
                                                                    name: name)
            case "renderer": return CVUParsedDefinition(type: .renderer, selector: "[renderer = \(name)]", name: name)
            case "language": return CVUParsedDefinition(type: .language, selector: "[language = \(name)]", name: name)
            default:
                throw CVUParseErrors.UnknownDefinition(type, typeToken)
            }
        }
        else {
            throw CVUParseErrors.ExpectedCharacter("=", lastToken!)
        }
    }

    func createExpression(_ code: String, startInStringMode: Bool = false) throws -> CVUValue.ExpressionNode {
        try CVUValue.ExpressionNode(code: code, stringMode: startInStringMode)
    }

    func parseDict(_: String? = nil) throws -> CVUDefinitionContent {
        var parsedContent = CVUDefinitionContent()
        
        var stack = [CVUValue]()

        var lastKey: String?
        var isArrayMode = false

        func setPropertyValue() {
            if let lastKey = lastKey {
                if isArrayMode || stack.count > 1 {
                    parsedContent.properties[lastKey] = .array(stack)
                } else if let first = stack.first {
                    parsedContent.properties[lastKey] = first
                }
                stack = []
            }
        }

        func addUIElement(_ type: CVUUIElementFamily, _ properties: CVUDefinitionContent) {
            parsedContent.children.append(CVUUINode(type: type,
                                                 children: properties.children,
                                                 properties: properties.properties))
        }

        while true {
//            print(peekCurrentToken())

            switch popCurrentToken() {
            case let .Bool(v, _, _):
                stack.append(CVUValue.constant(.bool(v)))
            case .BracketOpen:
                if stack.count == 0, lastKey != nil {
                    isArrayMode = true
                }
                else {
                    setPropertyValue()

                    var definition = try parseBracketsSelector(lastToken)
                    definition = try parseDefinition(definition)
                    parsedContent.definitions.append(definition)
                }
            case .BracketClose:
                if isArrayMode {
                    setPropertyValue()
                    isArrayMode = false
                    lastKey = nil
                }
                else {
                    throw CVUParseErrors.UnexpectedToken(lastToken!) // We should never get here
                }
            case .CurlyBracketOpen:
                guard let lastKey = lastKey else {
                    throw CVUParseErrors.ExpectedIdentifier(lastToken!)
                }

                stack.append(.subdefinition(try parseDict(lastKey)))
            case .CurlyBracketClose:
                setPropertyValue()
                return parsedContent // DONE
            case .Colon:
                throw CVUParseErrors.ExpectedKey(lastToken!)
            case let .Expression(v, _, _):
                try stack.append(CVUValue.expression(createExpression(v)))
            case let .Identifier(value, _, _):
                if lastKey == nil {
                    var nextToken = peekCurrentToken()
                    if case CVUToken.Colon = nextToken {
                        _ = popCurrentToken()
                        lastKey = value
                        nextToken = peekCurrentToken()
                    }

                    if lastKey == nil, let type = CVUUIElementFamily(rawValue: value.lowercased()) {
                        var properties = CVUDefinitionContent()
                        if case CVUToken.CurlyBracketOpen = nextToken {
                            _ = popCurrentToken()
                            properties = try parseDict(value)
                        }
                        addUIElement(type, properties)
                    }
                    else if value == "userstate" || value == "viewarguments" || value ==
                        "contextpane"
                    {
                        if case CVUToken.CurlyBracketOpen = nextToken {
                            _ = popCurrentToken()
                            let properties = try parseDict()
                            stack.append(.subdefinition(properties))
                        }
                    }
                    else if case CVUToken.CurlyBracketOpen = nextToken {
                        lastKey = value
                    } else if case CVUToken.Caret = nextToken {
                        index -= 1 // TODO better way of
                        let identifierNode = try parseIdentifierSelector()
                        parsedContent.definitions.append(try parseDefinition(identifierNode))
                    }
                }
//                else if let name = knownActions[value.lowercased()] {
//                    var options: [String: Any?] = [:]
//                    outerLoop: while true {
//                        switch peekCurrentToken() {
//                        case .Comma:
//                            if isArrayMode { _ = popCurrentToken() }
//                        case .CurlyBracketOpen:
//                            _ = popCurrentToken()
//                            options = try parseDict()
//                        default:
//                            break outerLoop
//                        }
//                    }
////
////                    if let actionFamily = ActionFamily(rawValue: name) {
////                        let ActionType = ActionFamily.getType(actionFamily)()
////                        // swiftformat:disable:next redundantInit
////                        stack.append(ActionType.init(context, values: options))
////                    }
////                    else {
////                        // TODO: ERROR REPORTING
////                    }
//                }
                else {
                    stack.append(CVUValue.constant(.argument(value)))
                }
            case .Newline:
                if stack.count == 0 { continue }
                else { fallthrough }
            case .Comma:
                if isArrayMode { continue } // IGNORE
                fallthrough
            case .SemiColon:
                setPropertyValue()
                lastKey = nil
            case .Nil:
                stack.append(CVUValue.constant(.nil))
            case let .Number(value, _, _):
                stack.append(CVUValue.constant(.number(value)))
            case let .String(value, _, _):
                if !isArrayMode,
                   case CVUToken.Colon = peekCurrentToken()
                {
                    setPropertyValue() // TODO: Is this every necessary?
                    _ = popCurrentToken()
                    lastKey = value
                }
                else if lastKey == nil { lastKey = value }
                else { stack.append(CVUValue.constant(.string(value))) }
            case let .StringExpression(v, _, _):
                try stack.append(CVUValue.expression(createExpression(v, startInStringMode: true)))
            case .Color(let string, _, _):
                stack.append(CVUValue.constant(.colorHex(string)))
            default:
                throw CVUParseErrors.UnexpectedToken(lastToken!)
            }
        }
    }

    func parseDefinition(_ selector: CVUParsedDefinition) throws -> CVUParsedDefinition {
        while true {
            if case CVUToken.Newline = peekCurrentToken() {
                _ = popCurrentToken()
            }
            else {
                guard case CVUToken.CurlyBracketOpen = popCurrentToken() else {
                    throw CVUParseErrors.ExpectedCharacter("{", lastToken!)
                }

                break
            }
        }
        var parsedVersion = selector
        parsedVersion.parsed = try parseDict()
        return parsedVersion
    }
}
