//
//  CVULookup.swift
//  MemriDatabase
//
//  Created by T Brennan on 23/12/20.
//

import Foundation
import InflectorKit

/// This struct can be used to resolve CVU values to a final value of the desired type.
/// For lookups you must provide a CVUContext which contains required information on the default item, viewArguments, etc to be used in the lookup.
struct CVULookupController {
    init(mockMode: LookupMock? = nil) {
        self.lookupMockMode = mockMode
    }
    
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> Int? {
        switch value {
        case .constant(let constant):
            return constant.asInt()
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> Double? {
        switch value {
        case .constant(let constant):
            return constant.asNumber()
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> String? {
        switch value {
        case .constant(let constant):
            return constant.asString()
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> Bool? {
        switch value {
        case .constant(let constant):
            return constant.asBool()
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> Date? {
        switch value {
        case .constant(let constant):
            return constant.asNumber().flatMap { DatabaseHelperFunctions.decode(from: Int($0)) }
        case .expression(let expression):
            return resolve(expression: expression, context: context, type: Double.self, db: db).flatMap { DatabaseHelperFunctions.decode(from: Int($0)) }
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> ItemRecord? {
        switch value {
        case .constant:
            return nil
        case .item(let itemId):
            return ItemRecord.fetchWithID(itemId)
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return nil
        }
    }
    
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> [ItemRecord] {
        switch value {
        case .item(let itemId):
            return ItemRecord.fetchWithID(itemId).map { [$0] } ?? []
        case .expression(let expression):
            return resolve(expression: expression, context: context, db: db)
        default:
            return []
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> [CVUValue] {
        switch value {
        case .array(let values) :
            return values
        default:
            return []
        }
    }
    
    func resolve(value: CVUValue, context: CVUContext, db: DatabaseController) -> [[String: [CVUValue.Constant]]] {
        switch value {
        case .array(let values) :
            return values.compactMap{$0.getSubdefinition()?.properties.compactMapValues{
                return resolve(value: $0)
                }
            }
        default:
            return []
        }
    }
    
    func resolve(value: CVUValue) -> [CVUValue.Constant]? {
        switch value {
        case .constant(let constant):
            return [constant]
        case .array(let values):
            return values.flatMap{return resolve(value: $0) ?? []}
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, defaultValue: Bool, context: CVUContext, db: DatabaseController) -> Binding<Bool>? {
        switch value {
        case let .expression(.lookup(nodes)):
            guard let (item, property) = resolveToItemAndProperty(nodes: nodes, context: context, db: db)
            else { return nil }
            return item.propertyBinding(name: property, defaultValue: defaultValue)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, defaultValue: String, context: CVUContext, db: DatabaseController) -> Binding<String>? {
        switch value {
        case let .expression(.lookup(nodes)):
            guard let (item, property) = resolveToItemAndProperty(nodes: nodes, context: context, db: db)
            else { return nil }
            return item.propertyBinding(name: property, defaultValue: defaultValue)
        default:
            return nil
        }
    }
    
    func resolve(value: CVUValue, defaultValue: String?, context: CVUContext, db: DatabaseController) -> Binding<String?>? {
        switch value {
        case let .expression(.lookup(nodes)):
            guard let (item, property) = resolveToItemAndProperty(nodes: nodes, context: context, db: db)
            else { return nil }
            return item.propertyBinding(name: property, defaultValue: defaultValue)
        default:
            return nil
        }
    }
    
    /// Lookup an edge from an ItemRecord
    func resolve(edge: String, for item: ItemRecord, db: DatabaseController) -> ItemRecord? {
        item.edgeItem(edge, db: db)
    }
    
    /// Lookup an edge array from an ItemRecord
    func resolve(edge: String, for item: ItemRecord, db: DatabaseController) -> [ItemRecord] {
        item.edgeItems(edge, db: db)
    }
    
    /// Lookup a property from an ItemRecord
    func resolve(property: String, for item: ItemRecord, db: DatabaseController) -> PropertyDatabaseValue? {
        return item.propertyValue(property, db: db)
    }

    /// Used for tests to mock a lookup result
    struct LookupMock {
        var bool: Bool
        var string: String
        var double: Double
        var int: Int
    }
    
    /// Used for tests to mock a lookup result
    let lookupMockMode: LookupMock?
    
    /// Represents the result of an intermediate step in the lookup
    private enum LookupStep {
        case items([ItemRecord])
        case values([PropertyDatabaseValue])
    }
    
    private func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> LookupStep? {
        var currentValue: LookupStep?
        for node in nodes {
            switch node.type {
            case .default:
                guard let currentItem = context.currentItem else { return nil }
                currentValue = .items([currentItem])
            case .localized:
                let lang = Languages()
                lang.load(CVUController().definitions)
                if let name = nodes.last?.name,
                   case let .constant(.string(value)) = lang.keywords[name] {
                   return .values([.string(value)])
                }
                return nil
            case .function(let args):
                switch node.name.lowercased() {
                case "item":
                    if let exp = args.first,
                       let itemUIDString = resolve(expression: exp, context: context, type: String.self, db: db),
                       let item = ItemRecord.fetchWithID(StringUUID(uid: itemUIDString), db: db)
                    {
                        currentValue = .items([item])
                    } else {
                        return nil
                    }
                case "joined":
                    guard case let .values(values) = currentValue else { return nil }
                    if let exp = args.first, let separator = resolve(expression: exp, context: context, type: String.self, db: db) {
                        let joined = values.compactMap { $0.asString()?.nilIfBlank }.joined(separator: separator)
                        currentValue = .values([.string(joined)])
                    } else {
                        let strings = values.compactMap { $0.asString()?.nilIfBlank }
                        let joined = ListFormatter.localizedString(byJoining: strings)
                        currentValue = .values([.string(joined)])
                    }
                case "joinwithcomma":
                    let strings = args.compactMap {
                        resolve(expression: $0, context: context, type: String.self, db: db)?.nilIfBlank
                    }
                    let joined = strings.joined(separator: ", ")
                    currentValue = .values([.string(joined)])
                case "joinnatural":
                    let strings = args.compactMap {
                        resolve(expression: $0, context: context, type: String.self, db: db)?.nilIfBlank
                    }
                    let joined = ListFormatter.localizedString(byJoining: strings)
                    currentValue = .values([.string(joined)])
                case "plainstring":
                    guard case let .values(values) = currentValue else { return nil }
                    let stripped = values.compactMap { value -> PropertyDatabaseValue? in
                        guard let htmlstring = value.asString() else { return nil }
                        return .string(HTMLHelper.getPlainText(html: htmlstring))
                    }
                    currentValue = .values(stripped)
                case "first":
                    if case let .values(values) = currentValue, let first = values.first { currentValue = .values([first]) }
                    else if case let .items(items) = currentValue, let first = items.first { currentValue = .items([first]) }
                    else { return nil }
                case "last":
                    if case let .values(values) = currentValue, let last = values.last { currentValue = .values([last]) }
                    else if case let .items(items) = currentValue, let last = items.last { currentValue = .items([last]) }
                    else { return nil }
                case "fullname":
                    switch currentValue {
                    case let .items(items):
                        if let first = items.first, first.type == "Person" {
                            let name = [
                                resolve(property: "firstName", for: first, db: db),
                                resolve(property: "lastName", for: first, db: db)
                            ].compactMap { $0?.asString() }.joined(separator: " ")
                            currentValue = .values([.string(name)])
                        } else {
                            return nil
                        }
                    default:
                        return nil
                }
                case "initials":
                    switch currentValue {
                    case let .values(values):
                        let initials = String(values.compactMap { $0.asString()?.first }).uppercased()
                        currentValue = .values([.string(initials)])
                    case let .items(items):
                        if let first = items.first, first.type == "Person" {
                            let initials = String([
                                resolve(property: "firstName", for: first, db: db),
                                resolve(property: "lastName", for: first, db: db)
                            ].compactMap { $0?.asString()?.first }).uppercased()
                            currentValue = .values([.string(initials)])
                        } else {
                            return nil
                        }
                    case .none:
                        return nil
                }
                case "describechangelog":
                    switch currentValue {
                    case let .items(items):
                        if let first = items.first, first.type == "Person" {
                            let dateCreated = first.dateCreated.formatted()
                            let timeSinceCreated = first.dateCreated.timeDelta
                            return .values([.string("You created this \(first.type) \(dateCreated) over the past \(timeSinceCreated ?? "")")])
                        } else {
                            return nil
                        }
                    default:
                        return nil
                }
                case "plural":
                    guard case let .values(values) = currentValue else { return nil }
                    let pluralized = values.map { value -> PropertyDatabaseValue in
                        guard let string = value.asString() else { return value }
                        return .string(string.pluralized)
                    }
                    
                    currentValue = .values(pluralized)
                case "camelcasetowords":
                    guard case let .values(values) = currentValue else { return nil }
                    let camelCased = values.map { value -> PropertyDatabaseValue in
                        guard let string = value.asString() else { return value }
                        return .string(string.camelCaseToWords())
                    }
                    
                    currentValue = .values(camelCased)
                case "titlecase":
                    guard case let .values(values) = currentValue else { return nil }
                    let titleCased = values.map { value -> PropertyDatabaseValue in
                        guard let string = value.asString() else { return value }
                        return .string(string.titleCase())
                    }
                    
                    currentValue = .values(titleCased)
                case "firstuppercased":
                    guard case let .values(values) = currentValue else { return nil }
                    let upperCased = values.map { value -> PropertyDatabaseValue in
                        guard let string = value.asString() else { return value }
                        return .string(string.firstUppercased)
                    }
                    currentValue = .values(upperCased)
                case "count":
                    switch currentValue {
                    case let .items(items):
                        currentValue = .values([.int(items.count)])
                    case let .values(values):
                        currentValue = .values([.int(values.count)])
                    default:
                        break
                    }
                    
                default: return nil
                }
            case .lookup(let subexpression):
                switch currentValue {
                case .items(let items):
                    currentValue = itemLookup(node: node, items: items, subexpression: subexpression, context: context, db: db)
                case .none:
                        // Check if there is a matching view argument
                        if let viewArgs = context.viewArguments, let argValue = viewArgs.args[node.name] {
                            switch argValue {
                            case .constant(let constant):
                                switch constant {
                                case let .argument(x):
                                    currentValue = .values([.string(x)])
                                case let .number(x):
                                    currentValue = .values([.double(x)])
                                case let .int(x):
                                    currentValue = .values([.int(x)])
                                case let .string(x):
                                    currentValue = .values([.string(x)])
                                case let .bool(x):
                                    currentValue = .values([.bool(x)])
                                case let .colorHex(x):
                                    currentValue = .values([.string(x)])
                                case .nil:
                                    currentValue = nil
                                }
                            case .expression(let expression):
                                let context = CVUContext(currentItem: viewArgs.argumentItem, viewArguments: viewArgs.parentArguments)
                                if let item: ItemRecord = resolve(expression: expression, context: context, db: db) {
                                    currentValue = .items([item])
                                } else if let number: Double = resolve(expression: expression, context: context, db: db) {
                                    currentValue = .values([.double(number)])
                                } else if let string: String = resolve(expression: expression, context: context, db: db) {
                                    currentValue = .values([.string(string)])
                                } else {
                                    let items: [ItemRecord] = resolve(expression: expression, context: context, db: db)
                                        if !items.isEmpty {
                                            currentValue = .items(items)
                                        } else {
                                            return nil
                                        }
                                }
                            default:
                                return nil
                            }
                        } else if node.name == "me" {
                            if let me = ItemRecord.me {
                                currentValue = .items([me])
                            }
                        } else if node.name == "type", let item = context.currentItem {
                            currentValue = .values([.string(item.type)])
                        } else if node.name == "uid", let item = context.currentItem, let itemRowId = item.rowId {
                            currentValue = .values([.int(Int(itemRowId))])
                        } else if node.name == "items" {
                            currentValue = .items(context.items)
                        } else if node.name.lowercased() == "currentindex" {
                            currentValue = .values([.int(context.currentIndex+1)])
                        }

                default:
                    return nil
                }
            }
        }
        return currentValue
    }
    
    func filter(items: [ItemRecord], by subexpression: CVUValue.ExpressionNode?, db: DatabaseController) -> [ItemRecord] {
        guard let exp = subexpression else { return items }
        return items.filter { item in
            let context = CVUContext(currentItem: item)
            return resolve(expression: exp, context: context, db: db) ?? false
        }
    }

    private func itemLookup(node: CVUValue.LookupNode, items: [ItemRecord], subexpression: CVUValue.ExpressionNode?, context: CVUContext, db: DatabaseController) -> LookupStep? {
        let trimmedName = node.name.trimmingCharacters(in: ["~"])
        switch node.name.first {
        case "~":
            /// LOOKUP REVERSE EDGE FOR EACH ITEM
            if node.isArray {
                let result = items.flatMap { (item) -> [ItemRecord] in
                    return item.reverseEdgeItems(trimmedName, db: db)
                }
                return .items(filter(items: result, by: subexpression, db: db))
            } else {
                let result = items.compactMap { (item) -> ItemRecord? in
                    return item.reverseEdgeItem(trimmedName, dbController: db)
                }
                return .items(filter(items: result, by: subexpression, db: db))
            }
        default:
            /// CHECK IF WE'RE EXPECTING EDGES OR PROPERTIES
            guard let itemType = items.first?.type else {
                return nil
            }
            /// Check if this is an intrinsic property
            switch node.name {
            case "uid":
                return .values(items.compactMap {(item: ItemRecord) -> PropertyDatabaseValue? in
                    guard let itemRowId = item.rowId else {
                        return nil
                    }
                   
                    return .int(Int(itemRowId))
                })
            case "id":
                return .values(items.map { return .string($0.id.uid) })
            case "dateModified":
                return .values(items.map { return .datetime($0.dateModified) })
            case "dateCreated":
                return .values(items.map { return .datetime($0.dateCreated) })
            case "label":
                return .items(items.first?.edgeItems("label") ?? [])
            default: break
            }
                        
            /// Find out if it is a property or an edge (according to schema)
            switch db.schema.expectedType(forItemType: itemType, propertyOrEdgeName: node.name) {
            case .property:
                /// LOOKUP PROPERTY FOR EACH ITEM
                let result = items.compactMap { (item) -> PropertyDatabaseValue? in
                    guard let value = item.propertyValue(node.name, db: db) else { return nil}
                    return value
                }
                return .values(result)
            case .edge:
                /// LOOKUP EDGE FOR EACH ITEM
                if node.isArray {
                    let result = items.flatMap { (item) -> [ItemRecord] in
                        return item.edgeItems(trimmedName, db: db)
                    }
                    return .items(filter(items: result, by: subexpression, db: db))
                } else {
                    let result = items.compactMap { (item) -> ItemRecord? in
                        return item.edgeItem(trimmedName, db: db)
                    }
                    return .items(filter(items: result, by: subexpression, db: db))
                }
            case .none:
                return nil
            }
        }
    }
    
    /// Lookup a variable using its CVU string and return the value as a double
    private func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> Double? {
        if let mock = lookupMockMode { return mock.double }
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return nil }
        switch lookupResult {
        case .values(let values):
            return values.first?.asDouble()
        default:
            return nil
        }
    }
    
    /// Lookup a variable using its CVU string and return the value as a double
    private func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> Int? {
        if let mock = lookupMockMode { return mock.int }
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return nil }
        switch lookupResult {
        case .values(let values):
            return values.first?.asInt()
        default:
            return nil
        }
    }
    
    /// Lookup a variable using its CVU string and return the value as a string
    private func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> String? {
        if let mock = lookupMockMode { return mock.string }
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return nil }
        switch lookupResult {
        case .values(let values):
            return values.first?.asString()
        default:
            return nil
        }
    }
    
    /// Lookup a variable using its CVU string and return the value as a bool
    private func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> Bool? {
        if let mock = lookupMockMode { return mock.bool }
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return nil }
        switch lookupResult {
        case .values(let values):
            return values.first?.asBool()
        case .items(let items):
         return !items.isEmpty
        }
    }
    
    /// Lookup using a CVU expression string and return the value as an item
    private func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> ItemRecord? {
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return nil }
        switch lookupResult {
        case .items(let items):
            return items.first
        default:
            return nil
        }
    }
    
    /// Lookup using a CVU expression string and return the value as an array of items
    private func resolve(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> [ItemRecord] {
        guard let lookupResult: LookupStep = resolve(nodes: nodes, context: context, db: db) else { return [] }
        switch lookupResult {
        case .items(let items):
            return items
        default:
            return []
        }
    }
    
    
    private func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: ItemRecord.Type = ItemRecord.self, db: DatabaseController) -> ItemRecord? {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .or(a, b):
            return resolve(expression: a, context: context, type: ItemRecord.self, db: db) ?? resolve(expression: b, context: context, db: db)
        default:
            return nil
        }
    }
    
    private func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: [ItemRecord].Type = [ItemRecord].self, db: DatabaseController) -> [ItemRecord] {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .and(a, b):
            return resolve(expression: a, context: context, type: [ItemRecord].self, db: db) + resolve(expression: b, context: context, db: db)
        case let .or(a, b):
            let resolvedA = resolve(expression: a, context: context, type: [ItemRecord].self, db: db)
            return resolvedA.count > 0 ? resolvedA : resolve(expression: b, context: context, db: db)
        case let .stringMode(nodes):
            let first: CVUValue.ExpressionNode? = nodes.first
            let last: CVUValue.ExpressionNode? = nodes.last
            switch (first, last) {
            case let (.constant(query), .lookup(lookupNodes)):
                if let (item, property) = resolveToItemAndProperty(nodes: lookupNodes, context: context, db: db),
                   let value = property == "uid" ? PropertyDatabaseValue(databaseValue: item.id.databaseValue, propertyType: .string) : resolve(property: property, for: item, db: db) {
                    switch value {
                    case .string(let valueString):
                        return (try? db.search(string: "\(query.asString()) \(valueString)")) ?? []
                    default:
                        break
                    }
                }
                return []
            default:
                return []
            }
        default:
            return []
        }
    }
    
    private func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: Double.Type = Double.self, db: DatabaseController) -> Double? {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .or(a, b):
            return resolve(expression: a, context: context, db: db) ?? resolve(expression: b, context: context, db: db)
        case .negation(_):
            print("CVU Expression error: Should not use ! operator on non-boolean value")
            return nil
        case let .addition(a, b):
            return (resolve(expression: a, context: context, db: db) ?? 0) + (resolve(expression: b, context: context, db: db) ?? 0)
        case let .subtraction(a, b):
            return (resolve(expression: a, context: context, db: db) ?? 0) - (resolve(expression: b, context: context, db: db) ?? 0)
        case let .constant(const):
            return const.asNumber()
        case let .multiplication(a, b):
            return (resolve(expression: a, context: context, db: db) ?? 0) * (resolve(expression: b, context: context, db: db) ?? 0)
        case let .division(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Double.self, db: db),
                  rhs != 0
            else { return nil }
            return lhs / rhs
        default:
            return nil
        }
    }
    
    private func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: Int.Type = Int.self, db: DatabaseController) -> Int? {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .or(a, b):
            return resolve(expression: a, context: context, db: db) ?? resolve(expression: b, context: context, db: db)
        case .negation(_):
            print("CVU Expression error: Should not use ! operator on non-boolean value")
            return nil
        case let .addition(a, b):
            return (resolve(expression: a, context: context, db: db) ?? 0) + (resolve(expression: b, context: context, db: db) ?? 0)
        case let .subtraction(a, b):
            return (resolve(expression: a, context: context, db: db) ?? 0) - (resolve(expression: b, context: context, db: db) ?? 0)
        case let .constant(const):
            return const.asInt()
        case let .multiplication(a, b):
            return (resolve(expression: a, context: context, db: db) ?? 0) * (resolve(expression: b, context: context, db: db) ?? 0)
        case let .division(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Int.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Int.self, db: db),
                  rhs != 0
            else { return nil }
            return lhs / rhs
        default:
            return nil
        }
    }
    
    private func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: String.Type = String.self, db: DatabaseController) -> String? {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .or(a, b):
            return resolve(expression: a, context: context, type: String.self, db: db)?.nilIfBlank ?? resolve(expression: b, context: context, db: db)
        case .negation(_):
            print("CVU Expression error: Should not use ! operator on non-boolean value")
            return nil
        case let .addition(a, b):
            return (resolve(expression: a, context: context, db: db) ?? "") + (resolve(expression: b, context: context, db: db) ?? "")
        case .subtraction(_, _):
            print("CVU Expression error: Should not use - operator on string value")
            return nil
        case let .constant(const):
            return const.asString()
        case let .stringMode(nodes: nodes):
            return nodes.compactMap { resolve(expression: $0, context: context, db: db) }.joined(separator: "")
        default:
            return nil
        }
    }
    
    private func resolve(expression: CVUValue.ExpressionNode, context: CVUContext, type: Bool.Type = Bool.self, db: DatabaseController) -> Bool? {
        switch expression {
        case let .lookup(nodes):
            return resolve(nodes: nodes, context: context, db: db)
        case let .conditional(condition, trueExp, falseExp):
            let conditionResolved = resolve(expression: condition, context: context, type: Bool.self, db: db) ?? false
            if conditionResolved {
                return resolve(expression: trueExp, context: context, db: db)
            } else {
                return resolve(expression: falseExp, context: context, db: db)
            }
        case let .and(a, b):
            return (resolve(expression: a, context: context, db: db) ?? false) && (resolve(expression: b, context: context, db: db) ?? false)
        case let .or(a, b):
            return (resolve(expression: a, context: context, db: db) ?? false) || (resolve(expression: b, context: context, db: db) ?? false)
        case let .negation(x):
            return resolve(expression: x, context: context, type: Bool.self, db: db).map { !$0 }
        case .addition(_ ,_):
            print("CVU Expression error: Should not use + operator on bool value")
            return nil
        case .subtraction(_, _):
            print("CVU Expression error: Should not use - operator on bool value")
            return nil
        case let .constant(const):
            return const.asBool()
        case let .lessThan(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Double.self, db: db)
            else { return nil }
            return lhs < rhs
        case let .greaterThan(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Double.self, db: db)
            else { return nil }
            return lhs > rhs
        case let .lessThanOrEqual(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Double.self, db: db)
            else { return nil }
            return lhs <= rhs
        case let .greaterThanOrEqual(a, b):
            guard let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
                  let rhs = resolve(expression: b, context: context, type: Double.self, db: db)
            else { return nil }
            return lhs >= rhs
        case let .areEqual(a, b):
            if let lhs = resolve(expression: a, context: context, type: ItemRecord.self, db: db),
               let rhs = resolve(expression: b, context: context, type: ItemRecord.self, db: db) {
                return lhs.rowId == rhs.rowId
            }
            else if let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
               let rhs = resolve(expression: b, context: context, type: Double.self, db: db) {
                return lhs == rhs
            } else if let lhs = resolve(expression: a, context: context, type: String.self, db: db),
                      let rhs = resolve(expression: b, context: context, type: String.self, db: db) {
                return lhs == rhs
            } else if let lhs = resolve(expression: a, context: context, type: Bool.self, db: db),
                      let rhs = resolve(expression: b, context: context, type: Bool.self, db: db) {
                return lhs == rhs
            } else { return false }
        case let .areNotEqual(a, b):
            if let lhs = resolve(expression: a, context: context, type: Double.self, db: db),
               let rhs = resolve(expression: b, context: context, type: Double.self, db: db) {
                return lhs != rhs
            } else if let lhs = resolve(expression: a, context: context, type: String.self, db: db),
                      let rhs = resolve(expression: b, context: context, type: String.self, db: db) {
                return lhs != rhs
            } else if let lhs = resolve(expression: a, context: context, type: Bool.self, db: db),
                      let rhs = resolve(expression: b, context: context, type: Bool.self, db: db) {
                return lhs != rhs
            } else { return true }
        default:
            return nil
        }
    }
}

import SwiftUI
extension CVULookupController {
    func resolveToItemAndProperty(nodes: [CVUValue.LookupNode], context: CVUContext, db: DatabaseController) -> (item: ItemRecord, property: String)? {
        var currentItem: ItemRecord?
        
        // Find the item referenced by the lookup
        for node in nodes.dropLast() {
            switch node.type {
            case .default:
                guard let defaultItem = context.currentItem else { return nil }
                currentItem = defaultItem
            case .localized:
                guard let defaultItem = context.currentItem else { return nil }
                currentItem = defaultItem
            case .lookup(let subexpression):
                if let nextItem = currentItem {
                    let step = itemLookup(node: node, items: [nextItem], subexpression: subexpression, context: context, db: db)
                    if case let .items(items) = step {
                        currentItem = items.first
                    }
                }
            case .function:
            return nil
            }
        }
        
        guard let targetItem = currentItem, let propertyLookup = nodes.last else { return nil }
        // Make a binding to the right property
        return (item: targetItem, property: propertyLookup.name)
    }
}
