//
//  CVUPropertyResolver.swift
//  MemriDatabase
//
//  Created by T Brennan on 8/1/21.
//

import Foundation
import SwiftUI

struct CVUPropertyResolver {
    var context: CVUContext
    var lookup: CVULookupController
    var db: DatabaseController
    var properties: [String: CVUValue]
    
    func replacingItem(_ item: ItemRecord) -> CVUPropertyResolver {
        var result = self
        result.context.currentItem = item
        return result
    }
    
    func value(_ key: String) -> CVUValue? {
        properties[key]
    }
    
    func valueArray(_ key: String) -> [CVUValue] {
        guard let value = properties[key] else { return [] }
        guard case let .array(values) = value else {
            return [value]
        }
        return values
    }
    
    
    func subdefinition(_ key: String) -> CVUPropertyResolver? {
        guard let value = properties[key],
              case let .subdefinition(subdef) = value
        else {
            return nil
        }
        return CVUPropertyResolver(context: context, lookup: lookup, db: db, properties: subdef.properties)
    }
    
    func double(_ key: String) -> Double? {
        guard let value = value(key) else { return nil }
        return lookup.resolve(value: value, context: context, db: db)
    }
    
    func cgFloat(_ key: String) -> CGFloat? {
        guard let value = value(key) else { return nil }
        return (lookup.resolve(value: value, context: context, db: db) as Double?).map { CGFloat($0) }
    }
    
    func int(_ key: String) -> Int? {
        guard let value = value(key) else { return nil }
        return (lookup.resolve(value: value, context: context, db: db) as Double?).map { Int($0) }
    }
    
    func intArray(_ key: String) -> [Int] {
        return valueArray(key).compactMap { lookup.resolve(value: $0, context: context, db: db) }
    }
    
    func string(_ key: String) -> String? {
        guard let value = value(key) else { return nil }
        return lookup.resolve(value: value, context: context, db: db)
    }
    
    func stringArray(_ key: String) -> [String] {
        return valueArray(key).compactMap { lookup.resolve(value: $0, context: context, db: db) }
    }
    
    func bool(_ key: String) -> Bool? {
        guard let value = value(key) else { return nil }
        return lookup.resolve(value: value, context: context, db: db)
    }
    
    func bool(_ key: String, defaultValue: Bool, defaultValueForMissingKey: Bool? = nil) -> Bool {
        guard let value = value(key) else { return defaultValueForMissingKey ?? defaultValue }
        return lookup.resolve(value: value, context: context, db: db) ?? defaultValue
    }
    
    func dateTime(_ key: String) -> Date? {
        guard let value = value(key) else { return nil }
        return lookup.resolve(value: value, context: context, db: db)
    }
    
    func item(_ key: String) -> ItemRecord? {
        guard let value = value(key) else { return nil }
        return lookup.resolve(value: value, context: context, db: db)
    }
    
    func items(_ key: String) -> [ItemRecord] {
        guard let value = value(key) else { return [] }
        return lookup.resolve(value: value, context: context, db: db)
    }
    
    func items(_ key: String) -> [[String: [CVUValue.Constant]]] {
        guard let value = value(key) else { return [] }
        return lookup.resolve(value: value, context: context, db: db)
    }
  
    func items(_ key: String) -> [CVUValue] {
        guard let value = value(key) else { return [] }
        return lookup.resolve(value: value, context: context, db: db)
    }
    
    func edge(key: String, edgeName: String) -> ItemRecord? {
        guard let value = value(key) else { return nil }
        guard let item: ItemRecord = lookup.resolve(value: value, context: context, db: db) else { return nil }
        return lookup.resolve(edge: edgeName, for: item, db: db)
    }
    
    func property(key: String, propertyName: String) -> PropertyDatabaseValue? {
        guard let value = value(key) else { return nil }
        guard let item: ItemRecord = lookup.resolve(value: value, context: context, db: db) else { return nil }
        return lookup.resolve(property: propertyName, for: item, db: db)
    }
    
    func property(of item: ItemRecord, propertyName: String) -> PropertyDatabaseValue? {
        return lookup.resolve(property: propertyName, for: item, db: db)
    }
    
    func binding(for key: String, defaultValue: Bool = false) -> Binding<Bool>? {
        guard let value = value(key) else { return nil }
        return lookup.resolve(value: value, defaultValue: defaultValue, context: context, db: db)
    }
    
    func binding(for key: String, defaultValue: String) -> Binding<String>? {
        guard let value = value(key) else { return nil }
        return lookup.resolve(value: value, defaultValue: defaultValue, context: context, db: db)
    }
    
    func binding(for key: String, defaultValue: String? = nil) -> Binding<String?>? {
        guard let value = value(key) else { return nil }
        return lookup.resolve(value: value, defaultValue: defaultValue, context: context, db: db)
    }
    
    func actions(key: String) -> [CVUAction]? {
        guard let value = value(key) else { return nil }
        switch value {
        case .constant(.argument(let actionName)):
            guard let type = cvuAction(named: actionName) else { return nil }
            return [type.init(vars: [:])]
        case .array(let array):
            var actions = [CVUAction]()
            for index in stride(from: 0, to: array.count, by: 1) {
                guard case let .constant(.argument(actionName)) = array[safe: index] else {
                    continue
                }
                var vars = [String: CVUValue]()
                if case let .subdefinition(def) = array[safe: index+1] {
                    let keys = def.properties.keys
                    for key in keys {
                        if let value = context.viewArguments?.args[key] ?? context.viewArguments?.parentArguments?.args[key]  ?? def.properties[key] {
                            vars[key] = value
                        }
                    }
                }
                guard let type = cvuAction(named: actionName) else { continue }

                actions.append(type.init(vars: vars))
            }
            return actions
        default:
            return nil
        }
    }
    
    func action(key: String) -> CVUAction? {
        guard let value = value(key) else { return nil }
        switch value {
        case .constant(.argument(let actionName)):
            guard let type = cvuAction(named: actionName) else { return nil }
            return type.init(vars: [:])
        case .array(let array):
            guard case let .constant(.argument(actionName)) = array[safe: 0],
                  case let .subdefinition(def) = array[safe: 1] else {
                return nil
            }
            guard let type = cvuAction(named: actionName) else { return nil }
            return type.init(vars: def.properties)
        default:
            return nil
        }
    }
    
    func fileUID(key: String) -> String? {
        if let file = item(key), file.type == "File",
           let filename = property(of: file, propertyName: "filename")?.asString() {
            return filename
        } else if let file = edge(key: key, edgeName: "file"),
                  let filename = property(of: file, propertyName: "filename")?.asString() {
            return filename
        }
        return nil
    }
    
    func fileURL(key: String) -> URL? {
        fileUID(key: key).flatMap { FileStorageController.getURLForFile(withUUID: $0) }
    }
    
    func sizingMode(_ key: String = "sizingMode") -> CVU_SizingMode {
        guard let value = value(key),
              let string: String = lookup.resolve(value: value, context: context, db: db)
        else { return .fit }
        return CVU_SizingMode(rawValue: string) ?? .fit
    }
    
    func color(key: String = "color") -> CVUColor? {
        guard let value = value(key),
              let string: String = lookup.resolve(value: value, context: context, db: db)
        else { return nil }
        if CVUColor.hasNamed(string) {
            return CVUColor.named(string)
        }
        else {
            return CVUColor.hex(string)
        }
    }
    
    func alignment(for propertyName: String = "alignment") -> Alignment {
        guard let value = value(propertyName) else { return Alignment.center }
        switch lookup.resolve(value: value, context: context, db: db) as String? {
        case "left", "leading": return Alignment.leading
        case "top": return Alignment.top
        case "right", "trailing": return Alignment.trailing
        case "bottom": return Alignment.bottom
        case "center", "centre", "middle": return Alignment.center
        case "lefttop", "topleft": return Alignment.topLeading
        case "righttop", "topright": return Alignment.topTrailing
        case "leftbottom", "bottomleft": return Alignment.bottomLeading
        case "rightbottom", "bottomright": return Alignment.bottomTrailing
        default: return Alignment.center
        }
    }

    func textAlignment(for propertyName: String = "textAlign") -> TextAlignment {
        guard let value = value(propertyName) else { return .leading }
        switch lookup.resolve(value: value, context: context, db: db) as String? {
        case "left", "leading": return .leading
        case "right", "trailing": return .trailing
        case "center", "middle": return .center
        default: return .leading
        }
    }
    
    func cgPoint(for propertyName: String) -> CGPoint? {
        let values = valueArray(propertyName)
        if let x = values[safe: 0].flatMap({ lookup.resolve(value: $0, context: context, db: db) as Double?}),
           let y = values[safe: 1].flatMap({ lookup.resolve(value: $0, context: context, db: db) as Double?})
        {
            return CGPoint(x: x, y: y)
        }
        else if let value = cgFloat(propertyName) {
            return CGPoint(x: value, y: value)
        }
        else {
            return nil
        }
    }
    
    var edgeInsets: UIEdgeInsets? {
        insets(for: "edgeInset")
    }
    
    var nsEdgeInset: NSDirectionalEdgeInsets? {
        guard let edgeInsets = self.edgeInsets else { return nil }
        return NSDirectionalEdgeInsets(
            top: edgeInsets.top,
            leading: edgeInsets.left,
            bottom: edgeInsets.bottom,
            trailing: edgeInsets.right
        )
    }

    func insets(for propertyName: String) -> UIEdgeInsets? {
        let values = valueArray(propertyName)
        let insetArray = values.compactMap({ lookup.resolve(value: $0, context: context, db: db) as Double? }).map({ CGFloat($0) })
        switch insetArray.count {
        case 2: return UIEdgeInsets(
            top: insetArray[1],
            left: insetArray[0],
            bottom: insetArray[1],
            right: insetArray[0]
        )
        case 4: return UIEdgeInsets(
            top: insetArray[0],
            left: insetArray[3],
            bottom: insetArray[2],
            right: insetArray[1]
        )
        case 1:
            let edgeInset = insetArray[0]
            return UIEdgeInsets(
                top: edgeInset,
                left: edgeInset,
                bottom: edgeInset,
                right: edgeInset
            )
        default: return nil
        }
    }
    
    
    func font(
        for propertyName: String = "font",
        baseFont defaultValue: CVUFont = CVUFont()
    ) -> CVUFont {
        let values = valueArray(propertyName)
        if let name = values[safe: 0].flatMap({ lookup.resolve(value: $0, context: context, db: db) as String? }),
           let size = values[safe: 1].flatMap({ lookup.resolve(value: $0, context: context, db: db) as Double? }) {
                return CVUFont(
                    name: name,
                    size: CGFloat(size),
                    weight: values[safe: 2].flatMap { (lookup.resolve(value: $0, context: context, db: db) as String? ) }.flatMap(Font.Weight.init) ?? defaultValue.weight
                )
            }
            else if let value = values[safe: 0] {
                if let size: Double = lookup.resolve(value: value, context: context, db: db) {
                    return CVUFont(
                        name: defaultValue.name,
                        size: CGFloat(size),
                        weight: values[safe: 1].flatMap { (lookup.resolve(value: $0, context: context, db: db) as String?) }.flatMap(Font.Weight.init) ?? defaultValue.weight
                    )
                } else if let weight = lookup.resolve(value: value, context: context, db: db).flatMap(Font.Weight.init) {
                    return CVUFont(name: defaultValue.name, size: defaultValue.size, weight: weight)
                }
            }
        return defaultValue
    }
    
    var showNode: Bool {
        bool("show", defaultValue: false, defaultValueForMissingKey: true)
    }

    var opacity: Double {
        double("opacity") ?? 1
    }
    
    var backgroundColor: CVUColor? {
        color(key: "background")
    }

    var borderColor: CVUColor? {
        color(key: "border")
    }
    
    var minWidth: CGFloat? {
        cgFloat("width") ?? cgFloat("minWidth")
    }

    var minHeight: CGFloat? {
        cgFloat("height") ?? cgFloat("minHeight")
    }

    var maxWidth: CGFloat? {
        cgFloat("width") ?? cgFloat("maxWidth")
    }

    var maxHeight: CGFloat? {
        cgFloat("height") ?? cgFloat("maxHeight")
    }

    var offset: CGSize {
        guard let value = cgPoint(for: "offset") else { return .zero }
        return CGSize(width: value.x, height: value.y)
    }

    var shadow: CGFloat? {
        guard let value = cgFloat("shadow"), value > 0 else { return nil }
        return value
    }
    
    var zIndex: Double? {
        double("zIndex")
    }

    var lineLimit: Int? {
        int("lineLimit")
    }

    var forceAspect: Bool {
        bool("forceAspect", defaultValue: false)
    }
    
    var padding: EdgeInsets {
        guard let uiInsets = insets(for: "padding") else { return .init() }
        return EdgeInsets(
            top: uiInsets.top,
            leading: uiInsets.left,
            bottom: uiInsets.bottom,
            trailing: uiInsets.right
        )
    }

    var margin: EdgeInsets {
        guard let uiInsets = insets(for: "margin") else { return .init() }
        return EdgeInsets(
            top: uiInsets.top,
            leading: uiInsets.left,
            bottom: uiInsets.bottom,
            trailing: uiInsets.right
        )
    }
    
    func cornerRadius() -> CGFloat {
        cgFloat("cornerRadius") ?? 0
    }
    
    var spacing: CGPoint? {
        cgPoint(for: "spacing")
    }
    
}
