//
// Authentication.swift
// Copyright © 2020 memri. All rights reserved.

import CryptoKit
import Foundation
import LocalAuthentication
import KeychainAccess

class Authentication {
    #if targetEnvironment(simulator) || (targetEnvironment(macCatalyst) && DEBUG)
        private static let autologin = true
    #endif

    private static var isOwnerAuthenticated: Bool = false

    /// To check that device has secure enclave or not
    public static var hasSecureEnclave: Bool {
        !isSimulator && hasBiometrics
    }

    /// To Check that this is this simulator
    public static var isSimulator: Bool {
        TARGET_OS_SIMULATOR == 1
    }

    /// Check that this device has Biometrics features available
    private static var hasBiometrics: Bool {
        // Local Authentication Context
        let localAuthContext = LAContext()
        var error: NSError?

        /// Policies can have certain requirements which, when not satisfied, would always cause
        /// the policy evaluation to fail - e.g. a passcode set, a fingerprint
        /// enrolled with Touch ID or a face set up with Face ID. This method allows easy checking
        /// for such conditions.
        let isValidPolicy = localAuthContext.canEvaluatePolicy(
            .deviceOwnerAuthenticationWithBiometrics, error: &error
        )

        guard isValidPolicy == true else {
            if #available(iOS 11, *) {
                return error!.code != LAError.biometryNotAvailable.rawValue
            }
            else {
                return error!.code != LAError.touchIDNotAvailable.rawValue
            }
        }

        return isValidPolicy
    }

    private static let RootKeyTag = "memriPrivateKey"

    static func getErrorDescription(errorCode: Int) -> String {
        switch errorCode {
        case LAError.authenticationFailed.rawValue:
            return "Authentication was not successful, because user failed to provide valid credentials."
        case LAError.appCancel.rawValue:
            return "Authentication was canceled by application (e.g. invalidate was called while authentication was in progress)."
        case LAError.invalidContext.rawValue:
            return "LAContext passed to this call has been previously invalidated."
        case LAError.notInteractive.rawValue:
            return "Authentication failed, because it would require showing UI which has been forbidden by using interactionNotAllowed property."
        case LAError.passcodeNotSet.rawValue:
            return "Authentication could not start, because passcode is not set on the device."
        case LAError.systemCancel.rawValue:
            return "Authentication was canceled by system (e.g. another application went to foreground)."
        case LAError.userCancel.rawValue:
            return "Authentication was canceled by user (e.g. tapped Cancel button)."
        case LAError.userFallback.rawValue:
            return "Authentication was canceled, because the user tapped the fallback button (Enter Password)."
        default:
            return "Error code \(errorCode) not found"
        }
    }

    static func authenticateOwnerByPasscode(_ callback: @escaping (Error?) -> Void) {
        #if targetEnvironment(simulator) || (targetEnvironment(macCatalyst) && DEBUG)
            if DatabaseController.isRunningXcodeTests || autologin {
                isOwnerAuthenticated = true
                callback(nil)
                return
            }

            authenticateOwner(callback)
            return
        #else

            let query: NSDictionary = [
                kSecClass: kSecClassGenericPassword,
                kSecAttrService: "PasscodeAuthentication",
                kSecUseOperationPrompt: "Sign in",
            ]

            var typeRef: CFTypeRef?

            let status: OSStatus = SecItemCopyMatching(query,
                                                       &typeRef) // This will prompt the passcode.

            switch status {
            case errSecSuccess:
                callback(nil)
            default:
                callback(StringError(description:"Authentication failed"))
            }
        #endif
    }

    static func authenticateOwner(_ callback: @escaping (Error?) -> Void) {
        #if targetEnvironment(simulator) || (targetEnvironment(macCatalyst) && DEBUG)
            if DatabaseController.isRunningXcodeTests || autologin {
                isOwnerAuthenticated = true
                DispatchQueue.global(qos: .background).async {
                    callback(nil)
                }
                return
            }
        #endif

        let localAuthenticationContext = LAContext()
        localAuthenticationContext.localizedFallbackTitle = "Please use your Passcode"

        var authorizationError: NSError?
        let reason = "Authentication is required for you to continue"
        if localAuthenticationContext.canEvaluatePolicy(
            LAPolicy.deviceOwnerAuthenticationWithBiometrics,
            error: &authorizationError
        ) {
            localAuthenticationContext.evaluatePolicy(
                LAPolicy.deviceOwnerAuthenticationWithBiometrics,
                localizedReason: reason
            ) { success, evaluationError in

                if success {
                    isOwnerAuthenticated = true

                    callback(nil)
                }
                else {
                    callback(evaluationError)
                }
            }
        }
        else {
            localAuthenticationContext.evaluatePolicy(
                LAPolicy.deviceOwnerAuthentication,
                localizedReason: reason
            ) { success, evaluationError in

                if success {
                    isOwnerAuthenticated = true

                    callback(nil)
                }
                else {
                    callback(evaluationError)
                }
            }
        }
    }

    static func createOwnerAndDBKey() throws {
        let dbKey = "\(UUID().uuidString)\(UUID().uuidString)".replace("-", "")

        let privateKey = Curve25519.KeyAgreement.PrivateKey()
        let publicKey = privateKey.publicKey

        try setOwnerAndDBKey(
            privateKey: privateKey.rawRepresentation.hexEncodedString(options: .upperCase),
            publicKey: publicKey.rawRepresentation.hexEncodedString(options: .upperCase),
            dbKey: dbKey
        )
    }

    static func setOwnerAndDBKey(privateKey: String, publicKey: String, dbKey: String) throws {
        try ItemRecord.setOwnerAndDBKey(privateKey: privateKey, publicKey: publicKey, dbKey: dbKey)
    }

    static func getOwnerAndDBKey(_ callback: @escaping (Error?, AuthKeys?) -> Void) {
        do {
            let keys = try ItemRecord.getOwnerAndDBKey()
            callback(nil, keys)
        } catch {
            callback(error, nil)
        }
    }
    
    // MARK: Secure Enclave
    
    public static var lastRootPublicKey: String?

    static func createRootKey(areYouSure: Bool) throws -> String {
        guard areYouSure == true else {
            throw StringError(description: "This is a destructive operation and user must agree")
        }

        lastRootPublicKey = nil

        let access = SecAccessControlCreateWithFlags(
            kCFAllocatorDefault,
            kSecAttrAccessibleAfterFirstUnlockThisDeviceOnly,
            .privateKeyUsage,
            nil
        )

        if let access = access {
            var attributes: [String: Any]

            let appTag = "\(RootKeyTag)".data(using: .utf8)!

            if hasSecureEnclave {
                attributes = [
                    kSecAttrKeyType as String: kSecAttrKeyTypeEC,
                    kSecAttrKeySizeInBits as String: 256,
                    kSecAttrTokenID as String: kSecAttrTokenIDSecureEnclave,
                    kSecPrivateKeyAttrs as String: [
                        kSecAttrIsPermanent as String: true,
                        kSecAttrApplicationTag as String: appTag,
                        kSecAttrAccessControl as String: access,
                    ],
                ]
            }
            else {
                attributes = [
                    kSecAttrKeyType as String: kSecAttrKeyTypeEC,
                    kSecAttrKeySizeInBits as String: 256,
                    kSecAttrApplicationTag as String: appTag,
                    kSecAttrIsPermanent as String: true,
                ]
            }

            var publicKey: SecKey?
            var privateKey: SecKey?
            let err = SecKeyGeneratePair(attributes as NSDictionary, &publicKey, &privateKey)
            if err == errSecSuccess {
                var error: Unmanaged<CFError>?
                guard
                    let publicKey = publicKey,
                    let data = SecKeyCopyExternalRepresentation(publicKey, &error) as Data?
                else {
                    throw error!.takeRetainedValue() as Error
                }

                return data.hexEncodedString(options: .upperCase)
            }
            else {
                throw NSError(domain: NSOSStatusErrorDomain, code: Int(err), userInfo: nil)
            }
        }
        else {
            throw StringError(description: "Unspecified error")
        }
    }

    static func getPublicRootKey(_ callback: @escaping (Error?, String?) -> Void) {
        do {
            guard isOwnerAuthenticated else {
                throw StringError(description: "Not yet authenticated")
            }

            if lastRootPublicKey == nil {
                lastRootPublicKey = try getPublicRootKeySync()
            }

            callback(nil, lastRootPublicKey)
        }
        catch {
            authenticateOwner { error in
                if let error = error {
                    // TODO: error handling
                    callback(error, nil)
                    return
                }

                do {
                    callback(nil, try getPublicRootKeySync())
                }
                catch {
                    callback(error, nil)
                }
            }
        }
    }

    static func getPublicRootKeySync() throws -> String {
        guard !isSimulator else {
            let key = "SIMULATOR"
            lastRootPublicKey = key
            return key
        }
        
        guard isOwnerAuthenticated else {
            throw StringError(description: "Not yet authenticated")
        }

        if let lastRootPublicKey = lastRootPublicKey {
            return lastRootPublicKey
        }

        let query: [String: Any] = [
            kSecClass as String: kSecClassKey,
            kSecAttrApplicationTag as String: RootKeyTag,
            kSecAttrKeyType as String: kSecAttrKeyTypeEC,
            kSecReturnRef as String: true,
            kSecMatchLimit as String: 10000,
        ]

        var item: CFTypeRef?
        let status = SecItemCopyMatching(query as CFDictionary, &item)
        switch status {
        case noErr: break
        case errSecItemNotFound: throw StringError(description: "Keychain item not found")
        default: throw StringError(description: "Unable to fetch public key")
        }

        if
            let privateKey = (item as? [SecKey])?.last,
            let publicKey = SecKeyCopyPublicKey(privateKey)
        {
            var error: Unmanaged<CFError>?
            guard let data = SecKeyCopyExternalRepresentation(publicKey, &error) as Data? else {
                throw error!.takeRetainedValue() as Error
            }

            let key = data.subdata(in: Range(0 ... 63)).hexEncodedString(options: .upperCase)
            lastRootPublicKey = key
            return key
        }
        else {
            throw StringError(description: "Unable to fetch public key")
        }
    }
}
