//
//  PodAPIPayloads.swift
//  MemriDatabase
//
//  Created by T Brennan on 19/12/20.
//

import Foundation

/// A type representing the most common `post` body for pod API requests.
/// This typically includes a database key, and a payload (different for each API endpoints)
struct PodAPIRequestBody<Payload: Encodable>: Encodable {
    init(connectionConfig: PodAPIConnectionDetails, payload: Payload) {
        self.auth = ["type": "ClientAuth", "databaseKey": connectionConfig.keys.dbKey]
        #if DEBUG
            print("AUTH: ", connectionConfig.keys.dbKey)
        #endif
        self.payload = payload
    }
    
    var auth: [String: String]
    var payload: Payload
}

/// A namespace for types representing Pod API payloads. Refer to pod API documentation for expected payloads for each API.
enum PodAPIPayload {
    struct EmptyPayload: Encodable {
        func encode(to encoder: Encoder) throws {
            var container = encoder.singleValueContainer()
            try container.encodeNil()
        }
    }
    
    struct FileSHA:  Encodable {
        var sha256: String
    }
    struct itemId:  Encodable {
        var uid: StringUUID
    }

    struct ItemUIDList: Encodable {
        var uids: Set<StringUUID>
        
        func encode(to encoder: Encoder) throws {
           var container = encoder.singleValueContainer()
            try container.encode(uids)
        }
    }
    
    struct ItemUIDWithServicePayload:  Encodable {
        var uid: StringUUID
        var servicePayload: ServicePayload
    }
    
    /// Some of the pod APIs require a `servicePayload` included in the main payload
    struct ServicePayload: Encodable {
        var databaseKey: String
        var ownerKey: String
        
        init(connectionConfig: PodAPIConnectionDetails) {
            self.databaseKey = connectionConfig.keys.dbKey
            self.ownerKey = connectionConfig.keys.ownerKey
        }
    }
    
    struct BulkAction: Encodable {
        var createItems: [[String: AnyEncodable]]
        var updateItems: [[String: AnyEncodable]]
        var deleteItems: [StringUUID]
        var createEdges: [[String: AnyEncodable]]
    }
}
