//
//  AddressBookController.swift
//  Memri
//

import ContactsUI
import Foundation
import Combine
import GRDB
import PhoneNumberKit

struct AddressBookController {
    
    static var appStateObservation: AnyCancellable?
    static let phoneNumberKit = PhoneNumberKit()
    static var syncing = false

    static func start() {
        appStateObservation = AppController.shared.permissionController.contactAuthorization
            .filter { $0 }
            .sink { state in
            if AppController.shared.state == .authenticated && state {
                sync {
                    DispatchQueue.main.async {
                        syncing = false
                    }
                    print("Addressbook Sync Complete")
                }
            }
        }
    }
    
    static func sync(_ completion: (() -> Void)? = nil) {
        guard self.syncing == false else {
            completion?()
            return
        }
        
        syncing = true
        DispatchQueue.global(qos: .background).async {
            AppController.shared.databaseController.createDatabasePools { (pools) in
                guard let poolToUse = pools?.0 else {
                    completion?()
                    return
                }
                
                let contacts = fetchContacts()
                var index = 0
                for contact in contacts {
                    index = index+1
                    upsertContact(contact, pool: poolToUse, index: index)
                }
                
                completion?()
            }
        }
    }
    
    static func personWithAddressBookID(_ addressBookId: String, db: Database) -> ItemRecord? {
        guard let addressBookReference = try? StringRecord.filter(StringRecord.Columns.name == "addressBookId" &&
                                                                    StringRecord.Columns.value == addressBookId).fetchOne(db),
              let person = try? ItemRecord.filter(ItemRecord.Columns.rowId == addressBookReference.item).fetchAll(db).first else {
            return nil
        }
        
        return person
    }
    
    static func phoneNumberItemForPhoneNumber(_ phoneNumber: String, db: Database) -> ItemRecord? {
        guard let phoneNumberItems = try? ItemRecord.filter(ItemRecord.Columns.type == "PhoneNumber").fetchAll(db) else {
            return nil
        }
        
        for phoneNumberItem in phoneNumberItems {
            if let phoneNumberProperties = phoneNumberItem.propertyValues("phoneNumber", db: db) {
                for phoneNumberProperty in phoneNumberProperties {
                    if case let .string(phoneNumberValue) = phoneNumberProperty,
                       phoneNumberValue == phoneNumber {
                        return phoneNumberItem
                    }
                }
            }
        }
        
        return nil
    }
    
    static func upsertContact(_ contact: CNContact, pool: DatabasePool, index: Int) {
        try? AppController.shared.databaseController.write(usingPool: pool) { (db) in
            var existingContact: ItemRecord? = personWithAddressBookID(contact.identifier, db: db)
            
            if existingContact == nil {
                let person = ItemRecord(type: "Person")
                try? person.save(db: db)
                try? person.setPropertyValue(name: "addressBookId", value: .string(contact.identifier), db: db)
                existingContact = person
            }
            
            guard let person = existingContact else {
                return
            }
            
            let phoneNumbers = contact.formattedPhoneNumbers
            for phoneNumber in phoneNumbers.prefix(10) {
                savePhoneNumber(phoneNumber, forPerson: person, db: db)
            }

            if !contact.givenName.isEmpty {
                try? person.setPropertyValue(name: "firstName", value: .string(contact.givenName), db: db)
            }
            
            if !contact.familyName.isEmpty {
                try? person.setPropertyValue(name: "lastName", value: .string(contact.familyName), db: db)
            }
            
            if let birthDateComponents = contact.birthday,
               let birthDate = Calendar.current.date(from: birthDateComponents) {
                try? person.setPropertyValue(name: "birthDate", value: .datetime(birthDate), db: db)
            }
            
            if let email = contact.emailAddresses.first?.value as String? {
                try? person.setPropertyValue(name: "email", value: .string(email), db: db)
            }

            if contact.imageDataAvailable ,
               let avatar = contact.imageData {
                saveAvatar(avatar, forPerson: person, db: db)
            }
        }
    }

    static func saveAvatar(_ data: Data, forPerson person: ItemRecord, db: Database) {
        if let personRowId = person.rowId {

            do {
                let newImageItem  = ItemRecord(type: "Photo")
                try newImageItem.save(db: db)

                guard let newImageId = newImageItem.rowId else {
                    return
                }
                _ = try EdgeRecord.createFor(source: personRowId, type: "profilePicture", target: newImageId, db: db)

                let newFileItem = ItemRecord(type: "File")
                try newFileItem.save(db: db)

                guard let newFileItemId = newFileItem.rowId else {
                    return
                }

                let fileName = "\(UUID().uuidString).jpg"

                try FileStorageController.writeData(data, toFileForUUID: fileName)
                let sha256 = FileStorageController.getHashForFile(withUUID: fileName)
                try newFileItem.setPropertyValue(name: "sha256", value: .string(sha256), db: db)
                try newFileItem.setPropertyValue(name: "filename", value: .string(fileName), db: db)
                _ = try EdgeRecord.createFor(source: newImageId, type: "file", target: newFileItemId, db: db)
            } catch {
                print("Error saving avatar : \(error.localizedDescription)")
            }
        }
    }
    
    static func savePhoneNumber(_ formattedPhoneNumber: String, forPerson person: ItemRecord, db: Database) {
        if let personRowId = person.rowId {
            var existingItem = phoneNumberItemForPhoneNumber(formattedPhoneNumber, db: db)
            
            // Create new
            if (existingItem == nil) {
                existingItem = ItemRecord(type: "PhoneNumber")
                try? existingItem?.save(db: db)
                try? existingItem?.setPropertyValue(name: "phoneNumber", value: .string(formattedPhoneNumber), db: db)
            }

            guard let existingItemRowId = existingItem?.rowId else {
                return
            }
            
            let existingEdge = try? EdgeRecord.filter(EdgeRecord.Columns.source == personRowId
                                                    && EdgeRecord.Columns.target == existingItemRowId
                                                    && EdgeRecord.Columns.name == "hasPhoneNumber").fetchOne(db)
            if (existingEdge == nil) {
                let _ = try? EdgeRecord.createFor(source: personRowId, type: "hasPhoneNumber", target: existingItemRowId, db: db)
            }
        }
    }
    
    static func fetchContacts() -> [CNContact] {
        let contactStore = CNContactStore()
        let keysToFetch = [
            CNContactFormatter.descriptorForRequiredKeys(for: .fullName),
            CNContactGivenNameKey,
            CNContactFamilyNameKey,
            CNContactBirthdayKey,
            CNContactPhoneNumbersKey,
            CNContactEmailAddressesKey,
            CNContactThumbnailImageDataKey,
            CNContactImageDataAvailableKey,
            CNContactImageDataKey,
        ] as [Any]

        var allContainers: [CNContainer] = []
        do {
            allContainers = try contactStore.containers(matching: nil)
        } catch {
            print("Error fetching containers")
        }

        var results: [CNContact] = []

        for container in allContainers {
            let fetchPredicate = CNContact.predicateForContactsInContainer(withIdentifier: container.identifier)

            do {
                let containerResults = try contactStore.unifiedContacts(matching: fetchPredicate, keysToFetch: keysToFetch as! [CNKeyDescriptor])
                results.append(contentsOf: containerResults)
            } catch {
                print("Error fetching containers")
            }
        }

        return results
    }
}

extension CNContact {
    static let phoneNumberKit = PhoneNumberKit()

    var databaseID: String {
        return "AddressBook-\(self.identifier)"
    }
    
    var formattedPhoneNumbers: [String] {
        return self.phoneNumbers.compactMap{
            if let phoneNumber = try? CNContact.phoneNumberKit.parse($0.value.stringValue) {
                return CNContact.phoneNumberKit.format(phoneNumber, toType: .e164)
            }
            
            return nil
        }
    }
}
