//
//  AppController.swift
//  MemriDatabase
//
//  Created by T Brennan on 14/12/20.
//

import Foundation
import KeychainAccess
import Combine
import SwiftUI

class AppController: ObservableObject {
    static let shared = AppController()

    var podURL: String {
        return (try? Settings.shared.getSetting("user/pod/host", type: String.self)) ?? "http://0.0.0.0:3030"
    }

    let databaseController: DatabaseController
    let syncController: SyncController
    let cvuController: CVUController
    let pubsubController: PubSubController
    let permissionController: PermissionsController

    private init() {
        let databaseController = DatabaseController()
        self.databaseController = databaseController
        self.syncController = SyncController(databaseController: databaseController)
        self.pubsubController = PubSubController()
        self.cvuController = CVUController()
        self.permissionController = PermissionsController()
    }
    
    enum AppState {
        case setup
        case authentication
        case authenticated
    }
    
    @Published
    var state: AppState = .setup
    let statePublisher = PassthroughSubject<AppState, Never>()

    func updateState() {
        guard databaseController.databasePool != nil, databaseController.isDatabaseSetup else {
            state = .setup
            statePublisher.send(state)
            try? syncController.sync()
            return
        }
        
        state = .authenticated
        statePublisher.send(state)
        try? syncController.sync()
    }
    
    func onLaunch() {
        databaseController.setupDatabase {[weak self] in
            self?.updateState()
            AddressBookController.start()
        }
    }
    
    // MARK: Setup
    func setupApp(config: SetupConfig, useDemoData: Bool = true, onCompletion: @escaping (_ error: Error?) -> Void) {
        switch config {
        case .local:
            self.setupDatabaseKey() {[weak self] (error) in
                guard error == nil else {
                    onCompletion(error)
                    return
                }
                
                self?.databaseController.setupDatabase {
                    try? self?.databaseController.importRequiredData()
                    if (useDemoData) {
                        try? self?.databaseController.setupWithDemoData()
                    }
                }
            }
        case .newPod:
            self.setupDatabaseKey() {[weak self] (error) in
                guard error == nil else {
                    onCompletion(error)
                    return
                }
                
                self?.databaseController.setupDatabase {
                    try? Authentication.createOwnerAndDBKey()
                    try? self?.databaseController.importRequiredData()
                    if (useDemoData) {
                        try? self?.databaseController.setupWithDemoData()
                    }
                }
            }
        case .existingPod:
            // TODO:
            break
        }

        onCompletion(nil)
    }
   
    private var podScheme: String {
        guard let url = URL(string: podURL),
              let scheme = url.scheme else {
            return "http"
        }
        
        return scheme
    }
    
    private var podHost: String {
        guard let url = URL(string: podURL),
              let host = url.host else {
            return "0.0.0.0"
        }
        
        return host
    }
    
    private var podPort: Int {
        guard let url = URL(string: podURL),
              let port = url.port else {
            return 3030
        }
        
        return port
    }
    
    // MARK: Pod connection
    var podConnectionConfig: PodAPIConnectionDetails? {
        do {
            return try databaseController.read { db in
                // Here you should retrieve the connection details stored in the database
                let keys = try ItemRecord.getOwnerAndDBKey()
                return PodAPIConnectionDetails(scheme: podScheme, host: podHost, port: podPort, keys: keys)
            }
        } catch {
            print(error)
            return nil
        }
    }
    
    static let keychain_databaseKey: String = "memri_databaseKey"
    
    func getDatabaseKey(_ callback: @escaping (Error?, String?) -> Void) {
        if let cachedKey = Authentication.lastRootPublicKey {
            callback(nil, cachedKey)
            return
        }
        
        self.state = .authentication
        Authentication.authenticateOwner {[weak self] (error) in
            DispatchQueue.main.async {
                guard error == nil else {
                    self?.state = .setup
                    callback(error, nil)
                    return
                }
                
                
                let key = try? Authentication.getPublicRootKeySync()
                callback(nil, key)
                self?.updateState()
            }
        }
    }
    
    func setupDatabaseKey(callback: @escaping (Error?) -> Void) {
        self.state = .authentication
        Authentication.authenticateOwner {[weak self] (error) in
            DispatchQueue.main.async {
                guard error == nil else {
                    callback(error)
                    self?.state = .setup
                    return
                }
                
                // Use existing key if already present
                let existingKey = try? Authentication.getPublicRootKeySync()
                guard existingKey == nil else {
                    callback(nil)
                    self?.updateState()
                    return
                }

                // Try creating a new one
                guard let _ = try? Authentication.createRootKey(areYouSure: true) else {
                    callback(StringError(description: "Error generating keypair"))
                    self?.state = .setup
                    return
                }

                callback(nil)
                self?.updateState()
            }
        }
    }
}


enum SetupConfig {
    case local
    case newPod(NewPodConfig)
    case existingPod(ExistingPodConfig)
    
    struct NewPodConfig {
        var podURL: String
    }
    
    struct ExistingPodConfig {
        var podURL: String
        var podPrivateKey: String
        var podPublicKey: String
        var podDatabaseKey: String
    }
}
