//
//  GRDBTest.swift
//  MemriDatabase
//
//  Created by T Brennan on 10/11/20.
//

import Foundation
import GRDB

/// The database controller provides access to the app's SQLite database. Generally only a single database controller will be used throughout the app
class DatabaseController {
    
    #if DEBUG
        static var isRunningXcodeTests = false
    #endif
    
    var url: URL
    
    /// Create a DatabaseController. Change the databaseName to create/access a different database file (eg. for testing purposes)
    init(databaseName: String = "memri") {
        let paths = FileManager.default.urls(for: .documentDirectory, in: .userDomainMask)
        let documentsDirectory = paths[0]
        self.url = documentsDirectory.appendingPathComponent("\(databaseName).sqlite")
        print(self.url)
    }
    
    func setupDatabase(_ completion: @escaping () -> Void) {
        guard databasePool == nil else {
            completion()
            return
        }
        
        createDatabasePools {[weak self] (pools) in
            self?.databasePool = pools?.0
            self?.readPool = pools?.1

            if !Constants.isRunningTests {
                try? self?.runMigrations()
            }
            
            if let pool = self?.databasePool {
                do {
                    try DemoData.importSchemaOnce(pool: pool)
                } catch let error {
                    print("ERROR: ", error)
                }
            }

            completion()
        }
    }
    
    func createDatabasePools(_ completion: @escaping ((DatabasePool?, DatabasePool?)?) -> Void) {
        AppController.shared.getDatabaseKey {[weak self] (error, key) in
            guard let passphrase = key,
                  let weakSelf = self else {
                completion(nil)
                return
            }
            
            let semaphore = DispatchSemaphore(value: 0)
            var config = Configuration()
            DispatchQueue.global(qos: .background).async {
                config.prepareDatabase {db in
                    try db.usePassphrase(passphrase)
                }
                
                semaphore.signal()
            }
            
            semaphore.wait()
            let pool = try? DatabasePool(path: weakSelf.url.path, configuration: config)
            let readPool = try? DatabasePool(path: weakSelf.url.path, configuration: config)
            completion((pool, readPool))
        }
    }
    
    var schema = Schema()
    
    /// This variable is `thread-local` and used to track whether there is a database `read` already in progress
    private var readSession: ThreadLocal<Database?> = .init(value: nil)
    
    /// This variable is `thread-local` and used to track whether there is a database `write` already in progress
    private var writeSession: ThreadLocal<Database?> = .init(value: nil)
    
    /// This is the connection to the database used throughout the app
    /// It uses SQLite's WAL mode to provide concurrent reads and allow for read/write from background threads
    var databasePool: DatabasePool?
    var readPool: DatabasePool?
    
    /// Prepare the database and register migration steps.
    func runMigrations() throws {
        guard let pool = databasePool else {
            throw StringError(description: "Cannot setup database, no pool available")
        }
        var migrator = DatabaseMigrator()
        
        if Constants.isRunningTests {
            /// *NEVER* set this outside of TESTING config (because it deletes all data)
            try pool.erase()
        }
        
        #if DEBUG
        /// Aid development by wiping the database EVERY LAUNCH
        /// *NEVER* set this outside of DEBUG config (because it deletes all data)
//        try databasePool.erase()
        
        /// Aid database development by wiping the database when migrations change
        /// *NEVER* set this outside of DEBUG config (because it deletes all data)
        migrator.eraseDatabaseOnSchemaChange = true
        #endif
        
        migrator.registerMigration("initialSetup") { db in
            
            // Item
            try db.create(table: "items") { t in
                t.column("rowId", .integer).primaryKey()
                t.column("id", .text).notNull()
                t.column("type", .text).notNull()
                t.column("dateCreated", .integer).notNull()
                t.column("dateModified", .integer).notNull()
                t.column("dateServerModified", .integer).notNull()
                t.column("deleted", .boolean).notNull(onConflict: .replace)
                
                /// iOS only (used for syncing decisions
                t.column("syncState", .text).defaults(to: SyncState.create).notNull(onConflict: .replace)
                t.column("fileState", .text).defaults(to: FileState.skip).notNull(onConflict: .replace)
                t.column("syncHasPriority", .boolean).defaults(to: false).notNull(onConflict: .replace)
            }
            try db.create(index: "idx_items_id", on: "items", columns: ["id"], unique: true)
            try db.create(index: "idx_items_type_dateServerModified", on: "items", columns: ["type", "dateServerModified"])
            try db.create(index: "idx_items_dateServerModified", on: "items", columns: ["dateServerModified"])

            
            // Edges
            try db.create(table: "edges") { t in
                t.column("item", .integer).primaryKey().references("items", onDelete: .cascade)
                t.column("source", .integer).notNull().references("items", onDelete: .cascade)
                t.column("name", .text).notNull()
                t.column("target", .integer).notNull().references("items")
                
                /// iOS only (used for syncing decisions
                t.column("syncState", .text).defaults(to: SyncState.create).notNull(onConflict: .replace)
                t.column("syncHasPriority", .boolean).defaults(to: false).notNull(onConflict: .replace)
            }
            try db.create(index: "idx_edges_source_name", on: "edges", columns: ["source", "name"])
            try db.create(index: "idx_edges_target_name", on: "edges", columns: ["target", "name"])

            
            // Integers
            try db.create(table: "integers") { t in
                t.column("item", .integer).notNull().references("items", onDelete: .cascade)
                t.column("name", .text).notNull()
                t.column("value", .integer).notNull()
            }
            try db.create(index: "idx_integers_item_name", on: "integers", columns: ["item", "name"], unique: true)
            try db.create(index: "idx_integers_name_value", on: "integers", columns: ["name", "value"])
            try db.create(index: "idx_integers_name_item", on: "integers", columns: ["name", "item"])

            // Strings
            try db.create(table: "strings") { t in
                t.column("item", .integer).notNull().references("items", onDelete: .cascade)
                t.column("name", .text).notNull()
                t.column("value", .text).notNull()
            }
            try db.create(index: "idx_strings_item_name", on: "integers", columns: ["item", "name"], unique: true)
            try db.create(index: "idx_strings_name_value", on: "integers", columns: ["name", "value"])
            try db.create(index: "idx_strings_name_item", on: "integers", columns: ["name", "item"])
            try db.create(virtualTable: "strings_search", using: FTS4()) { t in
                t.synchronize(withTable: "strings")
                t.tokenizer = .porter
                t.column("item").notIndexed()
                t.column("name").notIndexed()
                t.column("value")
            }

            // Reals
            try db.create(table: "reals") { t in
                t.column("item", .integer).notNull().references("items", onDelete: .cascade)
                t.column("name", .text).notNull()
                t.column("value", .double).notNull()
            }
            try db.create(index: "idx_reals_item_name", on: "integers", columns: ["item", "name"], unique: true)
            try db.create(index: "idx_reals_name_value", on: "integers", columns: ["name", "value"])
            try db.create(index: "idx_reals_name_item", on: "integers", columns: ["name", "item"])
            
            // Navigation state
            try db.create(table: "navigationState") { t in
                t.column("sessionID", .text).notNull().primaryKey()
                t.column("state", .blob).notNull()
            }
        }
                
        try migrator.migrate(pool)
    }
    
    
    func search(string searchString: String) throws -> [ItemRecord] {
        try read { (db) in
            guard let searchQuery = FTS3Pattern(matchingAllTokensIn: searchString) else {
                throw StringError(description: "Invalid search string: \(searchString)")
            }
            let refinedQuery = try FTS3Pattern(rawPattern: "\(searchQuery.rawPattern)*")
            let search = try ItemRecord.search(db, pattern: refinedQuery)
            return search
        }
    }
}

import SwiftUI
extension DatabaseController {
    /// Read from the database while constructing a SwiftUI view
    func read<Content: View>(@ViewBuilder content: (Database) throws -> Content) throws -> some View {
        guard let pool = readPool else {
            throw StringError(description: "Cannot read, no pool available")
        }

        if let existingReadDB = readSession.inner.value {
            return try? content(existingReadDB)
        } else {
            defer {
                readSession.inner.value = nil
            }
            return try? pool.read { (db) -> Content? in
                readSession.inner.value = db
                return try? content(db)
            }
        }
    }
    
    /// Read from the database and return a value
    func read<T>(_ block: (Database) throws -> T ) throws -> T {
        guard let pool = readPool else {
            throw StringError(description: "Cannot read, no pool available")
        }

        if let existingReadDB = readSession.inner.value {
            return try block(existingReadDB)
        } else {
            defer {
                readSession.inner.value = nil
            }
            return try pool.read { (db) -> T in
                readSession.inner.value = db
                return try block(db)
            }
        }
    }
    
    // Write to the database. This happens async by default
    func write(usingPool: DatabasePool? = nil, updates: @escaping (Database) throws -> Void ) throws {
        guard let pool = usingPool ?? databasePool else {
            throw StringError(description: "Cannot write, no pool available")
        }

        DispatchQueue.global(qos: .userInteractive).async {
            do {
                try pool.write { (db) in
                    try updates(db)
                }
            } catch {
                print(error)
            }
        }
    }
    
    // Write to the database immediately. Note that this may throw an error if inside a database read
    func writeSync<T>(_ updates: (Database) throws -> T ) throws -> T {
        guard let pool = databasePool else {
            throw StringError(description: "Cannot write, no pool available")
        }

        if let existingWriteDB = writeSession.inner.value {
            return try updates(existingWriteDB)
        } else {
            defer {
                writeSession.inner.value = nil
            }
            
            return try pool.write { (db) in
                writeSession.inner.value = db
                return try updates(db)
            }
        }
    }
}

extension DatabaseController {

    var isDatabaseSetup: Bool {
        return hasImportedSchema && hasImportedDefaultData
    }
    
    /// Check if the database has been setup
    var hasImportedDemoData: Bool {
        (try? read { db in
            try? ItemRecord
                .filter(Column(ItemRecord.Columns.type.rawValue) == "Note")
                .fetchOne(db) != nil
        }) ?? false
    }

    var hasImportedSchema: Bool {
        (try? read { db in
            try? ItemRecord
                .filter(Column(ItemRecord.Columns.type.rawValue) == "ItemPropertySchema")
                .fetchOne(db) != nil
        }) ?? false
    }
    
    var hasImportedDefaultData: Bool {
        (try? read { db in
            try? ItemRecord
                .filter(Column(ItemRecord.Columns.type.rawValue) == "NavigationItem")
                .fetchOne(db) != nil
        }) ?? false
    }
        
    func importRequiredData() throws {
        if !hasImportedDefaultData {
            try DemoData.importDefaultData()
        }
    }
    
    func setupWithDemoData() throws {
        guard !self.hasImportedDemoData else {
            // If there is already data set up, don't import
            return
        }
        try DemoData.importDemoData(databaseController: self)
    }
}
