//
//  DatabaseQuery.swift
//  MemriDatabase
//
//  Created by T Brennan on 7/12/20.
//

import Foundation
import GRDB
import Combine


enum ConditionOperator: String, Codable {
    case and = "AND"
    case or = "OR"
}

/// This type is used to describe a database query. It can be stored as a JSON (it is codable)
struct DatabaseQueryConfig: Equatable, Codable {
    /// A list of item types to include. Default is Empty -> ALL item types
    var itemTypes: [String] = ["Person", "Note", "Address", "Photo", "Indexer", "Importer"]
    
    /// A list of item UIDs to include. Default is Empty -> don't filter on UID
    var itemIDs: Set<Int64> = []
    
    /// A property to sort the results by
    var sortProperty: String? = "dateModified"
    
    var sortAscending: Bool = false
    
    /// Only include items modified after this date
    var dateModifiedAfter: Date?
    
    /// Only include items modified before this date
    var dateModifiedBefore: Date?
    
    /// Only include items created after this date
    var dateCreatedAfter: Date?
    
    /// Only include items created before this date
    var dateCreatedBefore: Date?
    
    var deleted: Bool = false
    
    /// The maximum number of items to fetch
    var pageSize: Int = 1000
    
    /// Use this to change which page is requested by the query (eg. if there are more than `pageSize` items)
    var currentPage: Int = 0
    
    /// A search string to match item properties against
    var searchString: String? = nil
    
    /// If enabled the search will find items that link to another item matching the search term. This only goes one edge deep for performance purposes.
    var includeImmediateEdgeSearch: Bool = true
    
    /// A list of conditions. eg. property name = "Demo note"
    var conditions: [DatabaseQueryCondition] = []
    
    var edgeTargetsOperator: ConditionOperator = .and
    
    var count: Int?
    
    private func constructFilteredRequest(withSearchUIDs searchUIDs: Set<Int64>?) -> QueryInterfaceRequest<ItemRecord> {

        let limit = count ?? pageSize
        let offset = count != nil ? 0 : pageSize * currentPage
        
        var queryConditions: [SQLExpression] = []
        /// Filter by item type
        if !itemTypes.isEmpty {
            queryConditions.append(itemTypes.contains(ItemRecord.Columns.type))
        }
        
        
        /// Filter to only include items matching the search term (AND if already filtered by UID, those that match both)
        if let searchUIDs = searchUIDs {
            if !itemIDs.isEmpty {
                let combinedUIDs = searchUIDs.intersection(itemIDs)
                queryConditions.append(combinedUIDs.contains(ItemRecord.Columns.rowId))
            } else {
                queryConditions.append(searchUIDs.contains(ItemRecord.Columns.rowId))
            }
        } else if !itemIDs.isEmpty {
            queryConditions.append(itemIDs.contains(ItemRecord.Columns.rowId))
        }
        
        /// Filter by date ranges
        if let dateModifiedBefore = dateModifiedBefore {
            queryConditions.append(ItemRecord.Columns.dateModified <= dateModifiedBefore)
        }
        if let dateModifiedAfter = dateModifiedAfter {
            queryConditions.append(ItemRecord.Columns.dateModified >= dateModifiedAfter)
        }
        if let dateCreatedBefore = dateCreatedBefore {
            queryConditions.append(ItemRecord.Columns.dateModified <= dateCreatedBefore)
        }
        if let dateCreatedAfter = dateCreatedAfter {
            queryConditions.append(ItemRecord.Columns.dateModified >= dateCreatedAfter)
        }
        queryConditions.append(ItemRecord.Columns.deleted == deleted)

        var request = ItemRecord.limit(limit, offset: offset).filter(queryConditions.joined(operator: .and)).distinct()
        
        // Property conditions
        let queryPropertyConditions: [SQLExpression] = conditions.compactMap { condition in
            switch condition {
            case let .propertyEquals(info):
                switch info.value {
                case let .string(stringValue):
                    return StringRecord.Columns.name == info.name && StringRecord.Columns.value == stringValue
                case let .bool(boolValue):
                    return IntegerRecord.Columns.name == info.name && IntegerRecord.Columns.value == (boolValue == true ? 1 : 0)
                }
            default:
                return nil
            }
        }
        if !queryPropertyConditions.isEmpty {
            let stringPropertyFilter = ItemRecord.stringProperties.filter(queryPropertyConditions.joined(operator: .and))
            let integerPropertyFilter = ItemRecord.integerProperties.filter(queryPropertyConditions.joined(operator: .and))
            let realPropertyFilter = ItemRecord.realProperties.filter(queryPropertyConditions.joined(operator: .and))

            request = request.having(stringPropertyFilter.isEmpty == false ||
                                        integerPropertyFilter.isEmpty == false ||
                                        realPropertyFilter.isEmpty == false)
        }
        
        // Edge conditions
        let queryEdgeConditions: [SQLExpression] = conditions.compactMap { condition in
            switch condition {
            case let .edgeHasTarget(info):
                return EdgeRecord.Columns.name == info.edgeName && EdgeRecord.Columns.target == info.target
            case let .edgeHasSource(info):
                return EdgeRecord.Columns.name == info.edgeName && EdgeRecord.Columns.source == info.source
            default:
                return nil
            }
        }
        if !queryEdgeConditions.isEmpty {
            if (hasEdgeSourceConditions) {
                let edgeFilter = ItemRecord.reverseEdges.filter(queryEdgeConditions.joined(operator: edgeTargetsOperator == .and ? .and : .or))
                request = request.having(edgeFilter.isEmpty == false)
            } else {
                let edgeFilter = ItemRecord.edges.filter(queryEdgeConditions.joined(operator: edgeTargetsOperator == .and ? .and : .or))
                request = request.having(edgeFilter.isEmpty == false)
            }
        }
        
        switch sortProperty?.nilIfBlank {
        case "dateCreated":
            request = request.order(sortAscending ? [ItemRecord.Columns.dateCreated, ItemRecord.Columns.dateModified] : [ItemRecord.Columns.dateCreated.desc, ItemRecord.Columns.dateModified.desc])
        case "dateModified":
            request = request.order(sortAscending ? [ItemRecord.Columns.dateModified, ItemRecord.Columns.dateCreated] : [ItemRecord.Columns.dateModified.desc, ItemRecord.Columns.dateCreated.desc])
        case "dateSent":
            let sortAlias = TableAlias()
            request = request.joining(optional:
                                        ItemRecord.realProperties.aliased(sortAlias)
                                        .filter(RealRecord.Columns.name == sortProperty))
        case .some(let sortProperty):
            let sortAlias = TableAlias()
            request = request.joining(optional:
                                        ItemRecord.stringProperties.aliased(sortAlias)
                                        .filter(StringRecord.Columns.name == sortProperty)
            )
            // todomigration:
//            request = request.joining(optional:
//                                        ItemRecord.integerProperties.aliased(sortAlias)
//                                        .filter(IntegerRecord.Columns.name == sortProperty)
//            )
//            request = request.joining(optional:
//                                        ItemRecord.realProperties.aliased(sortAlias)
//                                        .filter(RealRecord.Columns.name == sortProperty)
//            )

            .order(sortAscending ? [
                sortAlias[StringRecord.Columns.value].asc,
                ItemRecord.Columns.dateModified,
                ItemRecord.Columns.dateCreated
            ] : [
                sortAlias[StringRecord.Columns.value].desc,
                ItemRecord.Columns.dateModified.desc,
                ItemRecord.Columns.dateCreated.desc
            ])
        default: break
        }
        
        return request
    }
    
    private var hasEdgeSourceConditions: Bool {
        return (conditions.filter { condition in
            if case .edgeHasSource(_) = condition {
                return true
            }
            
            return false
        }).count > 0
    }
    
    private func constructSearchRequest() -> SQLRequest<Int64?>? {
        guard let searchString = searchString?.nilIfBlank,
              let searchQuery = FTS3Pattern(matchingAllTokensIn: searchString),
              let refinedQuery = try? FTS3Pattern(rawPattern: "\(searchQuery.rawPattern)*")
        else {
            return nil
        }
        return SQLRequest(sql:
            """
               SELECT DISTINCT \(StringRecord.Columns.item)
               FROM \(StringRecord.databaseSearchTableName)
               WHERE \(StringRecord.databaseSearchTableName) MATCH ?
            """, arguments: [refinedQuery])
    }
    
    func executeRequest(db: Database) throws -> [ItemRecord] {
        var searchUIDs = try constructSearchRequest().map { try Set($0.fetchAll(db).compactMap { $0 }) }
        if includeImmediateEdgeSearch {
            /// Find items connected the the search results by one or more edges. Eg. if a file is found based on the search term, we will also include a Photo or Note that links to it
            let edgeUIDs = try searchUIDs.map { searchUIDs -> [Int64] in
                try EdgeRecord.filter(searchUIDs.contains(EdgeRecord.Columns.target)).select([EdgeRecord.Columns.source]).fetchAll(db)
            }
            searchUIDs = searchUIDs?.union(edgeUIDs ?? [])
        }
        
        let filterRequest = constructFilteredRequest(withSearchUIDs: searchUIDs)
        
        return try filterRequest.fetchAll(db)
    }
}

enum PropertyValue: Codable, Equatable {
    case string(String)
    case bool(Bool)
    
    func encode(to encoder: Encoder) throws {
        var container = encoder.singleValueContainer()
        switch self {
        case .string(let stringValue):
            try container.encode(stringValue)
        case .bool(let boolValue):
            try container.encode(boolValue)
        }
    }
    
    init(from decoder: Decoder) throws {
        let container = try decoder.singleValueContainer()
        if let bool = try? container.decode(Bool.self) {
            self = .bool(bool)
        } else if let string = try? container.decode(String.self) {
            self = .string(string)
        }else {
            throw DecodingError.dataCorruptedError(in: container, debugDescription: "AnyDecodable value cannot be decoded")
        }
    }
}

enum DatabaseQueryCondition: Codable, Equatable {
    // A property of this item equals a particular value
    case propertyEquals(PropertyEquals)
    
    // This item has an edge pointing to 'x' item
    case edgeHasTarget(EdgeHasTarget)
    
    case edgeHasSource(EdgeHasSource)

    enum CodingKeys: CodingKey {
        case propertyEquals, edgeHasTarget, edgeHasSource
    }
    
    struct PropertyEquals: Codable, Equatable {
        var name: String
        var value: PropertyValue
    }
    
    struct EdgeHasTarget: Codable, Equatable {
        var edgeName: String
        var target: Int64
    }
    
    struct EdgeHasSource: Codable, Equatable {
        var edgeName: String
        var source: Int64
    }
    
    func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        switch self {
        case .propertyEquals(let info):
            try container.encode(info, forKey: .propertyEquals)
        case .edgeHasTarget(let info):
            try container.encode(info, forKey: .edgeHasTarget)
        case .edgeHasSource(let info):
            try container.encode(info, forKey: .edgeHasSource)
        }
    }
    
    init(from decoder: Decoder) throws {
        let container = try decoder.container(keyedBy: CodingKeys.self)
        let key = container.allKeys.first
        
        switch key {
        case .propertyEquals:
            try self = .propertyEquals(container.decode(PropertyEquals.self, forKey: .propertyEquals))
        case .edgeHasTarget:
            try self = .edgeHasTarget(container.decode(EdgeHasTarget.self, forKey: .edgeHasTarget))
        case .edgeHasSource:
            try self = .edgeHasSource(container.decode(EdgeHasSource.self, forKey: .edgeHasSource))
        case .none:
            throw StringError(description: "Unknown database query condition")
        }
    }
}

extension DatabaseQueryConfig {
    
    static func queryConfigWith(context: CVUContext,
                                          datasource: CVUParsedDefinition? = nil,
                                          inheritQuery: DatabaseQueryConfig? = nil,
                                          overrideUIDs: Set<Int64>? = nil,
                                          targetItem: ItemRecord? = nil,
                                          dateRange: DateInterval? = nil) -> DatabaseQueryConfig {
        
        let datasourceResolver = datasource?.parsed.propertyResolver(context: context, lookup: CVULookupController(), db: AppController.shared.databaseController)
        let uidList = overrideUIDs ?? datasourceResolver.map {
            Set($0.intArray("uids").map { Int64($0) })
        }
        let filterDef = datasourceResolver?.subdefinition("filter")

        let edgeTargets = filterDef?.subdefinition("edgeTargets")
        let edgeTargetConditions = edgeTargets?.properties.keys.compactMap { key -> DatabaseQueryCondition? in
            guard let target = edgeTargets?.int(key) else { return nil }
            return .edgeHasTarget(DatabaseQueryCondition.EdgeHasTarget(edgeName: key, target: Int64(target)))
        } ?? []

        let edgeSources = filterDef?.subdefinition("edgeSources")
        let edgeSourceConditions = edgeSources?.properties.keys.compactMap { key -> DatabaseQueryCondition? in
            guard let source = edgeSources?.int(key) else { return nil }
            return .edgeHasSource(DatabaseQueryCondition.EdgeHasSource(edgeName: key, source: Int64(source)))
        } ?? []

        let properties = filterDef?.subdefinition("properties")
        let propertyConditions = properties?.properties.keys.compactMap { key -> DatabaseQueryCondition? in
            if let value = properties?.bool(key)  {
               return .propertyEquals(DatabaseQueryCondition.PropertyEquals(name: key, value: .bool(value)))
            } else if let value = properties?.string(key)  {
                return .propertyEquals(DatabaseQueryCondition.PropertyEquals(name: key, value: .string(value)))
            }
            return nil
        } ?? []

        var queryConfig = inheritQuery ?? DatabaseQueryConfig()
        if let itemTypes = datasourceResolver?.stringArray("query") ?? targetItem.map({ [$0.type] })  {
            queryConfig.itemTypes = itemTypes
        }
        if let itemUIDs = uidList {
            queryConfig.itemIDs = itemUIDs
        }
        
        if case let .constant(.string(edgeTargetsOperator)) = datasourceResolver?.properties["edgeTargetsOperator"] {
            queryConfig.edgeTargetsOperator = edgeTargetsOperator == ConditionOperator.or.rawValue ? .or : .and
        }

        if let sortProperty = datasourceResolver?.string("sortProperty") {
            queryConfig.sortProperty = sortProperty
        }
        if let sortAscending = datasourceResolver?.bool("sortAscending") {
            queryConfig.sortAscending = sortAscending
        }
        if let dateModifiedRange = dateRange {
            queryConfig.dateModifiedAfter = dateModifiedRange.start
            queryConfig.dateModifiedBefore = dateModifiedRange.end
        }
        if !edgeTargetConditions.isEmpty || !edgeSourceConditions.isEmpty || !propertyConditions.isEmpty {
            queryConfig.conditions = edgeTargetConditions + edgeSourceConditions + propertyConditions
        }
        
        if let count = datasourceResolver?.int("count") {
            queryConfig.count = count
        }

        return queryConfig
    }
}
