//
//  DemoData.swift
//  MemriDatabase
//
//  Created by T Brennan on 30/11/20.
//

import Foundation
import GRDB

enum DemoData {
    static var types: [String: SchemaType] = [:]

    static func importSchemaOnce(pool: DatabasePool) throws {
    
        let schemaItem = try pool.read { db in
            try? ItemRecord
                .filter(Column(ItemRecord.Columns.type.rawValue) == "ItemPropertySchema" || Column(ItemRecord.Columns.type.rawValue) == "ItemEdgeSchema")
                    .fetchOne(db)
        }
        
        guard schemaItem == nil else {
            return
        }

        let fileURL = Bundle.main.url(forResource: "schema", withExtension: "json")!
        let fileData = try Data(contentsOf: fileURL)
                
        guard let items = try JSONSerialization.jsonObject(with: fileData, options: []) as? [String: Any],
        let properties = items["properties"] as? [[String: Any]],
        let edges = items["edges"] as? [[String: Any]] else {
            throw StringError(description: "Could not locate schema file")
        }

        try pool.write { (db) in
            for property in properties {
                if let itemType = property["item_type"] as? String,
                   let propertyName = property["property"] as? String,
                   let propertyValue = property["value_type"] as? String {
                    
                    let record = ItemRecord(type: "ItemPropertySchema")
                    try record.insert(db)
                    
                    if let recordRowId = record.rowId {
                        try StringRecord(item: recordRowId, name: "itemType", value: itemType).insert(db)
                        try StringRecord(item: recordRowId, name: "propertyName", value: propertyName).insert(db)
                        try StringRecord(item: recordRowId, name: "valueType", value: propertyValue).insert(db)
                    }
                }
            }
            
            for edge in edges {
                if let sourceType = edge["source_type"] as? String,
                   let edgeName = edge["edge"] as? String,
                   let targetType = edge["target_type"] as? String {
                    
                    let record = ItemRecord(type: "ItemEdgeSchema")
                    try record.insert(db)
                    
                    if let recordRowId = record.rowId {
                        try StringRecord(item: recordRowId, name: "sourceType", value: sourceType).insert(db)
                        try StringRecord(item: recordRowId, name: "edgeName", value: edgeName).insert(db)
                        try StringRecord(item: recordRowId, name: "targetType", value: targetType).insert(db)
                    }
                }
            }
        }
    }
    
    static func loadSchema() throws {
        let fileURL = Bundle.main.url(forResource: "schema", withExtension: "json")!
        let fileData = try Data(contentsOf: fileURL)
        
        func handleError(_ string: String) throws {
            if Constants.isRunningTests {
                throw(StringError(description: string))
            } else {
                print(string)
            }
        }
        
        let fileDecoded = try JSONDecoder().decode(SchemaFile.self, from: fileData)
        
        let groupedProperties = Dictionary(grouping: fileDecoded.properties) { $0.itemType }
        let groupedEdges = Dictionary(grouping: fileDecoded.edges) { $0.sourceType }
        
        let allTypes = Set(groupedProperties.keys).union(groupedEdges.keys)
        
        // Confirm that all edge target types actually exist in the schema
        let allEdgeTargetTypes = Set(fileDecoded.edges.map { $0.targetType })
        let undefinedEdgeTargetTypes = allEdgeTargetTypes.subtracting(allTypes).subtracting(["Any"])
        if !undefinedEdgeTargetTypes.isEmpty {
            try handleError("Edge target types in schema for types that don't exist: \(undefinedEdgeTargetTypes)")
        }
        
        // Collate the schema into a format we can efficiently traverse
        types = Dictionary(uniqueKeysWithValues: allTypes.map { type -> (String, SchemaType) in
            return (type,
                    SchemaType(type: type,
                              propertyTypes: groupedProperties[type]
                                .map { $0.map { ($0.property, $0) } }
                                .map { Dictionary($0, uniquingKeysWith: { a, b in a }) } ?? [:],
                              edgeTypes: groupedEdges[type]
                                .map { $0.map { ($0.edge, $0) } }
                                .map { Dictionary($0, uniquingKeysWith: { a, b in a }) } ?? [:])
            )
        })
    }
    
    static func importDefaultData(databaseController: DatabaseController = AppController.shared.databaseController, throwIfAgainstSchema: Bool = false) throws {
        try importData(fileName: "default_database", databaseController: databaseController, throwIfAgainstSchema: throwIfAgainstSchema)
    }
    
    static func importDemoData(databaseController: DatabaseController = AppController.shared.databaseController, throwIfAgainstSchema: Bool = false) throws {
        try importData(fileName: "demo_database", databaseController: databaseController, throwIfAgainstSchema: throwIfAgainstSchema)
    }
    
    static func importData(fileName: String, databaseController: DatabaseController, throwIfAgainstSchema: Bool) throws {
        let fileURL = Bundle.main.url(forResource: fileName, withExtension: "json")!
        let fileData = try Data(contentsOf: fileURL)
        guard let items = try JSONSerialization.jsonObject(with: fileData, options: []) as? [[String: Any]] else {
            return
        }
        try loadSchema()
        
        let processedItems: [DemoData_Item] = try items.flatMap { try processItemJSON(item: $0) }
        
        try databaseController.writeSync { (db) in
            var rowIdLookup: [String: Int64] = [:]
            
            for item in processedItems {
                let record = ItemRecord(id: item.id, type: item.type, dateCreated: item.dateCreated, dateModified: item.dateModified, dateServerModified: item.dateServerModified)
                if record.type == "File" {
                    record.fileState = .needsUpload
                }
                try record.insert(db)
                let id = item.tempID ?? item.id.uid
                if let rowId = record.rowId {
                    rowIdLookup[id] = rowId
                }
                
            }
            
            for item in processedItems {
                let itemTempId = item.tempID ?? item.id.uid

                guard let itemRowId = rowIdLookup[itemTempId] else {
                    continue
                }

                for property in item.properties {
                    var recordItem: BaseRecord? = nil
                    switch property.value {
                    case .string(let stringValue):
                        recordItem = StringRecord(item: itemRowId, name: property.name, value: stringValue)
                    case .bool(let boolValue):
                        recordItem = IntegerRecord(item: itemRowId, name: property.name, value: boolValue == true ? 1 : 0)
                    case .int(let intValue):
                        recordItem = IntegerRecord(item: itemRowId, name: property.name, value: Int64(intValue))
                    case .double(let doubleValue):
                        recordItem = RealRecord(item: itemRowId, name: property.name, value: doubleValue)
                    case .datetime(let dateValue):
                        recordItem = RealRecord(item: itemRowId, name: property.name, value: Double(DatabaseHelperFunctions.encode(dateValue)))
                    case .blob(let blob):
                        if let blobString = String(data: blob, encoding: .utf8) {
                            recordItem = StringRecord(item: itemRowId, name: property.name, value: blobString)
                        }
                    }
                    
                    if let recordItem = recordItem {
                        try recordItem.insert(db)
                    }
                }
                for edge in item.edges {
                    guard let targetRowId = rowIdLookup[edge.tempTarget] else {
                        continue
                    }
                    let _ = try EdgeRecord.createFor(source: itemRowId, type: edge.type, target:targetRowId, db: db)
                }
            }
        }
    }


    static func processItemJSON(item: [String: Any], overrideUID: String? = nil) throws -> [DemoData_Item] {
        func handleError(_ string: String) throws {
            if Constants.isRunningTests {
                // Used for testing: throw an error if error in demo data
                throw(StringError(description: string))
            } else {
                // Notify developer of error in demo data, but continue
                print(string)
            }
        }
        
        guard let itemType = item["_type"] as? String else {
            try handleError("BAD RECORD: \(item)")
            return []
        }
        guard types[itemType] != nil else {
            print("\(itemType) not in schema")
            return []
        }
        
        let itemTempUID = overrideUID ?? (item["id"] as? Int).map(String.init)
        
        var items: [DemoData_Item] = []
        var properties: [DemoData_Property] = []
        var edges: [DemoData_Edge] = []
        
        // Fake a recent date for the demo data
        var dateCreated = Date(timeIntervalSinceNow: -TimeInterval.random(in: 0...1814400))
        var dateModified = dateCreated

        propertyIterator: for (propertyName, propertyValue) in item {
            switch propertyName {
            case "_type", "id":
                continue
            case "allEdges":
                guard let edgeList = propertyValue as? [[String: Any]] else { continue }
                for edge in edgeList {
                    guard let edgeType = edge["_type"] as? String else { continue }
                    if let target = (edge["id"] as? Int).map(String.init) {
                        edges.append(DemoData_Edge(type: edgeType, tempTarget: target))
                    } else if let subitem = edge["_target"] as? [String: Any] {
                        // Sub-item declared as edge, add edge property AND item
                        let target = UUID().uuidString
                        edges.append(DemoData_Edge(type: edgeType, tempTarget: target))
                        items.append(contentsOf: try processItemJSON(item: subitem, overrideUID: target))
                    }
                }
            case "dateCreated":
                guard let dateInMilliseconds = propertyValue as? Int else {
                    continue
                }
                let date = DatabaseHelperFunctions.decode(from: dateInMilliseconds)
                dateCreated = date
                dateModified = date
            default:
                guard let expectedType = types[itemType]?.propertyTypes[propertyName] else {
                    try handleError("\(itemType).\(propertyName) property not in schema")
                    continue
                }
                do {
                    var databaseValue = try PropertyDatabaseValue(value: propertyValue, propertyType: expectedType.valueType, debugInfo: "\(itemType).\(propertyName)")
                    switch ((itemType, propertyName, databaseValue)) {
                    case ("File", "filename", let .string(fileName)) :
                        let newFileName = "\(UUID().uuidString).\(fileName.fileExtension ?? "jpg")"
                        let url = FileStorageController.getFileStorageURL().appendingPathComponent(newFileName)
                        
                        guard let demoDirectory = Bundle.main.resourceURL?.appendingPathComponent("demoAssets", isDirectory: true) else {
                            continue
                        }
                        let sourcePath = demoDirectory.appendingPathComponent("\(fileName.fileName ?? "").\(fileName.fileExtension ?? "jpg")")
                        try FileManager.default.copyItem(atPath: sourcePath.path, toPath: url.path)

                        // Also add sha256 property for item
                        let sha256 = try FileHelper.getSHAHash(url: url)
                        properties.append(DemoData_Property(name: "sha256", value: .string(sha256)))
                        
                        databaseValue = .string(newFileName)
                    default:
                        break
                    }
                    properties.append(DemoData_Property(name: propertyName, value: databaseValue))
                }
                catch {
                    try handleError(error.localizedDescription)
                    continue
                }
            }
        }
        
        items.append(DemoData_Item(type: itemType, tempID: itemTempUID , properties: properties, edges: edges, dateCreated: dateCreated, dateModified: dateModified))
        return items
    }

    struct DemoData_Item {
        var type: String
        var id: StringUUID = StringUUID()
        var tempID: String?
        var properties: [DemoData_Property]
        var edges: [DemoData_Edge]
        var dateCreated: Date?
        var dateModified: Date?
        var dateServerModified: Date?
    }

    struct DemoData_Property {
        var name: String
        var value: PropertyDatabaseValue
    }

    struct DemoData_Edge {
        var type: String
        var tempTarget: String
    }

    struct SchemaFile: Codable {
        var properties: [SchemaProperty]
        var edges: [SchemaEdge]
    }
}
