//
//  EdgeRecord+Sync.swift
//  Memri

import Foundation
import GRDB

struct SyncUploadEdge {
    var operation: SyncState
    var properties: [String: AnyEncodable]
}

extension EdgeRecord {

    static func fromSyncEdgeDict(item: Int64, fromSyncItemDict dict: [String: AnyDecodable], schema: Schema, db: Database) throws -> EdgeRecord? {
        
        do {
            guard let edgeName = dict["_edge"]?.value as? String,
                  let sourceItemId = dict["_source"]?.value as? String,
                  let targetItemId = dict["_target"]?.value as? String else {
                return nil
            }
            
            guard let sourceItem = try ItemRecord.fetchWithID(StringUUID(uid: sourceItemId), db: db),
                  let sourceItemRowId = sourceItem.rowId,
                  let targetItem = try ItemRecord.fetchWithID(StringUUID(uid: targetItemId), db: db),
                  let targetItemRowId = targetItem.rowId else {
                print("ERROR could not find source or target item for edge: ", dict)
                return nil
            }

            var existingEdge = try sourceItem.request(for: ItemRecord.edges).filter(EdgeRecord.Columns.name == edgeName
                                                                                        && EdgeRecord.Columns.name == edgeName
                                                                                        && EdgeRecord.Columns.source == sourceItemRowId
                                                                                        && EdgeRecord.Columns.target == targetItemRowId).fetchOne(db)
            if (existingEdge == nil) {
                let edge = try EdgeRecord.createFor(item: item, source: sourceItemRowId, type: edgeName, target: targetItemRowId, syncState: .noChanges)
                existingEdge = edge
            }
            
            return existingEdge
        } catch let error {
            print("ERROR creating edge from sync: ", error)
        }
        
        return nil
    }
    
    static func syncEdgesWithState(_ state: SyncState, maxItems: Int = 100) throws -> [[String: AnyEncodable]] {
        guard let pool = AppController.shared.databaseController.databasePool else {
            throw StringError(description: "Database pool not available")
        }
        
        let schema = AppController.shared.databaseController.schema
        
        /// Select the items to sync, giving priority to those marked as `syncHasPriority`
        let syncEdges = try (pool.read { db in
            try? EdgeRecord
                .filter(Column(EdgeRecord.Columns.syncState.rawValue) == state.rawValue)
                .limit(maxItems)
                .fetchAll(db)
                .compactMap{edge -> [String: AnyEncodable]? in
                    return edge.syncDict(db: db, schema: schema)
                }
        }) ?? []
        
        return syncEdges
    }
    
    func syncDict(db: Database, schema: Schema) -> [String: AnyEncodable]? {
        guard let sourceItem = try? ItemRecord.fetchWithID(source, db: db),
              sourceItem.syncState != .create,
              let targetItem = try? ItemRecord.fetchWithID(target, db: db),
              targetItem.syncState != .create else {
            return nil
        }
        
        return ["_source": sourceItem.id, "_target": targetItem.id, "_name": self.name].mapValues { AnyEncodable($0) }
    }
    
    static func didSyncEdges(syncItems: PodAPIPayload.BulkAction, error: Error?) throws {
        guard  error == nil else {
            throw StringError(description: "Sync Failed")
        }
        
        guard let pool = AppController.shared.databaseController.databasePool else {
            throw StringError(description: "Database pool not available")
        }
                
        _ = try pool.write { (db) in
            for syncedEdge in syncItems.createEdges {
                if let sourceItemId = (syncedEdge["_source"]?.value as? StringUUID)?.uid,
                   let targetItemId = (syncedEdge["_target"]?.value as? StringUUID)?.uid,
                   let name = syncedEdge["_name"]?.value as? String,
                   let sourceItem = try ItemRecord.filter(ItemRecord.Columns.id == sourceItemId).fetchOne(db),
                    let sourceItemRowId = sourceItem.rowId,
                    let targetItem = try ItemRecord.filter(ItemRecord.Columns.id == targetItemId).fetchOne(db),
                    let targetItemRowId = targetItem.rowId,
                    let edgeRecord = try EdgeRecord.filter(EdgeRecord.Columns.source == sourceItemRowId
                                                            && EdgeRecord.Columns.target == targetItemRowId
                                                            && EdgeRecord.Columns.name == name).fetchOne(db) {
                    try  EdgeRecord
                        .filter(Column(EdgeRecord.Columns.source.rawValue) == edgeRecord.source
                                && Column(EdgeRecord.Columns.target.rawValue) == edgeRecord.target
                                    && Column(EdgeRecord.Columns.item.rawValue) == edgeRecord.item
                                    && Column(EdgeRecord.Columns.name.rawValue) == edgeRecord.name
                         )
                        .updateAll(db, [Column(EdgeRecord.Columns.syncState.rawValue).set(to: SyncState.noChanges.rawValue)])
                } else {
                    print("ERROR: Count not locate edge for synced edge: ", syncedEdge)
                }
            }
        }
    }
}
