//
//  EdgeRecord.swift
//  MemriDatabase
//
//  Created by T Brennan on 21/12/20.
//

import Foundation
import GRDB

class EdgeRecord: BaseRecord {
    static override var databaseTableName: String { "edges" }
    
    let item: Int64
    let source: Int64
    let name: String
    let target: Int64
    
    var syncState: SyncState
    var syncHasPriority: Bool

    static let itemForeignKey = ForeignKey(["itemId"])

    enum Columns: String, ColumnExpression {
        case item, source, name, target, syncState, syncHasPriority
    }
        
    static var ownerForeignKey = ForeignKey([Columns.item], to: [ItemRecord.Columns.rowId])
    static var owningItem = hasOne(ItemRecord.self, using: ownerForeignKey)

    static var sourceForeignKey = ForeignKey([Columns.source], to: [ItemRecord.Columns.rowId])
    static var sourceItem = hasOne(ItemRecord.self, using: sourceForeignKey)
    
    static var targetForeignKey = ForeignKey([Columns.target], to: [ItemRecord.Columns.rowId])
    static var targetItem = hasOne(ItemRecord.self, using: targetForeignKey)

    func sourceItem(db: Database) -> ItemRecord? {
        try? ItemRecord.filter(key: source).fetchOne(db)
    }
    
    func targetItem(db: Database) -> ItemRecord? {
        try? ItemRecord.filter(key: target).fetchOne(db)
    }
    
    static func createFor(item: Int64? = nil, source: Int64, type: String, target: Int64, syncState: SyncState = .create, syncHasPriority: Bool = false, db: Database) throws -> EdgeRecord {
        guard let existingEdge = try EdgeRecord.filter(EdgeRecord.Columns.source == source
                                                && EdgeRecord.Columns.target == target
                                                            && EdgeRecord.Columns.name == type).fetchOne(db) else {
            
            var itemRowId = item
            if (item == nil) {
                let edgeItem = ItemRecord(type: "Edge")
                edgeItem.syncState = .create
                try edgeItem.save(db: db)
                itemRowId = edgeItem.rowId
            }
            
            guard let rowId = itemRowId else {
                print("Error saving edge item")
                throw StringError(description: "Error saving edge item")
            }

            let edge = EdgeRecord(item: rowId, source: source, type: type, target: target, syncState: syncState, syncHasPriority: syncHasPriority)
            try edge.save(db: db)
            
            return edge
        }
        
        
        return existingEdge
    }
    
    static func createFor(item: Int64? = nil, source: Int64, type: String, target: Int64, syncState: SyncState = .create, syncHasPriority: Bool = false, dbController: DatabaseController = AppController.shared.databaseController) throws -> EdgeRecord {
        try dbController.writeSync { db in
            return try createFor(item: item, source: source, type: type, target: target, syncState: syncState, syncHasPriority: syncHasPriority, db: db)
        }
    }
    
    internal init(item: Int64, source: Int64, type: String, target: Int64, syncState: SyncState = .create, syncHasPriority: Bool = false) {
        self.item = item
        self.source = source
        self.name = type
        self.target = target
        self.syncState = syncState
        self.syncHasPriority = syncHasPriority

        super.init()
    }
    
    required init(row: Row) {
        self.item = row[Columns.item]
        source = row[Columns.source]
        name = row[Columns.name]
        target = row[Columns.target]
        syncState = row[Columns.syncState]
        syncHasPriority = row[Columns.syncHasPriority]

        super.init(row: row)
    }
    
    override func encode(to container: inout PersistenceContainer) {
        container[Columns.item] = item
        container[Columns.source] = source
        container[Columns.name] = name
        container[Columns.target] = target
        container[Columns.syncState] = syncState
        container[Columns.syncHasPriority] = syncHasPriority
    }
    
    func _syncDict() -> [String: AnyEncodable] {
        let keyProperties: [String: AnyEncodable] = [
            "_name": name,
            "_source": source,
            "_target": target
        ].mapValues { AnyEncodable($0) }

        return keyProperties
    }
    
    func save(db: Database) throws {
        try save(db)
    }
    
    func save(db dbController: DatabaseController = AppController.shared.databaseController) throws {
        try dbController.writeSync { (db) in
            try save(db: db)
        }
    }
    
    func getEdgeItem(db: Database, deleted: Bool = false) throws -> ItemRecord? {
        guard let item = try ItemRecord.fetchWithID(self.item, db: db),
              item.deleted == deleted else {
            return nil
        }
        
        return item
    }
    
    func markDeleted(db: Database) throws -> Bool {
        guard let item = try getEdgeItem(db: db) else {
            return false
        }
        
        return try item.markDeleted(db: db)
    }
    
    func markDeleted(db dbController: DatabaseController = AppController.shared.databaseController) throws -> Bool {
        try dbController.writeSync { (db) in
            return try markDeleted(db: db)
        }
    }
}
