//
//  ItemRecord+Auth.swift
//  Memri
//
//  Created by Chaitanya Pandit on 01/05/21.
//

import Foundation
import GRDB

enum AuthKeyType: String {
    case dbKey = "Memri Database Key"
    case ownerKey = "Memri Owner Key"
}

struct AuthKey {
    let ownerId: Int64
    let type: String
    let role: String?
    let key: String
    let name: String
    let active: Bool

    func save() throws -> ItemRecord {
        let item = ItemRecord(type: "CryptoKey")
        try item.save()

        guard let itemRowId = item.rowId else {
            throw (StringError(description: "Error saving key"))
        }

        try StringRecord(item: itemRowId, name: "itemType", value: type).save()
        if let role = role {
            try StringRecord(item: itemRowId, name: "role", value: role).save()
        }
        try StringRecord(item: itemRowId, name: "keystr", value: key).save() // "keystr" because "key" is restricted in pod db
        try StringRecord(item: itemRowId, name: "name", value: name).save()
        try IntegerRecord(item: itemRowId, name: "active", value: active ? 1 : 0).save()

        return item
    }
}

struct AuthKeys: Codable {
    let ownerKey: String
    let dbKey: String
}

extension ItemRecord {
    static func deleteExistingDBKeys() throws {
        guard let dbKeys = try (AppController.shared.databaseController.read {db in
            try? ItemRecord
                .filter(Column(ItemRecord.Columns.type.rawValue) == "CryptoKey")
                .fetchAll(db)
        }) else {
            return
        }

        try AppController.shared.databaseController.write {db in
            for key in dbKeys {
                try key.delete(db)
            }
        }
    }

    static func setOwnerAndDBKey(privateKey: String, publicKey: String, dbKey: String) throws {
        do {
            try ItemRecord.deleteExistingDBKeys()

            guard let meRowId = try ItemRecord.me?.rowId ?? ItemRecord.createMe()?.rowId else {
                throw StringError(description: "Cound not find me user")
            }

            _ = try AuthKey(ownerId: meRowId, type: "64CharacterRandomHex", role: nil, key: dbKey, name: AuthKeyType.dbKey.rawValue, active: true).save()
            let privateKeyItem = try AuthKey(ownerId: meRowId, type: "ED25519", role: "private", key: privateKey, name: AuthKeyType.ownerKey.rawValue, active: true).save()
            let publicKeyItem = try AuthKey(ownerId: meRowId, type: "ED25519", role: "public", key: publicKey, name: AuthKeyType.ownerKey.rawValue, active: true).save()

            guard let privateKeyItemRowId = privateKeyItem.rowId,
                  let publicKeyItemRowId = publicKeyItem.rowId
            else {
                throw StringError(description: "Error saving keys, invalid rowIds")
            }

            try _ = EdgeRecord.createFor(source: privateKeyItemRowId, type: "owner", target: meRowId)
            try _ = EdgeRecord.createFor(source: publicKeyItemRowId, type: "owner", target: meRowId)
            try _ = EdgeRecord.createFor(source: privateKeyItemRowId, type: "publicKey", target: publicKeyItemRowId)
            try _ = EdgeRecord.createFor(source: publicKeyItemRowId, type: "privateKey", target: privateKeyItemRowId)
        } catch {
            print("ERROR: setOwnerAndDBKey", error)
            throw StringError(description: "Error deleting existing db keys")
        }
    }

    static func getOwnerAndDBKey() throws -> AuthKeys {
        guard let meRowId = ItemRecord.me?.rowId else {
            throw StringError(description: "Cound not find me user")
        }

        guard let items = try (AppController.shared.databaseController.read { db in
            try? ItemRecord
                .filter(Column(ItemRecord.Columns.type.rawValue) == "CryptoKey")
                .fetchAll(db)
        }),
            let ownerKey = items.filter({ $0.propertyValue("role") == .string("public")
                    && $0.propertyValue("active") == .bool(true)
                    && $0.edgeItem("owner")?.rowId == meRowId }).first,
            let dbKey = items.filter({ $0.propertyValue("name") == .string(AuthKeyType.dbKey.rawValue)
                && $0.propertyValue("active") == .bool(true)}).first,
            case let .string(ownerKeyValue) = ownerKey.propertyValue("keystr"),
            case let .string(dbKeyKeyValue) = dbKey.propertyValue("keystr")
        else {
            throw StringError(description: "Keys not found")
        }

        return AuthKeys(ownerKey: ownerKeyValue, dbKey: dbKeyKeyValue)
    }
}
