//
//  ItemRecord+Properties.swift
//  Memri
//
//  Created by Chaitanya Pandit on 19/04/21.
//

import Foundation
import GRDB

protocol ItemPropertyRecord {
    var item: Int64 { get }
    var name: String { get }
    func databaseValueForType(type: String, schema: Schema) -> PropertyDatabaseValue
}

extension ItemRecord {
    
    func propertyValue(_ name: String, db: DatabaseController = AppController.shared.databaseController) -> PropertyDatabaseValue? {
        guard let expectedType = db.schema.expectedType(forItemType: self.type, propertyName: name) else {
            return nil
        }
        
        switch expectedType {
        case .string:
            guard let record = StringRecord.propertyForItem(self, name: name) else {
                return nil
            }
            return .string(record.value)
        case .bool:
            guard let record = IntegerRecord.propertyForItem(self, name: name) else {
                return nil
            }
            return .bool(record.value == 1 ? true : false)
        case .int:
            guard let record = IntegerRecord.propertyForItem(self, name: name) else {
                return nil
            }
            return .int(Int(record.value))
        case .double:
            guard let record = RealRecord.propertyForItem(self, name: name) else {
                return nil
            }
            return .double(record.value)
        case .datetime:
            guard let record = RealRecord.propertyForItem(self, name: name) else {
                return nil
            }
            return .datetime(DatabaseHelperFunctions.decode(from: Int(record.value)))
        case .blob:
            guard let record = StringRecord.propertyForItem(self, name: name),
                  let data = record.value.data(using: .utf8) else {
                return nil
            }
            return .blob(data)
        }
    }
    
    func propertyValues(_ name: String, db: DatabaseController = AppController.shared.databaseController) -> [PropertyDatabaseValue] {
        return (try? db.read{ (db) in
            propertyValues(name, db: db)
        }) ?? []
    }
    
    func propertyValues(_ name: String, db: Database) -> [PropertyDatabaseValue]? {
        guard let rowId = self.rowId,
              // TODO: Make schema shared?
              let expectedType = AppController.shared.databaseController.schema.expectedType(forItemType: self.type, propertyName: name)  else {
            return []
        }
        
        switch expectedType {
        case .string:
            return StringRecord.propertiesForItem(rowId, name: name, db: db)?.map{.string($0.value)}
        case .bool:
            return IntegerRecord.propertiesForItem(rowId, name: name, db: db)?.map{.bool($0.value == 1 ? true : false)}
        case .int:
            return IntegerRecord.propertiesForItem(rowId, name: name, db: db)?.map{.int(Int($0.value))}
        case .double:
            return RealRecord.propertiesForItem(rowId, name: name, db: db)?.map{.double($0.value)}
        case .datetime:
            return RealRecord.propertiesForItem(rowId, name: name, db: db)?.map{.datetime(DatabaseHelperFunctions.decode(from: Int($0.value)))}
        case .blob:
            return []
        }
    }
    
    // eg: get phoneNumber property from Item "PhoneNumber" belonging to Person via edge "hasPhoneNumber"
    func edgePropertyValues(_ name: String, edgeName: String, forItem sourceItem: ItemRecord, db: Database) -> [PropertyDatabaseValue] {
        
        guard let edgeItems = sourceItem.edgeItems(edgeName, db: db) else {
            return []
        }

        return edgeItems.compactMap{$0.propertyValue(name)}
    }
    
    func setEdgePropertyValue(_ name: String, value: PropertyDatabaseValue, edgeName: String, forItem sourceItem: ItemRecord, schema: Schema, db: Database) {
        
        guard let edgeItems = sourceItem.edgeItems(edgeName, db: db) else {
            return
        }

        for item in edgeItems {
            try? item.setPropertyValue(name: name, value: value, db: db)
        }
    }
    
    func setPropertyValue(
          name: String,
          value: PropertyDatabaseValue?,
          addLog: Bool = false,
          db dbController: DatabaseController = AppController.shared.databaseController
      ) throws {
        try dbController.writeSync { (db) in
            try setPropertyValue(name: name, value: value, db: db)
            // If created, don't change to update else will go in sync payload as updatedItems
            if (self.syncState != .skip && self.syncState != .create) {
                self.syncState = .update
            }
            try self.update(db)
            if (addLog) {
                try addChangeLog(name: name, value: value, db: dbController)
            }
        }
    }
    
    func setPropertyValue(
          name: String,
          value: PropertyDatabaseValue?,
          db: Database
    ) throws {
        do {
            switch value {
            case .string(let stringValue):
                try StringRecord.setPropertyValue(name: name, value: stringValue, item: self, db: db)
            case .bool(let boolValue):
                try IntegerRecord.setPropertyValue(name: name, value: boolValue == true ? 1 : 0, item: self, db: db)
            case .int(let intValue):
                try IntegerRecord.setPropertyValue(name: name, value: intValue, item: self, db: db)
            case .double(let doubleValue):
                try RealRecord.setPropertyValue(name: name, value: doubleValue, item: self, db: db)
            case .datetime(let dateTimeValue):
                try RealRecord.setPropertyValue(name: name, value: Double(DatabaseHelperFunctions.encode(dateTimeValue)), item: self, db: db)
            case .blob(let data):
                if let stringValue = String(data: data, encoding: .utf8) {
                    try StringRecord.setPropertyValue(name: name, value: stringValue, item: self, db: db)
                }
            default :
                break
            }
        } catch let error {
           print("ERROR saving Property: ", error)
        }
    }

    func addChangeLog(
        name: String,
        value: PropertyDatabaseValue?,
        db dbController: DatabaseController = AppController.shared.databaseController
    ) throws {
        
        let auditItem = ItemRecord(type: "AuditItem")
        try auditItem.save()
        
        
        guard let rowId = self.rowId,
              let auditItemRowId = auditItem.rowId else {
            throw StringError(description: "Add changelog: Item doesn't have row id, possibly not saved?")
        }

        try IntegerRecord.setPropertyValue(name: "date", value: DatabaseHelperFunctions.encode(Date()), item: auditItem, db: dbController)
        try StringRecord.setPropertyValue(name: "content", value: name, item: auditItem, db: dbController)
        try StringRecord.setPropertyValue(name: "actionName", value: "edit", item: auditItem, db: dbController)
        try EdgeRecord.createFor(source: rowId, type: "changelog", target: auditItemRowId)
    }
}
