//
//  ItemRecord+Sync.swift
//  Memri
//
//  Created by Chaitanya Pandit on 04/05/21.
//

import Foundation
import GRDB

struct SyncUploadItem {
    var operation: SyncState
    var properties: [String: AnyEncodable]
}

struct SyncDownloadItem {
    var item: ItemRecord
    var properties: [BaseRecord]
    

}

extension ItemRecord {
    
    static func mapSchemaValueType(_ nativeType: String) -> String? {
        switch nativeType {
        case "string":
            return "Text"
        case "int":
            return "Integer"
        case "double":
            return "Real"
        case "bool":
            return "Bool"
        case "datetime":
            return "DateTime"
        case "blob":
            return "Text"
        default:
            return nativeType
        }
    }
    
    static func reverseMapSchemaValueType(_ valueType: String) -> String {
        switch valueType {
        case "Text":
            return "string"
        case "Integer":
            return "int"
        case "Real":
            return "double"
        case "Bool":
            return "bool"
        case "DateTime":
            return "datetime"
        default:
            return valueType
        }
    }
    
    static func escapeReservedKeywordsInPod(_ propertyName: String) -> String? {
        switch propertyName {
        case "action":
            return "actionName"
        case "query":
            return "queryName"
        case "key":
            return "keyName"
        case "type":
            return "typeName"
        default:
            return propertyName
        }
    }
    
    static func unEscapeReservedKeywordsInPod(_ propertyName: String) -> String {
        switch propertyName {
        case "actionName":
            return "action"
        case "queryName":
            return "query"
        case "keyName":
            return "key"
        case "typeName":
            return "type"
        default:
            return propertyName
        }
    }
    
    static func fromSyncItemDict(fromSyncItemDict dict: [String: AnyDecodable], schema: Schema, db: Database) throws -> ItemRecord? {
        
        guard let id = dict["id"]?.value as? String else {
            return nil
        }
        
        var item: ItemRecord? = try? ItemRecord.fetchWithID(StringUUID(uid: id), db: db)
        if (item == nil) {
            item = try ItemRecord(fromSyncDict: dict)
            try item?.save(db: db)
        }
        
        guard let createdItem = item,
              let itemRowId = createdItem.rowId else {
            return nil
        }

        if let dateServerModified = dict["dateServerModified"]?.value as? Int {
            let date = DatabaseHelperFunctions.decode(from: dateServerModified)
            createdItem.dateServerModified = date
            try createdItem.save(db: db)
        }
        
        try dict.forEach { (propertyName, decodableValue) in
            let unescapedPropertyName = unEscapeReservedKeywordsInPod(propertyName)
            
            guard !ItemRecord.intrinsicProperties.contains(unescapedPropertyName),
                  let expectedType = schema.expectedType(forItemType: createdItem.type, propertyName: unescapedPropertyName) else {
                return
            }
            
            var propertyValue = decodableValue.value
            if let itemType = dict["type"]?.value as? String,
               ["ItemPropertySchema", "ItemEdgeSchema"].contains(itemType),
               ["valueType", "sourceType", "targetType"].contains(propertyName),
               let mappedValue = propertyValue as? String {
                propertyValue = ItemRecord.reverseMapSchemaValueType(mappedValue)
            }
            
            let databaseValue = try PropertyDatabaseValue(value: propertyValue, propertyType: expectedType)
            try createdItem.setPropertyValue(name: unescapedPropertyName, value: databaseValue)
            
            // If the item has file and it does not exist on disk, mark the file to be downloaded
            if createdItem.type == "File",
               propertyName == "filename",
               let fileName = decodableValue.value as? String,
               !FileManager.default.fileExists(atPath: FileStorageController.getFileStorageURL().appendingPathComponent(fileName).path) {
                createdItem.fileState = .needsDownload
                try createdItem.save(db: db)
            }
        }

        // Check if it has edge and update
        let _ = try EdgeRecord.fromSyncEdgeDict(item: itemRowId, fromSyncItemDict: dict, schema: schema, db: db)

        return createdItem
    }
        
    func schemaPropertyDict(db: Database) -> [String: AnyEncodable]? {
        guard let itemRowId = self.rowId,
              let itemType = StringRecord.propertyForItem(itemRowId, name: "itemType", db: db)?.value,
              let propertyName = StringRecord.propertyForItem(itemRowId, name: "propertyName", db: db)?.value,
              let valueType = StringRecord.propertyForItem(itemRowId, name: "valueType", db: db)?.value,
              let schemaValueType = ItemRecord.mapSchemaValueType(valueType) else {
            return nil
        }
        
        return ["id": AnyEncodable(self.id.uid), "type": "ItemPropertySchema", "itemType":AnyEncodable(itemType), "propertyName": AnyEncodable(ItemRecord.escapeReservedKeywordsInPod(propertyName)), "valueType": AnyEncodable(schemaValueType)]
    }
    
    func schemaEdgeDict(db: Database) -> [String: AnyEncodable]? {
        guard let itemRowId = self.rowId,
              let edgeName = StringRecord.propertyForItem(itemRowId, name: "edgeName", db: db)?.value,
              let sourceType = StringRecord.propertyForItem(itemRowId, name: "sourceType", db: db)?.value,
              let targetType = StringRecord.propertyForItem(itemRowId, name: "targetType", db: db)?.value  else {
            return nil
        }
        
        return ["id": AnyEncodable(self.id.uid), "type": "ItemEdgeSchema", "sourceType":AnyEncodable(sourceType), "edgeName": AnyEncodable(ItemRecord.escapeReservedKeywordsInPod(edgeName)), "targetType": AnyEncodable(targetType)]
    }
    
    func syncDict(db: Database, schema: Schema) -> [String: AnyEncodable] {
        let stringProperties = try! request(for: ItemRecord.stringProperties).fetchAll(db)
        let integerProperties = try! request(for: ItemRecord.integerProperties).fetchAll(db)
        let realProperties = try! request(for: ItemRecord.realProperties).fetchAll(db)

        var properties = [ItemPropertyRecord]()
        properties.append(contentsOf: stringProperties)
        properties.append(contentsOf: integerProperties)
        properties.append(contentsOf: realProperties)

        return _syncDict(properties: properties, schema: schema)
    }
    
    func _syncDict(properties: [ItemPropertyRecord], schema: Schema) -> [String: AnyEncodable] { //, edges: [ItemEdgeRecord]
        let keyProperties: [String: AnyEncodable] = [
            "type": type,
            "id": id.uid,
            "dateCreated": DatabaseHelperFunctions.encode(dateCreated),
            "dateModified": DatabaseHelperFunctions.encode(dateModified),
            "deleted": deleted,
        ].mapValues { AnyEncodable($0) }

        let otherProperties: [(String, AnyEncodable)] = properties.compactMap {
            let propertyValue = $0.databaseValueForType(type: type, schema: schema)
            return ($0.name, AnyEncodable(propertyValue))
        }

        return keyProperties
            .merging(otherProperties, uniquingKeysWith: { a, b in a })
    }
    
    static func syncItemsWithState(_ state: SyncState, maxItems: Int = 100) throws -> [[String: AnyEncodable]] {
        guard let pool = AppController.shared.databaseController.databasePool else {
            throw StringError(description: "Database pool not available")
        }
        
        let schema = AppController.shared.databaseController.schema
        
        /// Select the items to sync, giving priority to those marked as `syncHasPriority`
        let syncItems = try (pool.read { db in
            try? ItemRecord
                .filter(Column(ItemRecord.Columns.type.rawValue) != "ItemPropertySchema"
                            && Column(ItemRecord.Columns.type.rawValue) != "ItemEdgeSchema"
                            && Column(ItemRecord.Columns.syncState.rawValue) == state.rawValue)
                .limit(maxItems)
                .fetchAll(db)
                .map{item -> [String: AnyEncodable] in
                    return item.syncDict(db: db, schema: schema)
                }
        }) ?? []
        
        return syncItems
    }
    
    static func didSyncItems(syncItems: PodAPIPayload.BulkAction, error: Error?) throws {
        guard  error == nil else {
            throw StringError(description: "Sync Failed")
        }
        
        guard let pool = AppController.shared.databaseController.databasePool else {
            throw StringError(description: "Database pool not available")
        }
        
        let createItemIDs = syncItems.createItems.compactMap{$0["id"]?.value as? String}
        let updateItemIDs = syncItems.updateItems.compactMap{$0["id"]?.value as? String}

        var allItems = createItemIDs
        allItems.append(contentsOf: updateItemIDs)
        let now = Date()
        
        _ = try pool.write { (db) in
            for itemId in allItems {
                if let item = try ItemRecord.filter(Columns.id == itemId).fetchOne(db) {
                    item.syncState = .noChanges
                    item.dateServerModified = now
                    try item.update(db)
                }
            }
        }
    }
    
    static func lastSyncedItem() throws -> ItemRecord? {
        guard let pool = AppController.shared.databaseController.databasePool else {
            throw StringError(description: "Database pool not available")
        }
                
        /// Select the items to sync, giving priority to those marked as `syncHasPriority`
        return try (pool.read { db in
            try? ItemRecord
                .order(ItemRecord.Columns.dateServerModified.desc)
                .fetchOne(db)
        })
    }
    
    static func fileItemRecordToUpload() throws -> (item: ItemRecord, fileName: String)? {
        guard let pool = AppController.shared.databaseController.databasePool else {
            throw StringError(description: "Database pool not available")
        }
                
        /// Select the items to sync, giving priority to those marked as `syncHasPriority`
        guard let item = try (pool.read { db in
            try? ItemRecord
                .filter(Column(ItemRecord.Columns.type.rawValue) == "File"
                            && Column(ItemRecord.Columns.fileState.rawValue) == FileState.needsUpload)
                .fetchOne(db)
        }) else {
            return nil
        }
        
        guard case let .string(fileName) = item.propertyValue("filename") else {
            return nil
        }
        
        return (item: item, fileName: fileName)
    }
    
    static func didUploadFileForItem(_ item: ItemRecord) throws {
        guard let rowId = item.rowId,
              let fetchedItem = ItemRecord.fetchWithID(rowId) else {
            return
        }
        
        fetchedItem.fileState = .noChanges
        try fetchedItem.save()
    }
    
    static func fileItemRecordToDownload() throws -> (item: ItemRecord, sha256: String, fileName: String)? {
        guard let pool = AppController.shared.databaseController.databasePool else {
            throw StringError(description: "Database pool not available")
        }
                
        /// Select the items to sync, giving priority to those marked as `syncHasPriority`
        guard let item = try (pool.read { db in
            try? ItemRecord
                .filter(Column(ItemRecord.Columns.type.rawValue) == "File"
                            && Column(ItemRecord.Columns.fileState.rawValue) == FileState.needsDownload)
                .fetchOne(db)
        }) else {
            return nil
        }
        
        guard case let .string(sha256) = item.propertyValue("sha256"),
              case let .string(fileName) = item.propertyValue("filename") else {
            return nil
        }
        
        return (item: item, sha256: sha256, fileName: fileName)
    }
    
    static func didDownloadFileForItem(_ item: ItemRecord) throws {
        guard let rowId = item.rowId,
              let fetchedItem = ItemRecord.fetchWithID(rowId) else {
            return
        }
        
        fetchedItem.fileState = .noChanges
        try fetchedItem.save()
    }
}
    
