//
//  ItemRecord.swift
//  MemriDatabase
//
//  Created by T Brennan on 21/12/20.
//

import Foundation
import GRDB

enum SyncState: String, Codable, Hashable, DatabaseValueConvertible {
    case skip
    case create
    case update
    case noChanges
    case failed
}

enum FileState: String, Codable, Hashable, DatabaseValueConvertible {
    case skip
    case needsUpload
    case needsDownload
    case noChanges
}

class ItemRecord: BaseRecord, Equatable, Hashable, Identifiable, Codable {
    static override var databaseTableName: String { "items" }
    
    var rowId: Int64?
    let id: StringUUID
    let type: String
    let dateCreated: Date
    var dateModified: Date
    var dateServerModified: Date
    var deleted: Bool
    
    var syncState: SyncState
    var fileState: FileState

    var syncHasPriority: Bool

    static let edges = hasMany(EdgeRecord.self, using: EdgeRecord.sourceForeignKey)
    static var stringProperties = hasMany(StringRecord.self, using: StringRecord.itemForeignKey)
    static var integerProperties = hasMany(IntegerRecord.self, using: IntegerRecord.itemForeignKey)
    static var realProperties = hasMany(RealRecord.self, using: RealRecord.itemForeignKey)

    enum Columns: String, ColumnExpression {
        case rowId, id, type, dateCreated, dateModified, dateServerModified, deleted, syncState, fileState, syncHasPriority
    }
    
    static func == (lhs: ItemRecord, rhs: ItemRecord) -> Bool {
        lhs.id == rhs.id && lhs.type == rhs.type
    }
    
    func hash(into hasher: inout Hasher) {
        hasher.combine(id)
        hasher.combine(type)
    }
    
    static func fetchWithID(_ id: StringUUID, db: DatabaseController = AppController.shared.databaseController) -> ItemRecord? {
        do {
            return try db.read { db in
                try ItemRecord.fetchWithID(id, db: db)
            }
        } catch let error {
            print("ERROR ItemRecord: fetchWithID: \(id) error: \(error)")
            return nil
        }
    }

    static func fetchWithID(_ id: StringUUID, db: Database) throws -> ItemRecord? {
        try ItemRecord.filter(Columns.id == id).fetchOne(db)
    }
    
    static func fetchWithID(_ id: Int64, db: DatabaseController = AppController.shared.databaseController) -> ItemRecord? {
        do {
            return try db.read { db in
                try fetchWithID(id, db: db)
            }
        } catch let error {
            print("ERROR ItemRecord: fetchWithID: \(id) error: \(error)")
            return nil
        }
    }
    
    static func fetchWithID(_ id: Int64, db: Database) throws -> ItemRecord? {
        try ItemRecord.filter(Columns.rowId == id).fetchOne(db)
    }
    
    func save(db: Database) throws {
        try save(db)
    }
    
    func save(db dbController: DatabaseController = AppController.shared.databaseController) throws {
        try dbController.writeSync { (db) in
            try save(db: db)
        }
    }
    
    func markDeleted(db: Database) throws -> Bool {
        self.deleted = true
        self.syncState = .update
        try save(db: db)
        
        let edges = (self.edges(nil, db: db) ?? []) + (self.reverseEdges(nil, db: db) ?? [])
        for edge in edges {
            _ = try edge.markDeleted(db: db)
        }
        
        return true
    }
    
    func markDeleted(db dbController: DatabaseController = AppController.shared.databaseController) throws -> Bool {
        try dbController.writeSync { (db) in
            try markDeleted(db: db)
        }
    }
    
    func edge(_ name: String, deleted: Bool = false, db: Database) -> EdgeRecord? {
        let edge = try? request(for: ItemRecord.edges).filter(EdgeRecord.Columns.name == name).fetchOne(db)
        
        // Ensure the edge item's deleted state matches as well
        guard let _ = try? edge?.getEdgeItem(db: db, deleted: deleted) else {
            return nil
        }
        
        return edge
    }
    
    func edge(_ name: String, deleted: Bool = false, db dbController: DatabaseController = AppController.shared.databaseController) -> EdgeRecord? {
        try? dbController.read { db in
            edge(name, deleted: deleted, db: db)
        }
    }
    
    func edges(_ name: String?, deleted: Bool = false, db: Database) -> [EdgeRecord]? {
        var request = request(for: ItemRecord.edges)
        if let name = name {
            request = request.filter(EdgeRecord.Columns.name == name)
        }

        var edgeRecords = (try? request.fetchAll(db)) ?? []
        edgeRecords = edgeRecords.filter {edge in
            guard let _ = try? edge.getEdgeItem(db: db, deleted: deleted) else {
                return false
            }
            
            return true
        }
        
        return edgeRecords
    }
    
    func edges(_ type: String? = nil, db dbController: DatabaseController = AppController.shared.databaseController) -> [EdgeRecord] {
        (try? dbController.read { db in
            edges(type, deleted: deleted, db: db)
        }) ?? []
    }
    
    func edgeItem(_ type: String, deleted: Bool = false, db dbController: DatabaseController = AppController.shared.databaseController) -> ItemRecord? {
        try? dbController.read { db in
            edge(type, deleted: deleted, db: db)?.targetItem(db: db)
        }
    }
    
    func edgeItems(_ type: String?, deleted: Bool = false, db: Database) -> [ItemRecord]? {
        edges(type, deleted: deleted, db: db)?.lazy.compactMap { $0.targetItem(db: db) }
    }
    
    func edgeItems(_ type: String? = nil, deleted: Bool = false, db dbController: DatabaseController = AppController.shared.databaseController) -> [ItemRecord] {
        (try? dbController.read { db in
            edgeItems(type, deleted: deleted, db: db)
        }) ?? []
    }
    
    static var reverseEdges = hasMany(EdgeRecord.self, using: EdgeRecord.targetForeignKey)
    
    func reverseEdge(_ name: String, deleted: Bool = false, db: Database) -> EdgeRecord? {
        let edge = try? request(for: ItemRecord.reverseEdges).filter(EdgeRecord.Columns.name == name).fetchOne(db)
        
        // Ensure the edge item's deleted state matches as well
        guard let _ = try? edge?.getEdgeItem(db: db, deleted: deleted) else {
            return nil
        }
        
        return edge
    }
    
    func reverseEdge(_ name: String, deleted: Bool = false, db dbController: DatabaseController = AppController.shared.databaseController) -> EdgeRecord? {
        try? dbController.read { db in
            reverseEdge(name, deleted: deleted, db: db)
        }
    }
    
    func reverseEdges(_ name: String?, deleted: Bool = false, db: Database) -> [EdgeRecord]? {
        var request = request(for: ItemRecord.reverseEdges)
        if let name = name {
            request = request.filter(EdgeRecord.Columns.name == name)
        }

        var edgeRecords = (try? request.fetchAll(db)) ?? []
        edgeRecords = edgeRecords.filter{edge in
            guard let _ = try? edge.getEdgeItem(db: db, deleted: deleted) else {
                return false
            }
            
            return true
        }
        
        return edgeRecords
    }
    
    func reverseEdges(_ type: String?, deleted: Bool = false, db dbController: DatabaseController = AppController.shared.databaseController) -> [EdgeRecord] {
        (try? dbController.read { db in
            reverseEdges(type, deleted:deleted, db: db)
        }) ?? []
    }
    
    func reverseEdgeItem(_ name: String?, deleted: Bool = false, dbController: DatabaseController = AppController.shared.databaseController) -> ItemRecord? {
        try? dbController.read { db in
            reverseEdge(type, deleted: deleted, db: db)?.targetItem(db: db)
        }
    }
    
    func reverseEdgeItems(_ name: String?, deleted: Bool = false, db: Database) -> [ItemRecord]? {
        reverseEdges(type, deleted: deleted, db: db)?.lazy.compactMap { $0.targetItem(db: db) }
    }
    
    func reverseEdgeItems(_ name: String? = nil, deleted: Bool = false, db: DatabaseController = AppController.shared.databaseController) -> [ItemRecord] {
        (try? db.read { db in
            reverseEdgeItems(name, deleted: deleted, db: db)
        }) ?? []
    }

    func reverseEdgeObjects(_ name: String? = nil, db: Database) -> [EdgeRecord]? {
        if let edgeName = name {
            return try? request(for: ItemRecord.reverseEdges).filter(EdgeRecord.Columns.name == edgeName).fetchAll(db)
        } else {
            return try? request(for: ItemRecord.reverseEdges).fetchAll(db)
        }
    }
    
    func reverseEdgeObjects(_ type: String? = nil, db dbController: DatabaseController = AppController.shared.databaseController) -> [EdgeRecord] {
        (try? dbController.read { db in
            reverseEdgeObjects(type, db: db)
        }) ?? []
    }

    internal init(rowId: Int64? = nil, id: StringUUID = StringUUID(), type: String, dateCreated: Date? = nil, dateModified: Date? = nil, dateServerModified: Date? = nil, deleted: Bool = false, syncState: SyncState = .create, fileState: FileState = .skip, syncHasPriority: Bool = false) {
        self.rowId = rowId
        self.id = id
        self.type = type
        self.dateCreated = dateCreated ?? Date()
        self.dateModified = dateModified ?? Date()
        self.dateServerModified = dateServerModified ?? Date.distantPast
        self.deleted = deleted
        self.syncState = syncState
        self.fileState = fileState
        self.syncHasPriority = syncHasPriority
        super.init()
    }
    
    required init(row: Row) {
        self.rowId = row[Columns.rowId]
        self.id = row[Columns.id]
        self.type = row[Columns.type]
        self.dateCreated = row[Columns.dateCreated]
        self.dateModified = row[Columns.dateModified]
        self.dateServerModified = row[Columns.dateServerModified]
        self.deleted = row[Columns.deleted]
        self.syncState = row[Columns.syncState]
        self.fileState = row[Columns.fileState]
        self.syncHasPriority = row[Columns.syncHasPriority]
        super.init(row: row)
    }
    
    override func encode(to container: inout PersistenceContainer) {
        container[Columns.rowId] = rowId
        container[Columns.id] = id
        container[Columns.type] = type
        container[Columns.dateCreated] = dateCreated
        container[Columns.dateModified] = dateModified
        container[Columns.dateServerModified] = dateServerModified
        container[Columns.deleted] = deleted
        container[Columns.syncState] = syncState
        container[Columns.fileState] = fileState
        container[Columns.syncHasPriority] = syncHasPriority
    }
    
    static let intrinsicProperties: Set<String> = ["type", "id", "dateCreated", "dateModified", "dateServerModified", "deleted"]

    init(fromSyncDict dict: [String: AnyDecodable]) throws {
        guard let uidString = dict["id"]?.value as? String,
              let type = dict["type"]?.value as? String,
              let dateCreated = (dict["dateCreated"]?.value as? Int).flatMap(DatabaseHelperFunctions.decode)
        else {
            throw NSError()
        }
        self.id = StringUUID(uid: uidString)
        self.type = type
        self.dateCreated = dateCreated
        self.dateModified = (dict["dateModified"]?.value as? Int).flatMap(DatabaseHelperFunctions.decode) ?? dateCreated
        self.dateServerModified = (dict["dateServerModified"]?.value as? Int).flatMap(DatabaseHelperFunctions.decode) ?? dateCreated
        self.deleted = dict["deleted"]?.value as? Bool ?? false
        
        // This is direct from the latest version so no need to sync again
        self.syncState = .noChanges
        self.fileState = .skip
        self.syncHasPriority = false
        super.init()
    }
    
    static func search(_ db: Database, pattern: FTS3Pattern) throws -> [ItemRecord] {
        try ItemRecord.fetchAll(db, sql:
        """
           SELECT DISTINCT \(ItemRecord.databaseTableName).*
           FROM \(ItemRecord.databaseTableName)
           JOIN \(StringRecord.databaseSearchTableName)
               ON \(StringRecord.databaseSearchTableName).item = \(ItemRecord.databaseTableName).id
               AND \(StringRecord.databaseSearchTableName) MATCH ?
        """, arguments: [pattern]
        )
    }
    
    override func didInsert(with rowID: Int64, for column: String?) {
        guard column == "rowId" else {
            return
        }
        rowId = rowID
    }
}

import SwiftUI
extension ItemRecord {
    func propertyBinding(name: String, defaultValue: String? = nil, db: DatabaseController = AppController.shared.databaseController) -> Binding<String?> {
        Binding {
            self.propertyValue(name, db: db)?.asString() ?? defaultValue
        } set: { (newValue) in
            try? self.setPropertyValue(name: name, value: newValue?.nilIfBlank.map { .string($0) }, db: db)
        }
    }
    
    func propertyBinding(name: String, defaultValue: String = "", db: DatabaseController = AppController.shared.databaseController) -> Binding<String> {
        Binding {
            self.propertyValue(name, db: db)?.asString() ?? defaultValue
        } set: { (newValue) in
            try? self.setPropertyValue(name: name, value: .string(newValue), db: db)
        }
    }
    
    func propertyBinding(name: String, db: DatabaseController = AppController.shared.databaseController) -> Binding<Bool?> {
        Binding {
            self.propertyValue(name, db: db)?.asBool()
        } set: { (newValue) in
            try? self.setPropertyValue(name: name, value: newValue.map { .bool($0) }, db: db)
        }
    }
    
    func propertyBinding(name: String, defaultValue: Bool = false, db: DatabaseController = AppController.shared.databaseController) -> Binding<Bool> {
        Binding {
            self.propertyValue(name, db: db)?.asBool() ?? defaultValue
        } set: { (newValue) in
            try? self.setPropertyValue(name: name, value: .bool(newValue), db: db)
        }
    }
    
    func propertyBinding(name: String, db: DatabaseController = AppController.shared.databaseController) -> Binding<Double?> {
        Binding {
            self.propertyValue(name, db: db)?.asDouble()
        } set: { (newValue) in
            try? self.setPropertyValue(name: name, value: newValue.map { .double($0) }, db: db)
        }
    }
    
    func propertyBinding(name: String, db: DatabaseController = AppController.shared.databaseController) -> Binding<Date?> {
        Binding {
            self.propertyValue(name, db: db)?.asDate()
        } set: { (newValue) in
            try? self.setPropertyValue(name: name, value: newValue.map { .datetime($0) }, db: db)
        }
    }
}

extension ItemRecord {
    static var me: ItemRecord? {
        let dbController: DatabaseController = AppController.shared.databaseController
        guard let edge = try? (dbController.read {db in
            try? EdgeRecord
                 .filter(Column(EdgeRecord.Columns.name.rawValue) == "me")
                    .fetchOne(db)
        }) else {
            return nil
        }
        
        return ItemRecord.fetchWithID(edge.target, db: dbController)
    }
    
    static func createMe() throws -> ItemRecord? {
        let myself = ItemRecord(type: "Person")
        try myself.save()
        guard let meRowId = myself.rowId else {
            return nil
        }
        
        try EdgeRecord.createFor(source: meRowId, type: "me", target: meRowId)
        
        return myself
    }
}
