//
//  IntegerRecord.swift
//  Memri
//
//  Created by Chaitanya Pandit on 15/04/21.
//

import Foundation
import GRDB

class RealRecord: BaseRecord, ItemPropertyRecord {
    static override var databaseTableName: String { "reals" }
    
    let item: Int64
    let name: String
    var value: Double
    
    func databaseValueForType(type: String, schema: Schema) -> PropertyDatabaseValue {
        let expectedType = schema.expectedType(forItemType: type, propertyName: name)
        switch expectedType {
        case .datetime:
            return .datetime(DatabaseHelperFunctions.decode(from: Int(value)))
        default:
            return .double(value)
        }
    }
    
    static let itemForeignKey = ForeignKey(["item"])
    static var item = belongsTo(ItemRecord.self, using: itemForeignKey)

    enum Columns: String, ColumnExpression {
        case item, name, value
    }
        
    internal init(item: Int64, name: String, value: Double) {
        self.item = item
        self.name = name
        self.value = value
        super.init()
    }
    
    required init(row: Row) {
        item = row[Columns.item]
        name = row[Columns.name]
        value = row[Columns.value]
        
        super.init(row: row)
    }
    
    override func encode(to container: inout PersistenceContainer) {
        container[Columns.item] = item
        container[Columns.name] = name
        container[Columns.value] = value
    }
    
    func _syncDict() -> [String: AnyEncodable] {
        let keyProperties: [String: AnyEncodable] = [
            "_item": item,
            "_name": name,
            "_value": value
        ].mapValues { AnyEncodable($0) }

        return keyProperties
    }
    
    static func propertyForItem(_ item: ItemRecord, name: String, dbController: DatabaseController = AppController.shared.databaseController) -> RealRecord? {
        guard let itemRowId = item.rowId else {
            return nil
        }
        
        return try! dbController.read { db in
            propertyForItem(itemRowId, name: name, db: db)
        }
    }
    
    static func propertyForItem(_ itemRowId: Int64, name: String, db: Database) -> RealRecord? {
        try? RealRecord
            .filter(Column(RealRecord.Columns.item.rawValue) == itemRowId &&
                         Column(RealRecord.Columns.name.rawValue) == name
            ).fetchOne(db)
    }
    
    static func propertiesForItem(_ item: ItemRecord, name: String, dbController: DatabaseController = AppController.shared.databaseController) -> [RealRecord] {
        guard let itemRowId = item.rowId else {
            return []
        }
        
        return (try? dbController.read { db in
            propertiesForItem(itemRowId, name: name, db: db)
        }) ?? []
    }
    
    static func propertiesForItem(_ itemRowId: Int64, name: String, db: Database) -> [RealRecord]? {
        try? RealRecord
            .filter(Column(RealRecord.Columns.item.rawValue) == itemRowId &&
                         Column(RealRecord.Columns.name.rawValue) == name
            ).fetchAll(db)
    }
    
    static func setPropertyValue(
        name: String,
        value: Double?,
        item: ItemRecord,
        db dbController: DatabaseController = AppController.shared.databaseController
    ) throws {
        try dbController.writeSync { (db) in
            /// Create or update the property
            try setPropertyValue(name: name, value: value, item: item, db: db)
        }
    }
    
    static func setPropertyValue(
        name: String,
        value: Double?,
        item: ItemRecord,
        db: Database
    ) throws {
        guard let itemRowId = item.rowId else {
            return
        }

        if let property: RealRecord = propertyForItem(item, name: name) {
            if let value = value {
               try  RealRecord
                    .filter(Column(RealRecord.Columns.item.rawValue) == itemRowId &&
                                Column(RealRecord.Columns.name.rawValue) == name
                    )
                .updateAll(db, [Column(RealRecord.Columns.value.rawValue).set(to: value)])
            } else {
                _ = try property.delete(db: db)
            }
        } else if let value = value{
            let property = RealRecord(item: itemRowId, name: name, value: value)
            try property.save(db: db)
        }
    }
    
    func save(db: Database) throws {
        try save(db)
    }
    
    func save(db dbController: DatabaseController = AppController.shared.databaseController) throws {
        try dbController.writeSync { (db) in
            try save(db: db)
        }
    }
    
    func delete(db: Database) throws -> Bool {
        try delete(db)
    }
    
    func delete(db dbController: DatabaseController = AppController.shared.databaseController) throws -> Bool {
        try dbController.writeSync { (db) in
            try delete(db: db)
        }
    }
}
