//
//  Schema.swift
//  MemriDatabase
//
//  Created by T Brennan on 30/11/20.
//

import Foundation
import GRDB

// MARK: SCHEMA STRUCTS

/// A schema definition. This is used to dynamically enforce supported types and properties
struct CachedProperty: Hashable {
    let itemType: String
    let propertyName: String
    
    func hash(into hasher: inout Hasher) {
        hasher.combine(itemType)
        hasher.combine(propertyName)
    }
    
    
    struct Result: Hashable {
        var value: SchemaValueType? = nil
        init(value: SchemaValueType?) {
            self.value = value
        }

        func hash(into hasher: inout Hasher) {
            hasher.combine(value)
        }
    }
}

struct CachedEdge: Hashable {
    let itemType: String
    let edgeName: String
    
    func hash(into hasher: inout Hasher) {
        hasher.combine(itemType)
        hasher.combine(edgeName)
    }
    
    struct Result: Hashable {
        var value: String? = nil
        init(value: String?) {
            self.value = value
        }
        
        func hash(into hasher: inout Hasher) {
            hasher.combine(value)
        }
    }
}


struct Schema {
    static var cachedProperties : [CachedProperty : CachedProperty.Result] = [:]
    static var cachedEdges:[CachedEdge : CachedEdge.Result] = [:]
    static var shared = Schema()

    /// Supported types in the schema
    func expectedType(forItemType itemType: String, propertyOrEdgeName: String) -> ResolvedType? {
        if let propertyType = expectedType(forItemType: itemType, propertyName: propertyOrEdgeName) {
            return .property(propertyType)
        } else if let edgeType = expectedTargetType(forItemType: itemType, edgeName: propertyOrEdgeName) {
            return .edge(targetType: edgeType)
        } else {
            return nil
        }
    }
    
    func propertyNamesForItemType(itemType: String, db: Database) -> [String] {
        var propertyNames: [String] = []
        let sortAlias = TableAlias()
        var request = ItemRecord
            .filter(Column(ItemRecord.Columns.type.rawValue) == "ItemPropertySchema")
        request = request.joining(required:
                                    ItemRecord.stringProperties.aliased(sortAlias)
                                    .filter((StringRecord.Columns.name == "itemType" && StringRecord.Columns.value == itemType)))

        if let itemRecords = try? request.fetchAll(db) {
            for itemRecord in itemRecords {
                if let itemRowId = itemRecord.rowId,
                   let properties = try? StringRecord.filter(Column(StringRecord.Columns.item.rawValue) == itemRowId).fetchAll(db) {
                    let properties = properties.filter({$0.name == "propertyName"})
                    propertyNames = properties.map{$0.value}
                }
            }
        }
        
        return propertyNames
    }
    
    func propertyNamesForItemType(itemType: String, controller: DatabaseController = AppController.shared.databaseController) -> [String] {
        return (try? controller.read { db in
            propertyNamesForItemType(itemType: itemType, db: db)
        }) ?? []
    }
    
    func expectedType(forItemType itemType: String, propertyName: String, db: Database) -> SchemaValueType? {
        let key = CachedProperty(itemType: itemType, propertyName: propertyName)
        let cachedValue = Schema.cachedProperties[key]
        if let cachedValue = cachedValue {
            return cachedValue.value
        }
        
        var schemaValueType: SchemaValueType? = nil
        
        let sortAlias = TableAlias()
        var request = ItemRecord
            .filter(Column(ItemRecord.Columns.type.rawValue) == "ItemPropertySchema")
        request = request.joining(required:
                                    ItemRecord.stringProperties.aliased(sortAlias)
                                    .filter((StringRecord.Columns.name == "itemType" && StringRecord.Columns.value == itemType)))
        if let itemRecords = try? request.fetchAll(db) {
            for itemRecord in itemRecords {
                if let itemRowId = itemRecord.rowId,
                   let properties = try? StringRecord.filter(Column(StringRecord.Columns.item.rawValue) == itemRowId).fetchAll(db) {
                    if let _ = properties.filter({$0.name == "propertyName" && $0.value == propertyName}).first,
                       let propertyValue = properties.filter({$0.name == "valueType"}).first {
                        schemaValueType = SchemaValueType(rawValue: propertyValue.value)
                        break
                    }
                }
            }
        }
        
        Schema.cachedProperties[key] = CachedProperty.Result(value: schemaValueType)
        
        return schemaValueType
    }
    
    func expectedType(forItemType itemType: String, propertyName: String, controller: DatabaseController = AppController.shared.databaseController) -> SchemaValueType? {
        return try? controller.read { db in
            expectedType(forItemType: itemType, propertyName: propertyName, db: db)
        }
    }
    
    func expectedTargetType(forItemType itemType: String, edgeName: String, db: Database) -> String? {
        let key = CachedEdge(itemType: itemType, edgeName: edgeName)
        let cachedValue = Schema.cachedEdges[key]
        if let cachedValue = cachedValue {
            return cachedValue.value
        }

        var targetType: String? = nil
        
        let sortAlias = TableAlias()
        var request = ItemRecord
            .filter(Column(ItemRecord.Columns.type.rawValue) == "ItemEdgeSchema")
        request = request.joining(required:
                                    ItemRecord.stringProperties.aliased(sortAlias)
                                    .filter((StringRecord.Columns.name == "sourceType" && StringRecord.Columns.value == itemType)))
        if let itemRecords = try? request.fetchAll(db) {
            for itemRecord in itemRecords {
                if let itemRowId = itemRecord.rowId,
                   let edges = try? StringRecord.filter(Column(StringRecord.Columns.item.rawValue) == itemRowId).fetchAll(db) {
                    if let _ = edges.filter({$0.name == "edgeName" && $0.value == edgeName}).first,
                       let edgeValue = edges.filter({$0.name == "targetType"}).first {
                        targetType = edgeValue.value
                        break
                    }
                }
            }
        }
        
        Schema.cachedEdges[key] = CachedEdge.Result(value: targetType)
        
        return targetType
    }
    
    func expectedTargetType(forItemType itemType: String, edgeName: String, controller: DatabaseController = AppController.shared.databaseController) -> String? {
        return try? controller.read { db in
            expectedTargetType(forItemType: itemType, edgeName: edgeName, db: db)
        }
    }
    
    enum ResolvedType {
        case property(SchemaValueType)
        case edge(targetType: String)
    }
}

/// A schema type definition. This defines the supported properties and their value type
struct SchemaType {
    var type: String
    /// Supported properties for this type in the schema. Value is the value type for the property - this must be a primitive (string, double, dateTime)
    var propertyTypes: [String: SchemaProperty]
    /// Supported properties for this type in the schema. Target type must be another type (eg. Photo, File, etc)
    var edgeTypes: [String: SchemaEdge]
}

/// A schema property definition.
struct SchemaProperty: Codable, Hashable {
    var itemType: String
    var property: String
    var valueType: SchemaValueType
    
    enum CodingKeys: String, CodingKey {
        case itemType = "item_type"
        case property = "property"
        case valueType = "value_type"
    }
}

enum SchemaValueType: String, Codable, Hashable {
    case string
    case bool
    case int
    case double
    case datetime
    case blob
}

/// A schema edge definition.
struct SchemaEdge: Codable {
    var sourceType: String
    var edge: String
    var targetType: String
    var allowsMultiple: Bool
    
    enum CodingKeys: String, CodingKey {
        case sourceType = "source_type"
        case edge = "edge"
        case targetType = "target_type"
    }
    
    init(from decoder: Decoder) throws {
        let container = try decoder.container(keyedBy: CodingKeys.self)
        sourceType = try container.decode(String.self, forKey: .sourceType)
        edge = try container.decode(String.self, forKey: .edge)
        
        let rawTargetType = try container.decode(String.self, forKey: .targetType)
        if rawTargetType.hasSuffix("[]") {
            targetType = String(rawTargetType.dropLast(2))
            allowsMultiple = true
        } else {
            targetType = rawTargetType
            allowsMultiple = false
        }
    }
    
    func encode(to encoder: Encoder) throws {
        var container = encoder.container(keyedBy: CodingKeys.self)
        try container.encode(sourceType, forKey: .sourceType)
        try container.encode(edge, forKey: .edge)
        try container.encode("\(targetType)\(allowsMultiple ? "[]" : "")", forKey: .targetType)
    }
}





