//
//  PermissionsController.swift
//  Memri

import AVFoundation
import Contacts
import Combine
import CoreLocation
import Foundation
import Photos
import SwiftUI


class PermissionsController: NSObject, ObservableObject {
    lazy var locationManager: CLLocationManager = {
        let locationManager = CLLocationManager()
        locationManager.desiredAccuracy = .greatestFiniteMagnitude
        locationManager.delegate = self
        return locationManager
    }()
    static let shared = PermissionsController()
    let contactAuthorization: PassthroughSubject<Bool, Never> = PassthroughSubject()
    var authorizationCompletionBlock:((Bool)->())? = {_ in}
    var locationPermission: Bool = CLLocationManager().authorizationStatus == .authorizedAlways || CLLocationManager().authorizationStatus == .authorizedWhenInUse
    

    var libraryPermission: Bool = PHPhotoLibrary.authorizationStatus() == .authorized

    var cameraPermission: Bool = AVCaptureDevice.authorizationStatus(for: AVMediaType.video) == AVAuthorizationStatus.authorized

    var microPhonePermission: Bool = AVAudioSession.sharedInstance().recordPermission == .granted

    var contactsPermission: Bool = CNContactStore.authorizationStatus(for: .contacts) == .authorized || CNContactStore.authorizationStatus(for: .contacts) == .restricted
    
    override init() {
        super.init()
        contactAuthorization.send(CNContactStore.authorizationStatus(for: .contacts) == .authorized || CNContactStore.authorizationStatus(for: .contacts) == .restricted )
  
    }
    
    func requestLocation(completion: ((Bool) -> ())? = nil) {
        if self.locationManager.authorizationStatus == .notDetermined {
            authorizationCompletionBlock = { isGranted in
                guard self.locationManager.authorizationStatus != .notDetermined else {
                    return
                }
                completion?(isGranted)
            }
            self.locationManager.requestAlwaysAuthorization()
        } else if self.locationManager.authorizationStatus != .authorizedAlways, self.locationManager.authorizationStatus != .authorizedWhenInUse {
            self.navigateToSystemSettings()
            completion?(false)
        } else {
            completion?(false)
        }
    }
    
    func requestPhotos(_ completion: ((Bool) -> ())? = nil) {
        if PHPhotoLibrary.authorizationStatus() == .notDetermined {
            PHPhotoLibrary.requestAuthorization(for: .readWrite) { status in
                completion?(status == .authorized)
            }
        } else if PHPhotoLibrary.authorizationStatus() != .authorized,
                  PHPhotoLibrary.authorizationStatus() != .limited
        {
            self.navigateToSystemSettings()
            completion?(false)
        } else {
            completion?(false)
        }
    }
    
    func requestCamera(_ completion: ((Bool) -> ())? = nil) {
        if AVCaptureDevice.authorizationStatus(for: AVMediaType.video) == AVAuthorizationStatus.notDetermined {
            AVCaptureDevice.requestAccess(for: AVMediaType.video, completionHandler: { status in
                completion?(status)
            })
        } else if AVCaptureDevice.authorizationStatus(for: AVMediaType.video) != AVAuthorizationStatus.authorized {
            self.navigateToSystemSettings()
            completion?(false)
        } else {
            completion?(false)
        }
    }
    
    func requestMicrophone(_ completion: ((Bool) -> ())? = nil) {
        if AVAudioSession.sharedInstance().recordPermission == .undetermined {
            AVAudioSession.sharedInstance().requestRecordPermission { status in
                completion?(status)
            }
        } else if AVAudioSession.sharedInstance().recordPermission != .granted {
            self.navigateToSystemSettings()
            completion?(false)
        } else {
            completion?(false)
        }
    }
    
    func requestContacts(_ completion: ((Bool) -> ())? = nil) {
        if CNContactStore.authorizationStatus(for: .contacts) == .notDetermined {
            CNContactStore().requestAccess(for: .contacts) { state, error in
                guard error == nil else {
                    self.contactAuthorization.send(false)
                    completion?(false)
                    return
                }
                self.contactAuthorization.send(state)
                completion?(state)
            }
        } else if CNContactStore.authorizationStatus(for: .contacts) != .authorized && CNContactStore.authorizationStatus(for: .contacts) != .restricted {
            self.navigateToSystemSettings()
            completion?(false)
        } else {
            completion?(false)
        }
    }
    
    func navigateToSystemSettings() {
        guard let settingsUrl = URL(string: UIApplication.openSettingsURLString) else {
            return
        }
        if UIApplication.shared.canOpenURL(settingsUrl) {
            UIApplication.shared.open(settingsUrl, completionHandler: { _ in })
        }
    }
}

extension PermissionsController: CLLocationManagerDelegate {
    func locationManager(_ manager: CLLocationManager, didChangeAuthorization status: CLAuthorizationStatus) {
        self.locationPermission = status == .authorizedAlways || status == .authorizedWhenInUse
        self.authorizationCompletionBlock?(status == .authorizedAlways || status == .authorizedWhenInUse)
    }

}
