//
//  PubSubController.swift
//  Memri

import Foundation
import Combine
import Alamofire

struct ItemSubscription {
    let item: ItemRecord
    let property: String
    var retryCount: Int = 0
    let desiredValue: PropertyDatabaseValue?
    let request: DataRequest? = nil
    let completion: ((PropertyDatabaseValue?, Error?) -> Void)
}

extension ItemSubscription: Hashable {
    static func == (lhs: ItemSubscription, rhs: ItemSubscription) -> Bool {
        let lhsHash = "\(lhs.item.id)/\(lhs.property)"
        let rhsHash = "\(rhs.item.id)/\(rhs.property)"
        
        return lhsHash == rhsHash
    }
    
    func hash(into hasher: inout Hasher) {
        hasher.combine("\(item.id)/\(property)")
    }
}


class PubSubController {
    static let pollInterval: TimeInterval = 3
    static let maxRetryAttempts: Int = 10

    private var subscribers: Set<ItemSubscription> = Set()
    private var cancellables: Set<AnyCancellable> = Set()
    private var subscriptions: Set<AnyCancellable> = []

    public func startObservingItemProperty(_ item: ItemRecord, property: String, desiredValue: PropertyDatabaseValue?, completion: @escaping ((PropertyDatabaseValue?, Error?) -> Void)) {
        stopObservingItemProperty(item, property: property)
        let subscription = ItemSubscription(item: item, property: property, desiredValue: desiredValue, completion: completion)
        subscribers.insert(subscription)
        startObserver(subscription)
    }
    
    public func stopObservingItemProperty(_ item: ItemRecord, property: String) {
        guard let subscription = subscriptionForItem(item, property: property) else {
            return
        }
        
        cancelSubscription(subscription, error: nil)
    }
    
    private func subscriptionForItem(_ item: ItemRecord, property: String) -> ItemSubscription? {
        return subscribers.filter{$0.item.id == item.id}.first
    }
  
    private func cancelSubscription(_ subscription: ItemSubscription, error: Error?) {
        subscription.completion(nil, error ?? StringError(description: "Cancelled"))
        subscribers.remove(subscription)
    }
    
    private func checkSubscriptionFulfilled(_ dicts: [[String: AnyDecodable]], subscription: ItemSubscription) {
        
        // Dict could be nil if item is still not synced
        guard let dict = dicts.first, dict.count > 0 else {
            reschedule(subscription)
            return
        }
        
        guard let newValue = propertyValueFromDict(subscription.property, type: subscription.item.type, dict: dict) else {
            cancelSubscription(subscription, error: StringError(description: "Property not found"))
            return
        }
        
        guard let databaseValue = ItemRecord.fetchWithID(subscription.item.id)?.propertyValue(subscription.property) else {
            cancelSubscription(subscription, error: StringError(description: "Database value not found"))
            return
        }
        
        // If we have expected value to look for, check if new value is same as that of expected value
        if let expectedValue = subscription.desiredValue {
            let containsValue = expectedValue == newValue
            guard !containsValue else {
                subscription.completion(newValue, nil)
                subscribers.remove(subscription)

                return
            }
        }
        
        // else check if the new value is different from db, in which case we just call the completion block but DO NOT stop observing
        // The part where we setup observer is responsible to cancel observation in this case
        if databaseValue != newValue {
            // Update item in db so that next time we will detect change
            if let dbItem = ItemRecord.fetchWithID(subscription.item.id) {
                try? dbItem.setPropertyValue(name: subscription.property, value: newValue)
            }
            subscription.completion(newValue, nil)
            reschedule(subscription)
            return
        }
        
        guard subscription.retryCount < PubSubController.maxRetryAttempts else {
            cancelSubscription(subscription, error: StringError(description: "Exceeded retry attempts"))
            return
        }
        
        var newSubscription = subscription
        newSubscription.retryCount = newSubscription.retryCount + 1
        subscribers.remove(subscription)
        subscribers.insert(newSubscription)
        reschedule(newSubscription)
    }
    
    private func reschedule(_ subscription: ItemSubscription) {
        guard subscribers.contains(subscription) else {
            return
        }
        
        DispatchQueue.main.asyncAfter(deadline: .now() + PubSubController.pollInterval) {[weak self] in
            self?.startObserver(subscription)
        }
    }
    
    private func startObserver(_ subscription: ItemSubscription) {
        guard subscription.request == nil else {
            print("Ignoring duplicate request for item: ", subscription.item)
            return
        }
        
        guard let connectionConfig = AppController.shared.podConnectionConfig else {
            cancelSubscription(subscription, error: StringError(description: "No pod connection config"))
            return
        }

        do {
            let request = PodAPIStandardRequest.searchAction(payload: ["id": AnyEncodable(stringLiteral: subscription.item.id.uid)])
            let networkCall = try request.execute(connectionConfig: connectionConfig)
            networkCall.sink {[weak self] (result) in
                guard result.error == nil,
                      let responseData = result.data,
                      let responseObjects = try? JSONSerialization.jsonObject(with: responseData) as? [[String: Any]] else {
                    self?.cancelSubscription(subscription, error: result.error ?? StringError(description: "Error fetching item"))
                    return
                }

                let itemDicts = responseObjects.compactMap{object -> [String : AnyDecodable]? in
                    return object.compactMapValues{AnyDecodable($0)}
                }
                                
                self?.checkSubscriptionFulfilled(itemDicts, subscription: subscription)
            } .store(in: &subscriptions)
        } catch let error {
            cancelSubscription(subscription, error: error)
        }
    }
    
    private func propertyValueFromDict(_ property: String, type: String, dict: [String: AnyDecodable]) -> PropertyDatabaseValue? {
        
        guard let decodableValue = dict[property] else {
            return nil
        }
        
        let schema = AppController.shared.databaseController.schema
        guard !ItemRecord.intrinsicProperties.contains(property),
              let expectedType = schema.expectedType(forItemType: type, propertyName: property) else {
            return nil
        }
        
        return try? PropertyDatabaseValue(value: decodableValue.value, propertyType: expectedType)
    }
}
