//
//  SceneController.swift
//  MemriDatabase
//
//  Created by T Brennan on 14/12/20.
//

import Foundation
import SwiftUI
import Combine
import GRDB

/// The scene controller is specific to a particular `window` of the app. On the iPhone there is usually only one. There may be multiple eg. if multitasking on iPad or multiple windows on mac
class SceneController: ObservableObject {
    let appController: AppController = .shared
    private(set) var topMostContext: ViewContextController?
    let navigationController = MemriUINavigationController()
    var appStateObservation: AnyCancellable?

    init() {
        setupObservations()
        navigationController.viewControllers = [UIHostingController(rootView: Text("Welcome to Memri"))]
        appStateObservation = appController.statePublisher.sink {[weak self] state in
            self?.setupObservations()
            
            guard state == .authenticated else {
                return
            }
            
            self?.restoreUIState()
        }
    }
    
    func restoreUIState() {
        let oldState = try? appController.databaseController.read {
            try NavigationStack.fetchOne($0)
        }
        if let navStack = oldState, let topView = navStack.state.last {
            navigationStack = navStack
            let context = makeContext(forConfig: topView)
            topMostContext = context
            navigationController.viewControllers = [UIHostingController(rootView: SceneContentView(context: context))]
        }
    }
    
    func toggleEditMode() {
        guard let topConfigHolder = topMostContext?.configHolder else {
            return
        }
        
        var viewArgs: CVUViewArguments? = topConfigHolder.config.viewArguments
        if (viewArgs == nil) {
            viewArgs = CVUViewArguments()
        }
        
        isInEditMode.toggle()

        if (!isInEditMode) {
            // Clear selection when ending edit mode
            topMostContext?.selectedItems = []
            try? AppController.shared.syncController.sync()
        }
        var currentArgs = viewArgs?.args ?? [:]
        currentArgs["readOnly"] = .constant(.bool(!isInEditMode))
        let newArgs = CVUViewArguments(args: currentArgs, argumentItem: viewArgs?.argumentItem, parentArguments: viewArgs?.parentArguments)
        topConfigHolder.config.viewArguments = newArgs
    }
    
    @Published
    var isContentFullscreen: Bool = false
    
    
    @Published
    var isInEditMode: Bool = false
    
    @Published
    var navigationIsVisible: Bool = false
    var navigationFilterText: String? {
        get { navigationQuery.searchString }
        set { navigationQuery.searchString = newValue?.nilIfBlank }
    }
    
    
    @Published
    var filterPanelIsVisible: Bool = false
    
    @Published
    var contextPaneVisible: Bool = false
    
    private var navigationQuery = DatabaseQueryConfig(itemTypes: ["NavigationItem"], sortProperty: nil) {
        didSet {
            if navigationQuery != oldValue {
                setupObservations()
            }
        }
    }
    
    var navigationItems: [NavigationElement] {
        navigationItemRecords.sorted(by: { (a, b) -> Bool in
            guard let aRowId = a.rowId, let bRowId = b.rowId else {return false}
            return aRowId < bRowId
        }).compactMap { (item) -> NavigationElement? in
           switch item.propertyValue("itemType")?.asString() {
           case "heading":
               guard let title = item.propertyValue("title")?.asString() else {
                   return nil
               }
               return .heading(title)
           case "line":
               return .line
           default:
               guard let title = item.propertyValue("title")?.asString(),
                     let targetViewName = item.propertyValue("sessionName")?.asString() else {
                   return nil
               }
               return .item(.init(name: title, targetViewName: targetViewName))
           }
       }
   }

    
    @Published
    private var navigationItemRecords: [ItemRecord] = []
    private var queryObservation: AnyCancellable?
    /// Sets up a database observation
    private func setupObservations() {
        /// Remove old observation
        queryObservation?.cancel()
        queryObservation = nil

        
        /// Note that the request must remain constant within the observation, hence constructed outside of the tracking (if changed, start a new observation)
        let config = navigationQuery
        let observation = ValueObservation
            .tracking { db in
                try config.executeRequest(db: db)
            }
        
        guard let pool = appController.databaseController.databasePool else {
            print("Error: Cannot setup observers, database pool not available")
            return
        }
        
        /// Subscribe to changes in the value, with the first result being reported immediately
        queryObservation = observation.publisher(
            in: pool,
            scheduling: .immediate)
        .sink(
            receiveCompletion: { completion in },
            receiveValue: { result in
                self.navigationItemRecords = result
            })
    }
    
    private var navigationStack: NavigationStack = NavigationStack() {
            willSet {
                if newValue != navigationStack {
                    objectWillChange.send()
                }
            }
        didSet {
            if navigationStack != oldValue {
                try? appController.databaseController.writeSync {
                    try navigationStack.save($0)
                }
            }
        }
    }
    
    func makeContext(forConfig config: ViewContextHolder) -> ViewContextController {
        ViewContextController(config: config, databaseController: appController.databaseController, cvuController: appController.cvuController)
    }
    
    var canNavigateBack: Bool {
        navigationStack.state.count > 1
    }
    
    func navigateBack() {
        guard navigationStack.state.count > 1, let newTopConfig = navigationStack.state[safe: navigationStack.state.endIndex - 2]
        else { return }
        _ = navigationStack.state.popLast()
        
        let context = makeContext(forConfig: newTopConfig)
        topMostContext = context
        
        let vc = UIHostingController(rootView: SceneContentView(context: context))
        navigationController.viewControllers.insert(vc, at: navigationController.viewControllers.endIndex - 1)
        navigationController.popToViewController(vc, animated: true)
    }
    
    func navigateToNewContext(defaultDefinition: CVUDefinitionContent? = nil,
                              defaultContext: CVUContext? = nil,
                              clearStack: Bool = false,
                              animated: Bool = true,
                              viewName: String? = nil,
                              inheritDatasource: Bool = false,
                              overrideRenderer: String? = nil,
                              defaultRenderer: String = "list",
                              targetItem: ItemRecord? = nil,
                              overrideUIDs: Set<Int64>? = nil,
                              dateRange: DateInterval? = nil,
                              customDefinition: CVUDefinitionContent? = nil,
                              viewArguments: CVUViewArguments? = nil) {
                
        let viewDefinition = defaultDefinition ?? appController.cvuController.viewDefinitionFor(viewName: viewName ?? "", customDefinition: customDefinition) ?? CVUDefinitionContent()
        let cvuContext = CVUContext(currentItem: targetItem, selector: nil, viewName: viewName, viewDefinition: viewDefinition, viewArguments: viewArguments)

        let rendererName = overrideRenderer ?? { () -> String in
            return viewDefinition.propertyResolver(context: cvuContext, lookup: CVULookupController(), db: appController.databaseController).string("defaultRenderer") ?? defaultRenderer
        }()
        
        let datasource = viewDefinition.definitions.first(where: { $0.type == .datasource })
        let queryConfig = DatabaseQueryConfig.queryConfigWith(context: cvuContext,
                                                              datasource: datasource,
                                                              inheritQuery: inheritDatasource ? topMostContext?.config.query : nil,
                                                              overrideUIDs: overrideUIDs,
                                                              targetItem: targetItem,
                                                              dateRange: dateRange)
        
        let config = ViewContext(viewName: viewName, rendererName: rendererName, viewDefinition: viewDefinition, query: queryConfig, viewArguments: viewArguments, focusedItem: targetItem)
        let holder = ViewContextHolder(config)
        
        let newViewContextController = makeContext(forConfig: holder)
        
        topMostContext = newViewContextController
        if clearStack {
            navigationStack.state = [holder]
        } else {
            navigationStack.state.append(holder)
        }
        navigationController.setViewControllers([UIHostingController(rootView: SceneContentView(context: newViewContextController))], animated: animated)
    }
    
    var closeStack =
        [Binding<PresentationMode>]() // A stack of bindings for the display state of presented popups
    public func addToStack(_ isPresentedBinding: Binding<PresentationMode>) {
        closeStack.append(isPresentedBinding)
    }

    public func closeLastInStack() {
        if let lastVisibleIndex = closeStack.lastIndex(where: { $0.wrappedValue.isPresented }) {
            closeStack[lastVisibleIndex].wrappedValue.dismiss()
            closeStack = Array(closeStack.prefix(upTo: lastVisibleIndex))
        }
    }
    
    func scheduleUIUpdate(updateWithAnimation: Bool = false) {
        guard let _ = topMostContext else {
            return
        }
        
        if updateWithAnimation {
            DispatchQueue.main.async {
                withAnimation {
                    self.topMostContext?.update()
                    self.objectWillChange.send()
                }
            }
        } else {
            self.topMostContext?.update()
            objectWillChange.send()
        }
    }
}
