//
//  Settings.swift
//  Memri
//
//  Created by Chaitanya Pandit on 28/06/21.
//

import Foundation
import UIKit
import GRDB
import SwiftUI

public class Settings {
    static let shared = Settings()

    private var listeners = [String: [UUID]]()
    private var callbacks = [UUID: (Any?) -> Void]()

    private var deviceID: String {
        return UIDevice.current.identifierForVendor!.uuidString
    }
    
    public func getBinding<T: Decodable>(_ path: String) -> Binding<T> {
        Binding<T>(
            get: { () -> T in
                // TODO: Error handling
                if let x: T = self.get(path) {
                    return x
                }
                else {
                    print("Could not get setting \(path)")

                    if T.self == String.self { return "" as! T }
                    if T.self == Double.self { return 0 as! T }
                    if T.self == Int.self { return 0 as! T }
                    if T.self == Bool.self { return false as! T }
                }

                return 0 as! T // Should never get here
            },
            set: {
                self.set(path, AnyCodable($0))
            }
        )
    }
    
    public func get<T: Decodable>(_ path: String, type _: T.Type = T.self) -> T? {
        do {
            for path in try getSearchPaths(path) {
                if let value: T = try getSetting(path) {
                    return value
                }
            }
        }
        catch {
            print("Could not fetch setting '\(path)': \(error)")
        }

        return nil
    }
    
    public func set(_ path: String, _ value: Any) {
        do {
            let searchPaths = try getSearchPaths(path)
            if searchPaths.count == 1 {
                throw StringError(description: "Missing scope 'user' or 'device' as the start of the path")
            }
            try setSetting(searchPaths[0], value as? AnyCodable ?? AnyCodable(value))
            fire(searchPaths[0], (value as? AnyCodable)?.value ?? value)
        }
        catch {
            print(error)
        }
    }
    
    private func getSearchPaths(_ path: String) throws -> [String] {
        let p = path.first == "/" ? String(path.suffix(path.count - 1)) : path
        let splits = p.split(separator: "/")
        let type = splits.first
        let query = splits.dropFirst().joined(separator: "/")

        if type == "device" {
            return ["\(deviceID)/\(query)", "defaults/\(query)"]
        }
        else if type == "user" {
            return ["user/\(query)", "defaults/\(query)"]
        }
        else {
            return ["defaults/\(query)"]
        }
    }
    
    public func getSetting<T: Decodable>(_ path: String, type: T.Type = T.self) throws -> T? {
        guard let settingItem = try getSettingItem(path),
              case let .blob(jsonData) = settingItem.propertyValue("json"),
              let json = try? JSONSerialization.jsonObject(with: jsonData, options: []) as? [String: T] else {
            return nil
        }
        
        return json["value"]
    }
    
    public func setSetting(_ path: String, _ value: AnyCodable) throws {
        var item = try getSettingItem(path)
        if (item == nil) {
            item = ItemRecord(type: "Setting")
            try item?.save()
            try item?.setPropertyValue(name: "keystr", value: .string(path))
        }
        
        guard let settingItem = item else {
            throw StringError(description: "Error saving item")
        }
        
        let jsonData = try JSONSerialization.data(withJSONObject: ["value": value.value], options: [])
        try settingItem.setPropertyValue(name: "json", value: .blob(jsonData))
    }
    
    private func getSettingItem(_ path: String) throws -> ItemRecord?  {
        guard let settingItems = try (AppController.shared.databaseController.read {db in
            try? ItemRecord
                .filter(Column(ItemRecord.Columns.type.rawValue) == "Setting")
                .fetchAll(db)
        }) else {
            return nil
        }
        
        return settingItems.filter({ $0.propertyValue("keystr") == .string(path)}).first
    }
    
    private func fire(_ path: String, _ value: Any?) {
        if let list = listeners[path] {
            for id in list {
                if let f = callbacks[id] {
                    f(value)
                }
            }
        }
    }
}

