//
//  SyncController.swift
//  MemriDatabase
//
//  Created by T Brennan on 1/12/20.
//

import Foundation
import GRDB
import Combine

enum SyncControllerState: String, Codable, Hashable, DatabaseValueConvertible {
    case idle
    case started
    case downloadedItems
    case uploadedSchemaProperties
    case uploadedSchemaEdges
    case uploadedItems
    case uploadedEdges
    case uploadedFiles
    case downloadedFiles
    case done
    case failed
}

class SyncController {
    var syncing = false
    
    var lastError: Error? = nil
    var completion: ((Error?) -> Void)? = nil
    let databaseController: DatabaseController
    public private(set) var subscriptions: Set<AnyCancellable> = []
    var state: SyncControllerState {
        didSet {
            switch state {
            case .started:
                syncing = true
                try? downloadItems()
            case .downloadedItems:
                try? uploadSchemaProperties()
            case .uploadedSchemaProperties:
                try? uploadSchemaEdges()
            case .uploadedSchemaEdges:
                try? uploadItems()
            case .uploadedItems:
                try? uploadEdges()
            case .uploadedEdges:
                try? uploadFiles()
            case .uploadedFiles:
                try? downloadFiles()
            case .downloadedFiles:
                finishSync()
            case .failed:
                finishSync()
            default:
                break
            }
        }
    }
    
    init(databaseController: DatabaseController) {
        self.databaseController = databaseController
        self.state = .idle
    }
    
    func sync(completion: ((Error?) -> Void)? = nil) throws {
        guard (try? ItemRecord.getOwnerAndDBKey()) != nil else {
            // Local mode or not setup
            throw StringError(description: "Not Setup")
        }
        
        guard !syncing else {
            throw StringError(description: "Already syncing")
        }
        
        self.completion = completion
        self.state = .started
    }
    
    private func downloadItems() throws {
        let dateServerModified = (try? ItemRecord.lastSyncedItem())?.dateServerModified ?? Date.distantPast
        let timestamp = DatabaseHelperFunctions.encode(dateServerModified)
                
        try searchAction(payload: timestamp > 0 ? ["dateServerModified>=": AnyEncodable.init(integerLiteral: timestamp)] : [:], completion: { (data, error) in
            guard error == nil,
                let responseData = data,
                let responseObjects = try? JSONSerialization.jsonObject(with: responseData) as? [[String: Any]] else {
                self.lastError = error
                self.state = .failed
                return
            }

            let responseItems = responseObjects.filter{$0["_source"] == nil || $0["_target"] == nil || $0["_edge"] == nil}
            let responseEdges = responseObjects.filter{$0["_source"] != nil && $0["_target"] != nil && $0["_edge"] != nil}

            try? AppController.shared.databaseController.writeSync {(db) in
                let schema = AppController.shared.databaseController.schema
                // We need to insert items first since edges need items to exist in db
               _ = responseItems.compactMap{object -> [String : AnyDecodable]? in
                   return object.compactMapValues{AnyDecodable($0)}
               }.compactMap { object -> ItemRecord? in
                   return try? ItemRecord.fromSyncItemDict(fromSyncItemDict: object, schema: schema, db: db)
               }
                               
                // Edges should be inserted only after all items are created
                // because currently there are scenarios where edges appear before items in server response
               _ = responseEdges.compactMap{object -> [String : AnyDecodable]? in
                   return object.compactMapValues{AnyDecodable($0)}
               }.compactMap { object -> ItemRecord? in
                   return try? ItemRecord.fromSyncItemDict(fromSyncItemDict: object, schema: schema, db: db)
               }
            }

            self.state = .downloadedItems
        })
    }
    
    func uploadSchemaProperties() throws {
        let syncPayload = try makeSyncSchemaPropertiesData()
        
        guard syncPayload.createItems.count > 0
            || syncPayload.updateItems.count > 0
            || syncPayload.deleteItems.count > 0 else {
            self.state = .uploadedSchemaProperties
            return
        }

        try bulkAction(bulkPayload: syncPayload, completion: { error in
            try? ItemRecord.didSyncItems(syncItems: syncPayload, error: error)

            guard error == nil else {
                self.lastError = error
                self.state = .failed
                return
            }
            
            self.state = .uploadedSchemaProperties
        })
    }
    
    func uploadSchemaEdges() throws {
        let syncPayload = try makeSyncSchemaEdgesData()
        
        guard syncPayload.createItems.count > 0
            || syncPayload.updateItems.count > 0
            || syncPayload.deleteItems.count > 0 else {
            self.state = .uploadedSchemaEdges
            return
        }

        try bulkAction(bulkPayload: syncPayload, completion: { error in
            try? ItemRecord.didSyncItems(syncItems: syncPayload, error: error)

            guard error == nil else {
                self.lastError = error
                self.state = .failed
                return
            }
            
            self.state = .uploadedSchemaEdges
        })
    }
    
    private func uploadItems() throws {
        let syncPayload = try makeSyncUploadData()
        
        guard syncPayload.createItems.count > 0
            || syncPayload.updateItems.count > 0
            || syncPayload.deleteItems.count > 0 else {
            self.state = .uploadedItems
            return
        }
            
        try bulkAction(bulkPayload: syncPayload, completion: {[weak self] error in
            try? ItemRecord.didSyncItems(syncItems: syncPayload, error: error)

            guard error == nil else {
                self?.lastError = error
                self?.state = .failed
                return
            }
            
            // Recurse until we run out of items to sync
            try? self?.uploadItems()
        })
    }
    
    private func uploadEdges(maxItems: Int = 100) throws {
        let syncPayload = try makeSyncEdgesData(maxItems: maxItems)
        
        guard syncPayload.createEdges.count > 0 else {
            self.state = .uploadedEdges
            return
        }
            
        try bulkAction(bulkPayload: syncPayload, completion: {[weak self] error in
            try? EdgeRecord.didSyncEdges(syncItems: syncPayload, error: error)

            guard error == nil else {
                self?.lastError = error
                self?.state = .failed
                return
            }
            
            // Recurse until we run out of edges to sync
            try? self?.uploadEdges(maxItems: maxItems)
        })
    }
    
    private func uploadFiles() throws {
        guard let (item: item, fileName: fileName) = try ItemRecord.fileItemRecordToUpload() else {
            self.state = .uploadedFiles
            return
        }
        
        print("Uploading File: ", fileName)

        try uploadFile(uuid: fileName) {[weak self] data, error in
            guard error == nil else {
                self?.lastError = error
                self?.state = .failed
                return
            }
            
            try? ItemRecord.didUploadFileForItem(item)
            try? self?.uploadFiles()
        }
    }
    
    private func downloadFiles() throws {
        guard let (item: item, sha256: sha256, fileName: fileName) = try ItemRecord.fileItemRecordToDownload() else {
            self.state = .downloadedFiles
            return
        }
        
        print("Downloading File: ", fileName)

        try downloadFile(sha256: sha256, fileName: fileName, completion: { [weak self] error in
            guard error == nil else {
                self?.lastError = error
                self?.state = .failed
                return
            }
            
            try? ItemRecord.didDownloadFileForItem(item)
            try? self?.downloadFiles()
        })
    }
    
    private func finishSync() {
        self.state = .done
        self.syncing = false
        self.completion?(self.lastError)
        
        self.state = .idle
        self.completion = nil
        print("Sync Complete!")
    }
    
    func makeSyncSchemaPropertiesData() throws -> PodAPIPayload.BulkAction {
        guard let pool = databaseController.databasePool else {
            throw StringError(description: "Database pool not available")
        }
        
        return try pool.read { (db) in
            let schemaPropertyItems = try ItemRecord
               .filter(Column(ItemRecord.Columns.type.rawValue) == "ItemPropertySchema"
                        && Column(ItemRecord.Columns.syncState.rawValue) == SyncState.create)
                    .fetchAll(db)
            let syncPropertyItems = schemaPropertyItems.compactMap{$0.schemaPropertyDict(db: db)}
                        
            let bulkAction = PodAPIPayload.BulkAction(
                createItems: syncPropertyItems,
                updateItems: [],
                deleteItems: [],
                createEdges: [])
            return bulkAction
        }
    }
    
    func makeSyncSchemaEdgesData() throws -> PodAPIPayload.BulkAction {
        guard let pool = databaseController.databasePool else {
            throw StringError(description: "Database pool not available")
        }
        
        return try pool.read { (db) in
            let schemaEdgeItems = try ItemRecord
               .filter(Column(ItemRecord.Columns.type.rawValue) == "ItemEdgeSchema"
                        && Column(ItemRecord.Columns.syncState.rawValue) == SyncState.create)
                    .fetchAll(db)
            let syncEdgeItems = schemaEdgeItems.compactMap{$0.schemaEdgeDict(db: db)}
            
            let bulkAction = PodAPIPayload.BulkAction(
                createItems: syncEdgeItems,
                updateItems: [],
                deleteItems: [],
                createEdges: [])
            return bulkAction
        }
    }
    
    private func makeSyncUploadData(maxItems: Int = 100) throws -> PodAPIPayload.BulkAction  {
        let createItems = try ItemRecord.syncItemsWithState(.create, maxItems: maxItems)
        let updateItems = try ItemRecord.syncItemsWithState(.update, maxItems: maxItems)
        
        let bulkAction = PodAPIPayload.BulkAction(
            createItems: createItems,
            updateItems: updateItems,
            deleteItems: [],
            createEdges: [])
        
        return bulkAction
    }
    
    private func makeSyncEdgesData(maxItems: Int = 100) throws -> PodAPIPayload.BulkAction  {
        let createEdges = try EdgeRecord.syncEdgesWithState(.create, maxItems: maxItems)

        let bulkAction = PodAPIPayload.BulkAction(
            createItems: [],
            updateItems: [],
            deleteItems: [],
            createEdges: createEdges)
        
        return bulkAction
    }
    
    private func bulkAction(bulkPayload: PodAPIPayload.BulkAction, completion: ((Error?) -> Void)?) throws {
        guard let connectionConfig = AppController.shared.podConnectionConfig
          else { throw StringError(description: "No pod connection config") }

        let request = PodAPIStandardRequest.bulkAction(payload: bulkPayload)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        networkCall.sink { (result) in
            if result.error != nil, let resultData = result.data {
                let errorString = String(data: resultData, encoding: .utf8)
                print("ERROR: ", errorString ?? "")
            }
            completion?(result.error)
        } .store(in: &subscriptions)
    }
        
    private func searchAction(payload: [String: AnyEncodable], completion: ((Data?, Error?) -> Void)?) throws {
        guard let connectionConfig = AppController.shared.podConnectionConfig
          else { throw StringError(description: "No pod connection config") }

        let request = PodAPIStandardRequest.searchAction(payload: payload)
        
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        networkCall.sink { (result) in
            #if DEBUG
                if result.error != nil, let resultData = result.data {
                    let errorString = String(data: resultData, encoding: .utf8)
                    print("ERROR: ", errorString ?? "")
                } else if let responseData = result.data {
                    print("SEARCH RESPONSE: ", String(data: responseData, encoding: .utf8) ?? "")
                }
            #endif

            completion?(result.data, result.error)
        } .store(in: &subscriptions)
    }
    
    public func uploadFile(uuid: String, completion: ((Data?, Error?) -> Void)?) throws {
        guard let connectionConfig = AppController.shared.podConnectionConfig
          else { throw StringError(description: "No pod connection config") }

        let fileURL = FileStorageController.getURLForFile(withUUID: uuid)
        let request = try PodAPIUploadRequest.uploadFile(fileURL: fileURL, connectionConfig: connectionConfig)
        
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        networkCall.sink { (result) in
            if result.error != nil, let resultData = result.data {
                let errorString = String(data: resultData, encoding: .utf8)
                if errorString == "Failure: File already exists" {
                    completion?(result.data, nil)
                    return
                }
                print("ERROR: ", errorString ?? "")
            }
            completion?(result.data, result.error)
        } .store(in: &subscriptions)
    }
    
    public func downloadFile(sha256: String, fileName: String, completion: ((Error?) -> Void)?) throws {
        guard let connectionConfig = AppController.shared.podConnectionConfig
          else { throw StringError(description: "No pod connection config") }

        let request = PodAPIDownloadRequest.downloadFile(fileSHAHash: sha256, fileUID: fileName)
        
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        networkCall.sink { (result) in
            if result.error != nil {
                print("ERROR: ", result.error?.localizedDescription ?? "")
            }
            completion?(result.error)
        } .store(in: &subscriptions)
    }

}
