//
// CalendarHelper.swift
// Copyright © 2020 memri. All rights reserved.

import Foundation

/// This struct contains a collection of useful functions for working with dates/times
struct CalendarHelper {
    /// The calendar to use for calculations. This defaults to the users system calendar
    var calendar = Calendar.current
    
    var dayFormatter: DateFormatter = {
        let formatter = DateFormatter()
        formatter.dateFormat = "d"
        return formatter
    }()

    var monthFormatter: DateFormatter = {
        let formatter = DateFormatter()
        formatter.dateFormat = "MMM"
        return formatter
    }()

    var monthYearFormatter: DateFormatter = {
        let formatter = DateFormatter()
        formatter.dateFormat = "MMM YYYY"
        return formatter
    }()

    /// Get an array of dates representing the start of each month in the specified period
    func getMonths(from startDate: Date, to endDate: Date) -> [Date] {
        guard endDate >= startDate,
              let firstMonth = startOfMonth(for: startDate),
              let lastMonth = startOfMonth(for: endDate)
        else { return [] }
        var dates: [Date] = []
        calendar.enumerateDates(
            startingAfter: firstMonth.addingTimeInterval(-1),
            matching: DateComponents(day: 1),
            matchingPolicy: .nextTime
        ) { date, _, stop in
            if let date = date {
                dates.append(date)
                if date >= lastMonth { stop = true }
            }
        }
        return dates
    }

    /// Get an array of dates representing the start of each day during the specified period
    func getDays(forMonth month: Date) -> [Date] {
        guard let endOfMonth = endOfMonth(for: month) else { return [] }
        var dates: [Date] = []
        calendar.enumerateDates(
            startingAfter: month.addingTimeInterval(-1),
            matching: DateComponents(hour: 0),
            matchingPolicy: .nextTime
        ) { date, _, stop in
            if let date = date {
                dates.append(date)
                if date >= endOfMonth { stop = true }
            }
        }
        return dates
    }

    /// Get an array of dates representing the start of each day in the specified period with nil values appended to the start in order to make index 0 represent the first day of a week
    /// (useful for padding the start of the month with empty cells in a calendar view)
    func getPaddedDays(forMonth month: Date) -> [Date?] {
        guard let weekdayAtStart = weekdayAtStartOfMonth(for: month),
              let endOfMonth = endOfMonth(for: month) else { return [] }
        let adjustedWeekday = (weekdayAtStart - 1) // 0 = Sunday
        var dates: [Date?] = .init(repeating: nil, count: adjustedWeekday)
        calendar.enumerateDates(
            startingAfter: month.addingTimeInterval(-1),
            matching: DateComponents(hour: 0),
            matchingPolicy: .nextTime
        ) { date, _, stop in
            if let date = date {
                dates.append(date)
                if date >= endOfMonth { stop = true }
            }
        }
        return dates
    }

    /// Returns true if the two dates are on the same day
    func areOnSameDay(_ a: Date, _ b: Date) -> Bool {
        calendar.isDate(a, inSameDayAs: b)
    }

    /// Returns true if the date is the same day as NOW
    func isToday(_ date: Date) -> Bool {
        calendar.isDateInToday(date)
    }

    /// Returns true if the given date has the same DateComponents as NOW.
    /// Eg. to check if in same year and month
    func isSameAsNow(_ date: Date, byComponents: Set<Calendar.Component>) -> Bool {
        let components = calendar.dateComponents(byComponents, from: date)
        return calendar.date(Date(), matchesComponents: components)
    }

    func dayString(for date: Date) -> String {
        dayFormatter.string(from: date)
    }

    func monthString(for date: Date) -> String {
        monthFormatter.string(from: date)
    }

    func monthYearString(for date: Date) -> String {
        if calendar.isDate(date, equalTo: Date(), toGranularity: .year) {
            return monthFormatter.string(from: date)
        }
        else {
            return monthYearFormatter.string(from: date)
        }
    }

    var daysInWeek: [String] {
        calendar.shortStandaloneWeekdaySymbols
    }

    func weekdayAtStartOfMonth(for date: Date) -> Int? {
        guard let startOfMonth = startOfMonth(for: date) else { return nil }
        return calendar.component(.weekday, from: startOfMonth)
    }

    /// Returns a date representing the start of the day on the supplied date
    func startOfDay(for date: Date) -> Date {
        calendar.startOfDay(for: date)
    }

    /// Returns a date representing the end of the day on the supplied date
    func endOfDay(for date: Date) -> Date {
        calendar.startOfDay(for: date) + 24*60*60
    }
    
    /// Returns a date interval representing the start and end of the day on the supplied date
    func wholeDay(for date: Date) -> DateInterval {
        DateInterval(start: calendar.startOfDay(for: date), duration: 24*60*60)
    }
    
    /// Returns a date representing the start of the first day in the same month as the supplied date
    func startOfMonth(for date: Date) -> Date? {
        guard let interval = calendar.dateInterval(of: .month, for: date) else { return nil }
        return interval.start
    }

    /// Returns a date representing the end of the last day in the same month as the supplied date
    func endOfMonth(for date: Date) -> Date? {
        guard let interval = calendar.dateInterval(of: .month, for: date) else { return nil }
        return calendar.date(byAdding: DateComponents(day: -1), to: interval.end)
    }

    /// Returns a date representing the start of the first day in the same year as the supplied date
    func startOfYear(for date: Date) -> Date? {
        guard let interval = calendar.dateInterval(of: .year, for: date) else { return nil }
        return interval.start
    }

    /// Returns a date representing the end of the last day in the same year as the supplied date
    func endOfYear(for date: Date) -> Date? {
        guard let interval = calendar.dateInterval(of: .year, for: date) else { return nil }
        return calendar.date(byAdding: DateComponents(day: -1), to: interval.end)
    }
}
