//
//  Browser.swift
//  Memri
//
//  Created by Chaitanya Pandit on 25/03/21.
//

import SwiftUI

struct BrowserView: View {
    var context: ViewContextController
    @EnvironmentObject var sceneController: SceneController

    /// Keep track of whether the search bar is currently open (keyboard shown)
    @State var searchBarOpen: Bool = false
    @Environment(\.presentationMode) var presentationMode: Binding<PresentationMode>

    @State var filterPanelIsVisible: Bool = false
    @GestureState var filterPanelGestureOffset: CGFloat = .zero

    @State var showContextPane: Bool = false

    /// Translates the rendererName to the correct Renderer view
    @ViewBuilder
    var renderer: some View {
        switch context.config.rendererName.lowercased() {
        case "list":
            ListRendererView(viewContext: context)
        case "grid":
            GridRendererView(viewContext: context)
        case "map":
            MapRendererView(viewContext: context)
        case "timeline":
            TimelineRendererView(viewContext: context)
        case "calendar":
            CalendarRendererView(viewContext: context)
        case "photoviewer":
            PhotoViewerRendererView(viewContext: context)
        case "chart":
            ChartRendererView(viewContext: context)
        case "singleitem":
            SingleItemRendererView(viewContext: context)
        case "noteeditor":
            NoteEditorRendererView(viewContext: context)
        case "labelannotation":
            LabelAnnotationRendererView(viewContext: context)
        case "fileviewer":
            FileRendererView(viewContext: context)
        case "generaleditor":
            GeneralEditorRendererView(viewContext: context)
        default:
            Text("No renderer selected").bold().frame(maxWidth: .infinity, maxHeight: .infinity)
        }
    }
    
    @ViewBuilder
    var topBarView: some View {
        VStack(spacing: 0) {
            HStack(spacing: 4) {
                Button(action: {
                    sceneController.closeLastInStack()
                }) {
                    Image(systemName: "xmark.circle")
                        .rotationEffect(sceneController.navigationIsVisible ? Angle(degrees: 90) : .zero)
                        .padding(10)
                        .contentShape(Rectangle())
                }
                if let title = context.viewDefinitionPropertyResolver.string("title") {
                    Spacer()
                    Text(title).font(.headline).frame(maxWidth: .infinity)
                    Spacer()
                }
            }
            .padding(.horizontal, 10)
            .font(Font.system(size: 20, weight: .medium))
            .frame(minHeight: 48)
            .background(Color(.secondarySystemBackground).edgesIgnoringSafeArea(.top))
            
            Divider()
        }
    }

    var body: some View {
        self.sceneController.addToStack(self.presentationMode)

        return VStack(spacing: 0) {
            topBarView
            renderer
            BottomBarView(context: context,
                          onSearchPressed: { searchBarOpen = true },
                          onFilterButtonPressed: {
                            withAnimation {
                                self.filterPanelIsVisible = true
                            }
                          })
            if (self.filterPanelIsVisible) {
             Color.black.opacity(0.15)
                 .onTapGesture {
                     withAnimation {
                        self.filterPanelIsVisible = false
                     }
                 }
                 .gesture(DragGesture()
                     .updating($filterPanelGestureOffset) { value, state, _ in
                         state = max(0, value.translation.height)
                     }.onEnded { value in
                         if value.predictedEndTranslation.height > 20 {
                             withAnimation {
                                self.filterPanelIsVisible = false
                             }
                         }
                     })
             VStack {
                 Capsule()
                     .fill(Color(.secondarySystemBackground))
                     .frame(width: 40, height: 5)
                     .frame(maxWidth: .infinity)
                     .frame(height: 15)
                FilterPanelView(context: self.context)
             }
             .offset(y: filterPanelGestureOffset)
             .transition(.move(edge: .bottom))
             .zIndex(9)
         }
         if (self.showContextPane) {
            ContextPane(context: self.context, onPaneClosed: {
                showContextPane = false
            })
            .zIndex(15)
         }
        }
        .onAppear {
            context.onAppear()
        }
        .onDisappear {
            context.onDisappear()
        }
    }
}

