//
//  MapView.swift
//  MemriDatabase
//
//  Created by T Brennan on 8/1/21.
//

import Combine
import Foundation
import MapKit
import SwiftUI

struct MapView: UIViewRepresentable {
    var config: MapViewConfig

    func makeUIView(context: Context) -> MKMapView {
        let mapView = MKMapView(frame: .zero)
        mapView.delegate = context.coordinator
        context.coordinator.setup(mapView)

        mapView.alpha = 0

        return mapView
    }

    func updateUIView(_: MKMapView, context: Context) {
        context.coordinator.mapModel.dataItems = config.dataItems
        context.coordinator.mapModel.locationResolver = config.locationResolver
        context.coordinator.mapModel.addressResolver = config.addressResolver
        context.coordinator.mapModel.labelResolver = config.labelResolver
        context.coordinator.mapView?.isScrollEnabled = config.moveable
    }

    func makeCoordinator() -> MapView.Coordinator {
        Coordinator(self)
    }

    // MARK: - Implementing MGLMapViewDelegate

    final class Coordinator: NSObject, MKMapViewDelegate {
        var parent: MapView
        var mapModel = MapModel()

        var mapView: MKMapView?

        var cancellableBag: Set<AnyCancellable> = []

        init(_ parent: MapView) {
            self.parent = parent
            super.init()

            mapModel.didChange.sink { [weak self] in
                self?.updateFromModel()
            }
            .store(in: &cancellableBag)
        }

        func setup(_ mapView: MKMapView) {
            self.mapView = mapView
            mapView.register(
                MapMarkerView_AppleMaps.self,
                forAnnotationViewWithReuseIdentifier: MKMapViewDefaultAnnotationViewReuseIdentifier
            )
        }

        var firstSet: Bool = true
        func updateFromModel() {
            guard let mapView = mapView else { return }

            let newAnnotations = mapModel.items.map { item -> MKAnnotation in
                let annotation = MapAnnotation_AppleMaps(
                    title: item.label,
                    coordinate: item.coordinate,
                    dataItem: item.dataItem
                )
                return annotation
            }
            mapView.removeAnnotations(mapView.annotations)
            mapView.addAnnotations(newAnnotations)

            if !mapModel.items.isEmpty {
                UIView.animate(withDuration: 0.3) {
                    mapView.alpha = 1
                }
                fitDataInView(animated: !firstSet)
                firstSet = false
            }
        }

        func fitDataInView(animated: Bool = true) {
            mapView?.showAnnotations(mapView?.annotations ?? [], animated: animated)
        }

        //        func mapView(_ mapView: MKMapView, viewFor annotation: MKAnnotation) -> MKAnnotationView? {
//
        //        }

        func mapView(
            _: MKMapView,
            annotationView view: MKAnnotationView,
            calloutAccessoryControlTapped _: UIControl
        ) {
            // ONPRESS
            guard let annotation = view.annotation as? MapAnnotation_AppleMaps,
                  let dataItem = annotation.dataItem else { return }
            parent.config.onPress?(dataItem)
        }
    }
}

class MapAnnotation_AppleMaps: NSObject, MKAnnotation {
    let title: String?
    let coordinate: CLLocationCoordinate2D
    let dataItem: ItemRecord?

    init(
        title: String?,
        coordinate: CLLocationCoordinate2D,
        dataItem: ItemRecord?
    ) {
        self.title = title
        self.coordinate = coordinate
        self.dataItem = dataItem

        super.init()
    }
}

class MapMarkerView_AppleMaps: MKMarkerAnnotationView {
    override var annotation: MKAnnotation? {
        willSet {
            canShowCallout = true
            calloutOffset = .zero
            rightCalloutAccessoryView = UIButton(type: .detailDisclosure)
            // markerTintColor
        }
    }
}
