//
//  ContextPane.swift
//  Memri
//
//  Created by Chaitanya Pandit on 05/04/21.
//

import SwiftUI

struct ContextPane: View {
    @ObservedObject var context: ViewContextController
    var widthRatio: CGFloat = 0.75
    var onPaneClosed: (() -> Void)

    @GestureState(reset: { _, transaction in
        transaction.animation = .default
    }) var offset: CGFloat = .zero

    @State var isVisible: Bool = false

    func paneWidth(_ geom: GeometryProxy) -> CGFloat {
        min(geom.size.width * widthRatio, 300)
    }

    func cappedOffset(_ geom: GeometryProxy) -> CGFloat {
        min(max(0, offset), paneWidth(geom))
    }

    func fractionVisible(_ geom: GeometryProxy) -> Double {
        1 - Double(abs(cappedOffset(geom)) / paneWidth(geom))
    }

    var body: some View {
        GeometryReader { geom in
            self.body(withGeom: geom)
        }
    }

    
    func body(withGeom geom: GeometryProxy) -> some View {
        ZStack(alignment: .trailing) {
            if isVisible {
                ContextPaneBackground()
                    .opacity(fractionVisible(geom) * 0.5)
                    .edgesIgnoringSafeArea(.vertical)
                    .transition(.opacity)
                    .gesture(TapGesture()
                        .onEnded { _ in
                            withAnimation {
                                isVisible = false
                                onPaneClosed()
                            }
                        })
                    .zIndex(-1)
                ContextPaneForeground(context: context)
                    .frame(width: paneWidth(geom))
                    .offset(x: cappedOffset(geom))
                    .edgesIgnoringSafeArea(.vertical)
                    .transition(.move(edge: .trailing))
            }
        }
        .simultaneousGesture(contextPaneDragGesture)
        .onAppear{
            withAnimation {
                isVisible = true
            }
        }
    }

    var contextPaneDragGesture: some Gesture {
        DragGesture()
            .updating($offset, body: { value, offset, _ in
                offset = value.translation.width
            })
            .onEnded { value in
                if value.predictedEndTranslation.width > 100, abs(value.translation.width) > 10 {
                    isVisible = false
                    onPaneClosed()
                }
            }
    }
}

struct ContextPaneBackground: View {
    var body: some View {
        Color.gray
    }
}

struct BackgroundPane_Previews: PreviewProvider {
    static var previews: some View {
        ContextPaneBackground()
    }
}

struct ContextPaneForeground: View {
    @EnvironmentObject var sceneController: SceneController
    var paddingLeft: CGFloat = 25
    let context: ViewContextController
    
    var definition: CVUDefinitionContent? {
        guard let def = context.itemPropertyResolver,
              let actionDef = def.properties["contextPane"]?.getSubdefinition() else {
            return nil
        }
        
        return actionDef
    }
    
    var resolver: CVUPropertyResolver? {
        guard let def = definition else {
            return nil
        }
        
        return context.propertyResolverFor(def)
    }
    
    var labels: [ItemRecord] {
        return resolver?.items("labels") ?? []
    }
    
    var buttons: [CVUAction] {
        return resolver?.actions(key: "buttons") ?? []
    }
    
    var actions: [CVUAction] {
        return resolver?.actions(key: "actions") ?? []
    }
    
    var navigations: [CVUAction] {
        let navigations = resolver?.actions(key: "navigate") ?? []
        return navigations
    }
    
    var describeChangelog: String? {
        let changeLog = resolver?.string("changeLog")
        return changeLog
    }
    
    var title: String? {
        let title = self.resolver?.string("title")
        return title
    }
    
    var subtitle: String? {
        return nil
    }
    
    func titleForAction(_ action: CVUAction) -> String? {
        let actionResolver = context.propertyResolverFor(action.vars)
        return actionResolver.string("title")
    }
    
    func itemValueForKey(_ key: String, _ item: ItemRecord) -> String? {
        guard case let .string(stringValue) = item.propertyValue(key) else {
            return nil
        }

        return stringValue
    }
    
    
    var body: some View {
        return
            ScrollView(.vertical) {
                VStack(alignment: .leading) {
                    VStack(alignment: .leading) {
                        Text(title ?? "") // TODO: make this generic
                            .font(.system(size: 23, weight: .regular, design: .default))
                            .fontWeight(.bold)
                            .opacity(0.75)
                            .padding(.horizontal, paddingLeft)
                            .padding(.vertical, 5)
                        Text(subtitle ?? "")
                            .font(.body)
                            .opacity(0.75)
                            .padding(.horizontal, paddingLeft)

                        HStack {
                            ForEach(buttons,
                                    id: \.id) { actionItem in
                                ActionButton(action: actionItem, context: context.getCVUContext(forItem: context.focusedItem))
                            }
                        }
                        .padding(.horizontal, paddingLeft)
                        .padding(.bottom, 15)

                        Divider()
                        (describeChangelog)
                            .map { description in
                                Text(description)
                            }
                            .padding(.horizontal, paddingLeft)
                            .padding(.vertical, 10)
                            .multilineTextAlignment(.center)
                            .font(.system(size: 14, weight: .regular, design: .default))
                            .opacity(0.6)
                        Divider()
                    }
                    HStack {
                        Text("ACTIONS")
                            .fontWeight(.bold)
                            .opacity(0.4)
                            .font(.system(size: 16, weight: .regular, design: .default))
                            .padding(.horizontal, paddingLeft)
                        Spacer()
                    }
                    .padding(.top, 15)
                    .padding(.bottom, 10)
                    VStack(alignment: .leading, spacing: 0) {
                        ForEach(actions,
                                id: \.id) { actionItem in
                            Button(action: {
                                actionItem.execute(sceneController: sceneController, context: context.getCVUContext(forItem: context.focusedItem), completion: nil)
                            }) {
                                Text(titleForAction(actionItem) ?? "")
                                    .foregroundColor(Color(.label))
                                    .opacity(0.6)
                                    .font(.system(size: 20, weight: .regular, design: .default))
                                    .padding(.vertical, 10)
                            }
                        }
                    }
                    .padding(.horizontal, self.paddingLeft)
                    Divider()
                    HStack {
                        Text("NAVIGATE")
                            .fontWeight(.bold)
                            .opacity(0.4)
                            .font(.system(size: 16, weight: .regular, design: .default))
                            .padding(.horizontal, paddingLeft)
                        Spacer()
                    }
                    .padding(.top, 15)
                    .padding(.bottom, 10)
                    VStack(alignment: .leading, spacing: 0) {
                        ForEach(navigations, id: \.id) { navigateItem in
                            Button(action: {
                                navigateItem.execute(sceneController: sceneController, context: context.getCVUContext(forItem: context.focusedItem), completion: nil)
                            }) {
                                Text(titleForAction(navigateItem) ?? "")
                                    .foregroundColor(Color(.label))
                                    .opacity(0.6)
                                    .font(.system(size: 20, weight: .regular, design: .default))
                                    .padding(.vertical, 10)
                            }
                            .padding(.horizontal, self.paddingLeft)
                        }
                    }
                    Divider()
                    HStack {
                        Text("LABELS")
                            .fontWeight(.bold)
                            .opacity(0.4)
                            .font(.system(size: 16, weight: .regular, design: .default))
                            .padding(.horizontal, paddingLeft)
                        Spacer()
                    }
                    .padding(.top, 15)
                    .padding(.bottom, 15)
                    VStack(alignment: .leading, spacing: 10) {
                        ForEach(labels, id: \.id) { labelItem in
                            Button(action: {
                               // todo noop context.executeAction(addLabelAction, with: labelItem)
                            }) {
                                Text(itemValueForKey("name", labelItem) ?? "")
                                    .foregroundColor(Color(.label))
                                    .opacity(0.6)
                                    .font(.system(size: 20, weight: .regular, design: .default))
                                    .padding(.vertical, 5)
                                    .padding(.horizontal, 15)
                                    .frame(minWidth: 150, alignment: .leading)
                            }
                            .background(Color(hex: itemValueForKey("color", labelItem) ?? "#ffd966ff"))
                            .cornerRadius(5)
                            .padding(.horizontal, self.paddingLeft)
                        }
                    }
                    Spacer()
                }
                .padding(.top, 60)
            }
            .background(Color(.secondarySystemGroupedBackground))
    }
}
