//
//  FilterPanelView.swift
//  Memri
//
//  Created by T Brennan on 31/1/21.
//

import Foundation
import SwiftUI

struct FilterPanelView: View {
    @ObservedObject var context: ViewContextController
    @State var currentTab: FilterPanelTab = .rendererOptions
    static let excludedSortFields = ["uid", "deleted", "externalId", "version", "allEdges"]
    static let defaultSortFields = ["dateCreated", "dateModified"]

    var clipShape: some Shape {
        RoundedCornerRectangle(radius: 20, corners: [.topLeft, .topRight])
    }
    
    var currentTabTitle: String {
        switch currentTab {
        case .renderer:
            return "Renderer selection"
        case .rendererOptions:
            return "Renderer options"
        case .filterOptions:
            return "Filter options"
        case .sortOptions:
            return "Sort options"
        }
    }
    
    var body: some View {
        VStack(spacing: 0) {
            HStack(spacing: 0) {
                tabButton(icon: Image(systemName: "note"), tab: .renderer)
                Divider()
                tabButton(icon: Image(systemName: "gear"), tab: .rendererOptions)
                Divider()
                tabButton(icon: Image(systemName: "clock.arrow.circlepath"), tab: .filterOptions)
                Divider()
                tabButton(icon: Image(systemName: "arrow.up.arrow.down.square"), tab: .sortOptions)
                Divider()
            }
            .frame(height: 40)
            Divider()
            NavigationView {
                Form {
                    Group {
                        switch currentTab {
                        case .renderer:
                            rendererTab
                        case .filterOptions:
                            filterOptionsTab
                        case .rendererOptions:
                            rendererOptionsTab
                        case .sortOptions:
                            sortOptionsTab
                        }
                    }
                }
                .navigationTitle(currentTabTitle)
                .navigationBarItems(trailing:
                    HStack {
                        if currentTab == .sortOptions {
                            Button(action: {
                                $context.config.query.sortAscending.wrappedValue = !$context.config.query.sortAscending.wrappedValue
                            }) {
                                Image(systemName: !$context.config.query.sortAscending.wrappedValue ? "arrow.up" : "arrow.down")
                                    .padding(5)
                            }
                        }
                    }
                )
                .navigationBarTitleDisplayMode(.inline)
            }
        }
        .frame(maxWidth: .infinity)
        .frame(height: 340)
        .clipShape(clipShape)
        .background(
            clipShape
                .fill(Color(.systemBackground))
                .shadow(radius: 10)
                .edgesIgnoringSafeArea([.bottom])
        )
    }
    
    @ViewBuilder
    func tabButton(icon: Image, tab: FilterPanelTab) -> some View {
        Button(action: {
            currentTab = tab
        }) {
            icon
                .padding(5)
                .frame(maxWidth: .infinity, maxHeight: .infinity)
                .background(
                    currentTab == tab ? Color(.systemFill) : Color(.secondarySystemBackground)
                )
        }
    }
    
    @ViewBuilder
    func optionalDateRow(title: String, selection: Binding<Date?>, initialSet: Date = Date()) -> some View {
        OptionalDatePicker(title: title, selection: selection, initialSet: initialSet)
    }
    
    enum FilterPanelTab: String, Identifiable, Hashable {
        case renderer
        case filterOptions
        case rendererOptions
        case sortOptions
        
        var id: Int { hashValue }
    }
    
    var rendererTab: some View {
        Group {
            ForEach(context.supportedRenderers.sorted(), id: \.self) { rendererName in
                Button(action: {
                    context.config.rendererName = rendererName
                }) {
                    Text(rendererName.localizedCapitalized)
                        .if(rendererName == context.config.rendererName) {
                            $0.bold()
                        }
                }
            }
        }
    }
    
    var rendererOptionsTab: some View {
        Group {
            switch context.config.rendererName.lowercased() {
            case "timeline":
                TimelineRendererSettingsView(context: context)
            case "chart":
                ChartRendererSettingsView(context: context)
            case "grid":
                GridRendererSettingsView(context: context)
            default:
                Text("No configurable settings for this renderer.")
            }
        }
    }
    
    var filterOptionsTab: some View {
        Group {
            optionalDateRow(title: "Modified after",
                            selection: $context.config.query.dateModifiedAfter, initialSet: Date.oneWeekAgo)
            optionalDateRow(title: "Modified before",
                            selection: $context.config.query.dateModifiedBefore)
            optionalDateRow(title: "Created after",
                            selection: $context.config.query.dateCreatedAfter, initialSet: Date.oneWeekAgo)
            optionalDateRow(title: "Created before",
                            selection: $context.config.query.dateCreatedBefore)
        }
    }
    
    var sortFields: [String]? {
        if let item = context.items.first  {
            var fields = context.databaseController.schema.propertyNamesForItemType(itemType: item.type)
            fields.append(contentsOf: FilterPanelView.defaultSortFields)
            fields.sort()
            return fields
        }
        
        return nil
    }
    
    var sortOptionsTab: some View {
        Group {
            if let fields = sortFields {
                ForEach(fields, id: \.id) { field in
                    if !FilterPanelView.excludedSortFields.contains(field) {
                        FilterPanelSortItemView(property: field, selection: $context.config.query.sortProperty)
                    }
                }
            } else {
                Text("No sort options available.")
            }
        }
    }
}
