//
// NavigationView.swift
// Copyright © 2020 memri. All rights reserved.

import ASCollectionView
import SwiftUI

/// This view manages displaying the navigation pane, as well as its gestures
struct NavigationWrapperView<Content: View>: View {
    @ObservedObject var sceneController: SceneController
    
    init(sceneController: SceneController, @ViewBuilder content: () -> Content) {
        self.sceneController = sceneController
        self.content = content()
    }

    var content: Content
    var widthRatio: CGFloat = 0.8
    
    @GestureState(reset: { _, transaction in
        transaction.animation = .default
    }) var offset: CGFloat = .zero

    func navWidth(_ geom: GeometryProxy) -> CGFloat {
        min(300, geom.size.width * widthRatio)
    }

    func cappedOffset(_ geom: GeometryProxy) -> CGFloat {
        if sceneController.navigationIsVisible {
            return max(min(0, offset), -navWidth(geom))
        }
        else {
            return min(max(0, offset), navWidth(geom))
        }
    }

    func fractionVisible(_ geom: GeometryProxy) -> Double {
        let fraction = Double(abs(cappedOffset(geom)) / navWidth(geom))
        return sceneController.navigationIsVisible ? 1 - fraction : fraction
    }

    var body: some View {
        Group {
            if Constants.memri_shouldUseLargeScreenLayout {
                GeometryReader { geom in
                    self.bodyForLargeScreen(withGeom: geom)
                }
            }
            else {
                GeometryReader { geom in
                    self.body(withGeom: geom)
                }
            }
        }
    }

    func body(withGeom geom: GeometryProxy) -> some View {
        ZStack(alignment: .leading) {
            content
                .frame(width: geom.size.width, height: geom.size.height, alignment: .topLeading)
                .offset(x: sceneController.navigationIsVisible ? navWidth(geom) + cappedOffset(geom) : cappedOffset(geom))
                .disabled(sceneController.navigationIsVisible)
                .zIndex(-1)
            Color.clear
                .contentShape(Rectangle())
                .frame(minWidth: 10, maxWidth: 10, maxHeight: .infinity)
                .simultaneousGesture(
                    navigationDragGesture,
                    including: (sceneController.isContentFullscreen) ? .none : .all
                )
            if sceneController.navigationIsVisible || offset > 0 {
                Color.black
                    .opacity(fractionVisible(geom) * 0.5)
                    .edgesIgnoringSafeArea(.all)
                    .onTapGesture {
                        withAnimation {
                            sceneController.navigationIsVisible = false
                        }
                    }
                    .simultaneousGesture(navigationDragGesture)
                    .zIndex(10)
                NavigationPaneView(sceneController: sceneController)
                    .frame(width: navWidth(geom))
                    .edgesIgnoringSafeArea(.all)
                    .offset(
                        x: sceneController.navigationIsVisible ? cappedOffset(geom) : (-navWidth(geom) + cappedOffset(geom)),
                        y: 0
                    )
                    .simultaneousGesture(navigationDragGesture)
                    .transition(.move(edge: .leading))
                    .zIndex(15)
            }
        }
    }

    func bodyForLargeScreen(withGeom _: GeometryProxy) -> some View {
        HStack(spacing: 0) {
            NavigationPaneView(sceneController: sceneController)
                .frame(width: 300)
                .edgesIgnoringSafeArea(.all)
            content
        }
    }

    var navigationDragGesture: some Gesture {
        DragGesture()
            .updating($offset, body: { value, offset, _ in
                offset = value.translation.width
            })
            .onEnded { value in
                if
                    self.sceneController.navigationIsVisible ? value.predictedEndTranslation.width < -140 : value.translation
                    .width > 50,
                    abs(value.predictedEndTranslation.width) > 20
                {
                    withAnimation {
                        self.sceneController.navigationIsVisible.toggle()
                    }
                }
            }
    }
}
