//
//  GeneralEditorRenderer.swift
//  Memri
//
//  Created by T Brennan on 6/2/21.
//

import Foundation
import Combine
import SwiftUI

struct GeneralEditorLayoutItem {
    var id: String
    var dict: [String: CVUValue]
    
    func has(_ propName: String) -> Bool {
        dict[propName] != nil
    }
    
    func get<T>(_ propName: String, _: T.Type = T.self, _ item: ItemRecord? = nil) -> T? {
        guard let propValue = dict[propName] else {
            if propName == "section" {
                print("ERROR: tri")
            }

            return nil
        }

        if T.self == [String].self {
            switch propValue {
            case .array(let values):
                return (values.compactMap{(value) -> String? in
                    switch value {
                        case .constant(let constant) :
                            return constant.asString()
                    default:
                        return nil
                    }
                }) as? T
            case .constant(let constant):
                return [constant.asString()] as? T
            default:
                return nil
            }
        }
            
        if T.self == String.self {
            switch propValue {
            case .constant(let constant):
                return constant.asString() as? T
            default:
                return nil
            }
        }
        
        return propValue as? T
    }
}

/// The GeneralEditorRenderer
/// This presents an editor for a single item
struct GeneralEditorRendererView: View {
    @EnvironmentObject var sceneController: SceneController
    @ObservedObject var viewContext: ViewContextController

    var layout: [GeneralEditorLayoutItem] {
        guard let currentItem = viewContext.focusedItem else {
            return []
        }
        
        let possibleViewDefs = viewContext.cvuController.viewDefinitionFor(itemRecord: currentItem)?.definitions.first?["layout"] ??
            viewContext.cvuController.viewDefinitionFor(viewName: viewContext.config.viewName ?? "")?.definitions.first?["layout"]
        
        let viewDefs_ = possibleViewDefs.map { (def)  -> [[String: CVUValue]] in
            switch def {
            case .array(let values):
                return values.compactMap{(value) -> [String: CVUValue]? in
                    return value.getSubdefinition()?.properties
                }
            default:
                return []
            }
        }
                
        let generalDefs_ = viewContext.cvuController.rendererDefinitionFor(selector: "[renderer = \(self.viewContext.config.rendererName)]")?.properties["layout"].map { (def)  -> [[String: CVUValue]] in
            switch def {
            case .array(let values):
                return values.compactMap{(value) -> [String: CVUValue]? in
                    return value.getSubdefinition()?.properties
                }
            default:
                return []
            }
        }

        guard let viewDefs = viewDefs_, let generalDefs = generalDefs_ else {
            return []
        }
        
        var showDefaultLayout = true
        if case let .constant(.bool(show)) = viewContext.config.viewDefinition.properties["showDefaultLayout"] {
            showDefaultLayout = show
        }
        
        var mergedDefinitions = viewDefs
        if (showDefaultLayout) {
            // Merge layout from default generalEditor cvu
            mergedDefinitions.append(contentsOf: generalDefs)
        }
        
        var sections = [String]()
        var sectionInfos = [String: [String: CVUValue]]()
        for definition in mergedDefinitions {
            if let sectionName = definition["section"]?.toCVUString(depth: 0, tab: " ", includeInitialTab: false) {
                if (!sections.contains(sectionName)) {
                    sections.append(sectionName)
                }
                var newSection = sectionInfos[sectionName] ?? [:]
                newSection.merge(dict: definition)
                sectionInfos[sectionName] = newSection
            }
        }
        
        return sections.compactMap{
            if let value = sectionInfos[$0] {
                return GeneralEditorLayoutItem(id: $0, dict: value)
            }
            
            return nil
        }
    }
    
    var usedFields: Set<String> {
        var uniqueFields = Set<String>()
        
        for layoutItem in layout {
            if let fields = layoutItem.get("fields", [String].self) {
                for field in fields {
                    uniqueFields.insert(field)
                }
            }
        }
        
        return uniqueFields
    }
    
    @ViewBuilder
    var stackContent: some View {
        if let currentItem = viewContext.focusedItem {
            ForEach(layout, id: \.id) { layoutSection in
                GeneralEditorSection(viewContext: viewContext, isEditing: $sceneController.isInEditMode, layout: layoutSection, item: currentItem, usedFields: usedFields)
            }
        }
    }
    
    var body: some View {
        ScrollView(.vertical) {
            VStack(alignment: .leading, spacing: 0) {
                stackContent
            }
        }
    }
    
    @ViewBuilder
    func row(forProperty property: SchemaProperty, currentItem: ItemRecord) -> some View {
        let prettyName = property.property
        switch property.valueType {
        case .string:
            HStack {
                Text(prettyName)
                MemriTextField<String?>(value: currentItem.propertyBinding(name: property.property))
                    .padding(2)
                    .background(RoundedRectangle(cornerRadius: 6).fill(Color(.secondarySystemBackground)))
            }
        case .double:
            HStack {
                Text(prettyName)
                Spacer()
                MemriTextField<Double?>(value: currentItem.propertyBinding(name: property.property))
                    .padding(2)
                    .background(RoundedRectangle(cornerRadius: 6).fill(Color(.secondarySystemBackground)))
            }
        case .bool:
            Toggle(prettyName, isOn: currentItem.propertyBinding(name: property.property, defaultValue: false))
        case .datetime:
            OptionalDatePicker(title: prettyName, selection: currentItem.propertyBinding(name: property.property))
        default:
            Text(prettyName)
        }
    }
}

struct GeneralEditorSection: View {
    @EnvironmentObject var sceneController: SceneController
    @ObservedObject var viewContext: ViewContextController
    var isEditing: Binding<Bool>
    var layout: GeneralEditorLayoutItem
    
    var item: ItemRecord
    var usedFields: Set<String>
    
    var shouldShowTitle: Bool {
        guard let showTitle = self.nodeDefinition?.properties["showTitle"] else {
            return true
        }
        
        switch showTitle {
        case .constant(let value):
            switch value {
            case .bool(let boolValue):
                return boolValue
            default: break
            }
        default: break
        }
        
        return true
    }
    
    var sectionTitle: String? {
        guard let title = viewContext.cvuController.viewDefinitionFor(itemRecord: item)?.definitions.first?[layout.id]?.getSubdefinition()?.properties["title"] else {
            return layout.id
        }
    
        return viewContext.lookupController.resolve(value: title, context: viewContext.getCVUContext(forItem: item), db: viewContext.databaseController)
    }
        
    var isEmpty: Bool {
        if isEditing.wrappedValue {
            return false
        }
        return self.layout.has("edges") && self.fields.count == 0 && self.edgeItems.count == 0
    }
    
    var fields: [String] {
        var fields = layout.get("fields", [String].self) ?? []
        if fields == ["*"] {
            fields = []
            for propertyType in viewContext.databaseController.schema.propertyNamesForItemType(itemType: item.type) {
                if (!usedFields.contains(propertyType)) {
                    fields.append(propertyType)
                }
            }
        }
        
        return fields
    }
    
    var edgeItems: [ItemRecord] {
        guard let edges = layout.get("edges", [String].self) else {
            return []
        }
        
        var items = [ItemRecord]()
        for edge in edges {
            let edgeItems: [ItemRecord] = viewContext.lookupController.resolve(edge: edge, for: item, db: viewContext.databaseController)
            items.append(contentsOf: edgeItems)
        }
        
        return items
    }
    
    var nodeDefinition: CVUDefinitionContent? {
        if let nodeDefinition = viewContext.cvuController.viewDefinitionFor(itemRecord: item)?.definitions.first?[layout.id]?.getSubdefinition() {
            return nodeDefinition
        } else if let nodeDefinition = viewContext.cvuController.rendererDefinitionFor(viewName: viewContext.config.rendererName)?.properties[layout.id]?.getSubdefinition() {
            return nodeDefinition
        } else if let viewName = viewContext.config.viewName, let nodeDefinition = viewContext.cvuController.viewDefinitionFor(viewName: viewName)?.properties[layout.id]?.getSubdefinition() {
            return nodeDefinition
        }
        
        return nil
    }
    
    var viewArguments: CVUViewArguments? {
        guard isEditing.wrappedValue,
              let edgeType = layout.get("edges", [String].self)?.first,

              layout.has("edges") else {
            return nil
        }

        var args = viewContext.config.viewArguments?.args ?? [:]
        args["query"] = .constant(.string(item.type))
        args["type"] = .constant(.string(edgeType))
        args["subject"] = .item(item.id)
        args["item"] = .item(item.id)
        args["edgeType"] = .constant(.string(edgeType))
        args["distinct"] = .constant(.bool(false))
        
        return CVUViewArguments(args: args, argumentItem: item)
    }
    
    var action: CVUAction_OpenViewByName? {
        guard isEditing.wrappedValue,
              let edgeType = layout.get("edges", [String].self)?.first,
              let itemType = layout.get("type", String.self),
              layout.has("edges") else {
            return nil
        }
        
        let properties: [String: CVUValue] = ["query": .constant(.string(item.type)), "type": .constant(.string(edgeType)), "subject": .item(item.id), "item": .item(item.id), "edgeType": .constant(.string(edgeType)), "distinct": .constant(.bool(false))]
        let vars: [String: CVUValue] = ["viewArguments": .subdefinition(.init(definitions: [], children: [], properties: properties))]
        
        return CVUAction_OpenViewByName(vars: vars, viewName: "itemByQuery", renderer: "list", itemType: itemType)
    }

    var body: some View {        
        if isEmpty == true {
            EmptyView()
        } else {
            Section( header: Group {
                HStack(alignment: .bottom) {
                    if let title = sectionTitle, shouldShowTitle {
                        Text(title.uppercased())
                            .generalEditorHeader()
                        if let action = action {
                            Spacer()
                            ActionPopupButton(action: action)
                        }
                    }
                }.padding(.trailing, 20)
            },
            content: {
                Divider()
                if let nodeDefinition = self.nodeDefinition {
                    viewContext.render(item: item, nodeDefinition: nodeDefinition)
                } else {
                    ForEach(fields, id: \.id) { field in
                        if let fieldProperty = sceneController.appController.databaseController.schema.expectedType(forItemType: item.type, propertyName: field) {
                            let schemaProperty = SchemaProperty(itemType: item.type, property: field, valueType: fieldProperty)
                            DefaultGeneralEditorRow(viewContext:viewContext, isEditing: isEditing, property: schemaProperty, currentItem: item, prop: field, isLast: fields.last == field, item: item)
                        }
                    }
                    
                    if let edgeItems = self.edgeItems, edgeItems.count > 0 {
                        ForEach(edgeItems, id: \.id) { edgeItem in
                            if let nodeDefinition = viewContext.cvuController.edgeDefinitionFor(itemRecord: edgeItem) {
                                viewContext.render(item: edgeItem, nodeDefinition: nodeDefinition, viewArguments: viewArguments)
                            }
                        }
                    }
                }
                Divider()
            })
        }
    }
}

struct DefaultGeneralEditorRow: View {
    @EnvironmentObject var sceneController: SceneController
    @ObservedObject var viewContext: ViewContextController
    var isEditing: Binding<Bool>

    var property: SchemaProperty
    var currentItem: ItemRecord
    var prop: String
    var isLast: Bool
    var hasGroup = false
    var item: ItemRecord

    var isEmpty: Bool {
        return false
    }
    
    var nodeDefinition: CVUDefinitionContent? {
        if let nodeDefinition = viewContext.cvuController.viewDefinitionFor(itemRecord: currentItem)?.definitions.first?[prop]?.getSubdefinition() {
            return nodeDefinition
        } else if let nodeDefinition = viewContext.cvuController.rendererDefinitionFor(viewName: viewContext.config.rendererName)?.properties[prop]?.getSubdefinition() {
            return nodeDefinition
        } else if let viewName = viewContext.config.viewName, let nodeDefinition = viewContext.cvuController.viewDefinitionFor(viewName: viewName)?.properties[prop]?.getSubdefinition() {
            return nodeDefinition
        }
        
        return nil
    }
    
    var body: some View {
               
        return VStack(spacing: 0) {
            let propType = property.valueType
            
            VStack(alignment: .leading, spacing: 4) {
                Text(prop
                    .camelCaseToWords()
                    .lowercased()
                    .capitalizingFirst())
                    .generalEditorLabel()

                if let nodeDefinition = self.nodeDefinition {
                    viewContext.render(item: item, nodeDefinition: nodeDefinition)
                } else {
                    switch propType {
                    case .string:
                        stringRow()
                    case .bool:
                        boolRow()
                    case .datetime:
                        dateRow()
                    case .int:
                        intRow()
                    case .double:
                        doubleRow()
                    default:
                        defaultRow()
                    }
                }
            }
            .fullWidth()
            .padding(.bottom, 10)
            .padding(.horizontal)
            .background(Color(.systemBackground))

            if !isLast {
                Divider().padding(.leading, 35)
            }
        }
    }

    func stringRow() -> some View {
          let binding = Binding<String>(
              get: { return currentItem.propertyValue(property.property)?.asString() ?? ""},
              set: {
                  try? currentItem.setPropertyValue(name: self.prop, value: .string($0))
              }
          )

          return MemriTextField(value: binding,
                                clearButtonMode: .whileEditing,
                                isEditing: isEditing,
                                isSharedEditingBinding: true)
              .generalEditorCaption()
      }

    func boolRow() -> some View {
           let binding = Binding<Bool>(
               get: { return currentItem.propertyValue(property.property)?.asBool() ?? false},
               set: { _ in
                   do {
                       let currentVal = currentItem.propertyValue(property.property)?.asBool() ?? false
                       try? currentItem.setPropertyValue(name: self.prop, value: .bool(!currentVal))
                   }
               }
           )

           return Toggle(isOn: binding) {
               Text(prop
                   .camelCaseToWords()
                   .lowercased()
                   .capitalizingFirst())
           }
           .generalEditorCaption()
       }

    func intRow() -> some View {
          let binding = Binding<Int>(
              get: { return currentItem.propertyValue(property.property)?.asInt() ?? 0},
              set: {
                  try? currentItem.setPropertyValue(name: self.prop, value: .int($0))
              }
          )

          return MemriTextField(value: binding,
                                clearButtonMode: .whileEditing,
                                isEditing: isEditing,
                                isSharedEditingBinding: true)
              .generalEditorCaption()
      }

      func doubleRow() -> some View {
          let binding = Binding<Double>(
              get: { return currentItem.propertyValue(property.property)?.asDouble() ?? 0},
              set: {
                  try? currentItem.setPropertyValue(name: self.prop, value: .double($0))
              }
          )

          return MemriTextField(value: binding,
                                clearButtonMode: .whileEditing,
                                isEditing: isEditing,
                                isSharedEditingBinding: true)
              .generalEditorCaption()
      }

    @ViewBuilder
       func dateRow() -> some View {
           let binding = Binding<Date>(
               get: { return currentItem.propertyValue(property.property)?.asDate() ?? Date() },
               set: {
                   try? currentItem.setPropertyValue(name: self.prop, value: .datetime($0))
               }
           )

           DatePicker("", selection: binding, displayedComponents: .date)
               .labelsHidden()
       }

       func defaultRow(_ caption: String? = nil) -> some View {
           Text(caption ?? prop.camelCaseToWords().lowercased().capitalizingFirst())
               .generalEditorCaption()
       }

}

public extension View {
    func generalEditorLabel() -> some View { modifier(GeneralEditorLabel()) }
    func generalEditorCaption() -> some View { modifier(GeneralEditorCaption()) }
    func generalEditorHeader() -> some View { modifier(GeneralEditorHeader()) }
    func generalEditorInput() -> some View { modifier(GeneralEditorInput()) }
}

private struct GeneralEditorInput: ViewModifier {
    func body(content: Content) -> some View {
        content
            .fullHeight()
            .font(.system(size: 16, weight: .regular))
            .padding(10)
            .generalEditorCaption()
    }
}

private struct GeneralEditorLabel: ViewModifier {
    func body(content: Content) -> some View {
        content
            .foregroundColor(Color(hex: "#38761d"))
            .font(.system(size: 14, weight: .regular))
            .padding(.top, 10)
    }
}

private struct GeneralEditorCaption: ViewModifier {
    func body(content: Content) -> some View {
        content
            .font(.system(size: 18, weight: .regular))
            .foregroundColor(Color(hex: "#223322"))
    }
}

private struct GeneralEditorHeader: ViewModifier {
    func body(content: Content) -> some View {
        content
            .font(Font.system(size: 15, weight: .regular))
            .foregroundColor(Color(hex: "#434343"))
            .padding(.bottom, 5)
            .padding(.top, 24)
            .padding(.horizontal)
            .foregroundColor(Color(hex: "#333"))
    }
}
