//
// ListRenderer.swift
// Copyright © 2020 memri. All rights reserved.

import ASCollectionView
import Combine
import Foundation
import SwiftUI

/// The list renderer
/// This presents the data in a list (aka tableView)
struct ListRendererView: View {
    @EnvironmentObject var sceneController: SceneController
    @ObservedObject var viewContext: ViewContextController

    var isInEditMode: Bool {
        return viewContext.viewDefinitionPropertyResolver.bool("editMode", defaultValue: sceneController.isInEditMode) // sceneController.isInEditMode
    }
    
    var body: some View {
        let insets = viewContext.rendererDefinitionPropertyResolver.edgeInsets ?? .init(top: 10, left: 10, bottom: 10, right: 10)
        let spacing = viewContext.rendererDefinitionPropertyResolver.spacing ?? CGPoint(x: 10, y: 10)
        let backgroundColor = viewContext.rendererDefinitionPropertyResolver.backgroundColor ?? CVUColor.system(.systemBackground)
        let separatorsEnabled = !viewContext.rendererDefinitionPropertyResolver.bool("hideSeparators", defaultValue: false)
        
        VStack {
            if viewContext.hasItems {
                ASTableView(editMode: isInEditMode, section:
                    ASSection(id: 0,
                              data: viewContext.items,
                              dataID: \.id,
                              selectionMode: selectionMode,
                              onSwipeToDelete: { _, item in
                                let action = CVUAction_Delete()
                                action.execute(sceneController: sceneController, context: viewContext.getCVUContext(forItem: item), completion: nil)
                                return true
                              },
                              contextMenuProvider: contextMenuProvider) { dataItem, cellContext in
                        viewContext.render(item: dataItem)
                            .frame(maxWidth: .infinity, alignment: .leading)
                            .padding(EdgeInsets(top: cellContext.isFirstInSection ? 0 : spacing.y / 2,
                                leading: insets.left,
                                bottom: cellContext.isLastInSection ? 0 : spacing.y / 2,
                                trailing: insets.right))
                    })
                    .separatorsEnabled(separatorsEnabled)
                    .alwaysBounce()
                    .contentInsets(.init(
                        top: insets.top,
                        left: 0,
                        bottom: insets.bottom,
                        right: 0
                    ))
                    .background(backgroundColor.color)
            }
            else {
                Text("No items")
                    .multilineTextAlignment(.center)
                    .font(.system(size: 16, weight: .regular, design: .default))
                    .opacity(0.7)
                .padding(30)
                .frame(maxWidth: .infinity, maxHeight: .infinity)
                    .background(backgroundColor.color)
            }
        }
    }

    var selectionMode: ASSectionSelectionMode {
        if isInEditMode {
            return .selectMultiple(viewContext.selectedIndicesBinding)
        }
        else {
            return .selectSingle { index in
                guard let item = viewContext.items[safe: index],
                      let presses = viewContext.rendererDefinitionPropertyResolver.actions(key: "onPress") ?? viewContext.nodePropertyResolver(for: item)?.actions(key: "onPress") else {
                    return
                }
                
                presses.forEach { (press) in
                    press.execute(sceneController: sceneController, context: viewContext.getCVUContext(forItem: item), completion: nil)
                }
            }
        }
    }

    func contextMenuProvider(index: Int, item: ItemRecord) -> UIContextMenuConfiguration? {
        nil
//        UIContextMenuConfiguration(identifier: nil,
//                                   previewProvider: nil) { [weak viewContext] (_) -> UIMenu? in
//            let children: [UIMenuElement] = viewContext?.config.contextMenuActions
//                .map { [weak controller] action in
//                    UIAction(title: action.getString("title"),
//                             image: nil) { [weak controller] _ in
//                        controller?.context.executeAction(action, with: item)
//                    }
//                } ?? []
//            return UIMenu(title: "", children: children)
//        }
    }
}
