//
// PhotoViewerRenderer.swift
// Copyright © 2020 memri. All rights reserved.

import Foundation
import SwiftUI

/// The photo viewer renderer
/// This presents the data items in a photo viewer, that can page horizontally between images
/// The CVU for ItemType > map {...} must define a `file` expression pointing to a `File` item
struct PhotoViewerRendererView: View {
    @ObservedObject var viewContext: ViewContextController
    
    
    func photoItemProvider(forIndex index: Int) -> PhotoViewerController.PhotoItem? {
        guard let item = viewContext.items[safe: index],
              let fileURL = viewContext.nodePropertyResolver(for: item)?.fileURL(key: "file")
        else {
            return nil
        }
        let overlay = viewContext.render(item: item, blankIfNoDefinition: true).eraseToAnyView()
        return PhotoViewerController.PhotoItem(index: index, imageURL: fileURL, overlay: overlay)
    }

    func onToggleOverlayVisibility(_ visible: Bool) {
        withAnimation {
            self.isFullScreen = !visible
        }
    }

    func toggleFullscreen() {
        isFullScreen.toggle()
    }

    var isFullScreen: Bool {
        get { viewContext.isFullScreen }
        nonmutating set { viewContext.isFullScreen = newValue }
    }

    var body: some View {
        Group {
            if viewContext.hasItems {
                ZStack(alignment: .topLeading) {
                    PhotoViewerView(
                        photoItemProvider: photoItemProvider,
                        initialIndex: viewContext.focusedIndex,
                        onIndexChange: { (index) in
                            viewContext.focusedIndex = index
                        },
                        onToggleOverlayVisibility: onToggleOverlayVisibility
                    )
                    .edgesIgnoringSafeArea(viewContext.isFullScreen ? .all : [])
                }
            }
            else {
                Text("No photos found").bold().frame(maxWidth: .infinity, maxHeight: .infinity)
            }
        }
    }
}
