//
//  ViewContextControllerController.swift
//  Memri
//
//  Created by T Brennan on 30/1/21.
//

import Foundation
import Combine
import GRDB
import SwiftUI

class ViewContextController: ObservableObject {
    let configHolder: ViewContextHolder
    
    let databaseController: DatabaseController
    let cvuController: CVUController
    let lookupController: CVULookupController
    
    var configObservation: AnyCancellable?
    var queryObservation: AnyCancellable?
    var appStateObservation: AnyCancellable?

    var config: ViewContext {
        get { configHolder.config }
        set {
            if configHolder.config != newValue {
                configHolder.config = newValue
                updateCachedValues()
            }
        }
    }
    
    init(config: ViewContextHolder,
         databaseController: DatabaseController = AppController.shared.databaseController,
         cvuController: CVUController = AppController.shared.cvuController) {
        self.databaseController = databaseController
        self.cvuController = cvuController
        self.lookupController = CVULookupController()
        self.configHolder = config
        
        updateCachedValues()
        
        setupQueryObservation()
        
        // Watch for changes to the config
        configObservation = configHolder.configPublisher.sink { [weak self] _ in
            self?.setupQueryObservation()
        }
        
        appStateObservation = AppController.shared.statePublisher.sink {[weak self] _ in
            self?.setupQueryObservation()
        }
    }
    
    var focusedIndex: Int {
        get {
            guard let focusedItem = config.focusedItem else { return 0 }
            return items.firstIndex(of: focusedItem) ?? 0
        }
        set {
            if let item = items[safe: newValue] {
                config.focusedItem = item
            } else {
                config.focusedItem = nil
            }
        }
    }
    
    var focusedItem: ItemRecord? {
        get { config.focusedItem }
        set { config.focusedItem = newValue }
    }
    
    func onAppear() {
        isObservingQuery = true
    }
    
    func onDisappear() {
        isObservingQuery = false
    }

    func render(nodeDefinition: CVUDefinitionContent) -> some View {
        return cvuController.render(cvuContext: getCVUContext(), nodeDefinition: nodeDefinition, lookup: lookupController, db: databaseController, blankIfNoDefinition: false)
    }
    
    /// Return a SwiftUI view for the given item based on it's CVU definition.
    /// Set `overrideRenderer` if you want to render the item as though it is in a different renderer to the context (eg. "list" to get the list-specific appearance)
    func render(item: ItemRecord, items:[ItemRecord] = [], nodeDefinition: CVUDefinitionContent? = nil, viewArguments: CVUViewArguments? = nil, overrideRenderer: String? = nil, blankIfNoDefinition: Bool = false) -> some View {
        return cvuController.render(cvuContext: getCVUContext(forItem: item, items:items, viewArguments:viewArguments, overrideRenderer: overrideRenderer), nodeDefinition: nodeDefinition, lookup: lookupController, db: databaseController, blankIfNoDefinition: blankIfNoDefinition)
    }
    
    func getCVUContext(forItem item: ItemRecord? = nil, items: [ItemRecord] = [], viewArguments: CVUViewArguments? = nil, overrideRenderer: String? = nil) -> CVUContext {
       CVUContext(currentItem: item, items: items, selector: nil, viewName: config.viewName, rendererName: overrideRenderer ?? config.rendererName, viewDefinition: config.viewDefinition, viewArguments: viewArguments ?? config.viewArguments)
    }
    
    
    private func updateCachedValues() {
        
        if let viewName = config.viewName,
           let newDef = cvuController.viewDefinitionFor(viewName: viewName) {
            config.viewDefinition = newDef
        } else if let item = config.focusedItem,
                  let newDef = cvuController.viewDefinitionFor(itemRecord: item) {
            config.viewDefinition = newDef
        }
                
        rendererDefinition = cvuController.rendererDefinitionFor(getCVUContext())?.parsed ?? CVUDefinitionContent()
        
        viewDefinitionPropertyResolver = config.viewDefinition.propertyResolver(context: getCVUContext(), lookup: CVULookupController(), db: databaseController)
        
        rendererDefinitionPropertyResolver = rendererDefinition.propertyResolver(context: getCVUContext(), lookup: CVULookupController(), db: databaseController)
    }
    
    
    var supportedRenderers: Set<String> {
        let defined = viewDefinitionPropertyResolver.stringArray("supportedRenderers").map { $0.lowercased() }
        if !defined.isEmpty {
            return Set(defined)
        } else {
            let defaultRenderer = viewDefinitionPropertyResolver.string("defaultRenderer")?.lowercased()
            var list: [String] = config.viewDefinition.definitions.compactMap { def -> String? in
                guard def.type == .renderer else { return nil }
                return def.name?.lowercased()
            }
            list.appendOptional(defaultRenderer)
            return Set(list)
        }
    }
    
    lazy var rendererDefinition: CVUDefinitionContent = CVUDefinitionContent()
    
    lazy var viewDefinitionPropertyResolver: CVUPropertyResolver = {
        config.viewDefinition.propertyResolver(context: getCVUContext(), lookup: CVULookupController(), db: databaseController)
    }()
    
    lazy var rendererDefinitionPropertyResolver: CVUPropertyResolver = {
        rendererDefinition.propertyResolver(context: getCVUContext(), lookup: CVULookupController(), db: databaseController)
    }()
    
    var itemPropertyResolver: CVUPropertyResolver? {
        guard let viewDefinition = cvuController.viewDefinitionFor(itemRecord: focusedItem) else {
            return nil
        }
        
        return viewDefinition.propertyResolver(context: getCVUContext(forItem: focusedItem), lookup: CVULookupController(), db: databaseController)
    }
    
    func propertyResolverFor(_ definition: CVUDefinitionContent) -> CVUPropertyResolver {
        return propertyResolverFor(definition.properties)
    }
    
    func propertyResolverFor(_ properties: [String: CVUValue]) -> CVUPropertyResolver {
        return CVUPropertyResolver(context: getCVUContext(forItem: focusedItem), lookup: lookupController, db: databaseController, properties: properties)
    }
    
    func setRendererProperty(renderer: String, property: String, value: CVUValue) {
        if let index = config.viewDefinition.definitions.firstIndex(where: { $0.type == .renderer && $0.name == renderer }) {
            config.viewDefinition.definitions[index].parsed.properties[property] = value
        } else {
            let props = CVUDefinitionContent(properties: [property: value])
            let def = CVUParsedDefinition(type: .renderer, name: renderer, parsed: props)
            config.viewDefinition.definitions.append(def)
        }
        updateCachedValues()
    }
    
    func nodePropertyResolver(for item: ItemRecord) -> CVUPropertyResolver? {
        let context = getCVUContext(forItem: item)
        return cvuController.nodeDefinitionFor(context)?.propertyResolver(context: context, lookup: CVULookupController(), db: databaseController)
    }
    
    // MARK: Query RESULT
    /// This holds the array of results for the query. If `isObserving` is set to true, this will be up to date with any changes in the database.
    @Published
    private(set) var items: [ItemRecord] = []
    
    private(set) lazy var itemsChangePublisher: AnyPublisher<[ItemRecord], Never> = $items.removeDuplicates().eraseToAnyPublisher()
    
    /// Convenience variable returning whether there are any items
    var hasItems: Bool {
        !items.isEmpty
    }
    
    // MARK: Selection State
    @Published
    var selectedItems: Set<StringUUID> = []
    
    var selectedIndicesBinding: Binding<Set<Int>> {
        Binding { [weak self] in
            guard let self = self else { return [] }
            return Set(self.selectedItems.compactMap { uid in self.items.firstIndex(where: { $0.id == uid }) })
        } set: { [weak self] newValue in
            guard let self = self else { return }
            self.selectedItems = Set(newValue.compactMap { self.items[safe: $0]?.id })
        }
    }
    
    var allItemsSelected: Bool {
        return self.selectedItems.count > 0 && self.selectedItems.count == self.items.count
    }
    
    func selectAll() {
        self.selectedItems = Set(self.items.map{$0.id})
    }
    
    func deselectAll() {
        self.selectedItems = Set()
    }
    
    // MARK: Search State
    var searchString: String? {
        get { config.query.searchString }
        set { config.query.searchString = newValue }
    }
    
    // MARK: OTHER
    var isFullScreen: Bool = false
    
    // MARK: Configuration
    /// Determines if the query is actively observing for changes. If set to false, it won't respond to changes.
    /// Use this to avoid unneccessary observation if a query isn't being shown in the UI currently
    var isObservingQuery: Bool = true {
        didSet { if isObservingQuery != oldValue { setupQueryObservation() }}
    }

    /// Sets up a database observation so that all
    func setupQueryObservation() {
        /// Remove old observation
        queryObservation?.cancel()
        queryObservation = nil
        
        /// Only set up an observation if `isObserving` is true
        guard isObservingQuery else { return }
        
        let queryConfig = config.query
        let observation = ValueObservation
            .tracking { db in
                try queryConfig.executeRequest(db: db)
            }
        
        guard let pool = databaseController.databasePool else {
            print("Error: Cannot setup observers, database pool not available")
            return
        }

        /// Subscribe to changes in the value, with the first result being reported immediately
        queryObservation = observation.publisher(
            in: pool,
            scheduling: .immediate)
        .sink(
            receiveCompletion: {_ in},
            receiveValue: { result in
                self.items = result
                self.update()
            })
    }
    
    func update() {
        self.items = items.compactMap{ItemRecord.fetchWithID($0.id)}
        if let currentItem = self.config.focusedItem {
            self.config.focusedItem = ItemRecord.fetchWithID(currentItem.id)
            if (self.items.count == 0) {
                self.focusedItem = nil
            }
        }
        
        updateCachedValues()
        objectWillChange.send()
    }
}
