//
// ExprParserTests.swift
// Copyright © 2020 memri. All rights reserved.

@testable import Memri
import XCTest

class ExprParserTests: XCTestCase {
    typealias Lookup = Memri.CVUValue.LookupNode
    override func setUpWithError() throws {}
    
    override func tearDownWithError() throws {}
    
    private func parse(_ snippet: String, stringMode: Bool = false) throws -> CVUValue.ExpressionNode {
        let lexer = CVUExpressionLexer(input: snippet, startInStringMode: stringMode)
        let tokens = try lexer.tokenize()
        let parser = CVUExpressionParser(tokens)
        return try parser.parse()
    }
    
    func testArithmeticOperators() throws {
        let snippet = "(5 + 10 * 4 - 3 / 10) / 10"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.division(
                .subtraction(
                    .addition(
                        .constant(.number(5)),
                        .multiplication(
                            .constant(.number(10)),
                            .constant(.number(4))
                        )
                    ),
                    .division(
                        .constant(.number(3)),
                        .constant(.number(10))
                    )
                ),
                .constant(.number(10))
            )
        )
    }
    
    func testAnd() throws {
        let snippet = "true and false"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.and(.constant(.bool(true)), .constant(.bool(false)))
        )
    }
    
    func testOr() throws {
        let snippet = "true or false"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.or(.constant(.bool(true)), .constant(.bool(false)))
        )
    }
    
    func testSimpleCondition() throws {
        let snippet = "true ? 'yes' : 'no'"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.conditional(condition: .constant(.bool(true)), trueExp: .constant(.string("yes")), falseExp: .constant(.string("no")))
        )
    }
    
    func testMultiCondition() throws {
        let snippet = "true ? false and true ? -1 : false or true ? 'yes' : 'no' : -1"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.conditional(condition: .constant(.bool(true)),
                                                trueExp: .conditional(condition: .and(.constant(.bool(false)), .constant(.bool(true))),
                                                                      trueExp: .subtraction(.constant(.number(0)), .constant(.number(1))),
                                                                      falseExp: .conditional(condition: .or(.constant(.bool(false)), .constant(.bool(true))),
                                                                                             trueExp: .constant(.string("yes")),
                                                                                             falseExp: .constant(.string("no")))),
                                                falseExp: .subtraction(.constant(.number(0)), .constant(.number(1))))
        )
    }
    
    func testConditionEquals() throws {
        let snippet = "true = false"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.areEqual(.constant(.bool(true)), .constant(.bool(false)))
        )
    }
    
    func testConditionNotEquals() throws {
        let snippet = "true != false"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.areNotEqual(.constant(.bool(true)), .constant(.bool(false)))
        )
    }
    
    func testConditionGreaterThan() throws {
        let snippet = "5 > 10"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.greaterThan(.constant(.number(5)), .constant(.number(10)))
        )
    }
    
    func testConditionGreaterThanOrEqual() throws {
        let snippet = "5 >= 10"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.greaterThanOrEqual(.constant(.number(5)), .constant(.number(10)))
        )
    }
    
    func testConditionLessThan() throws {
        let snippet = "5 < 10"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.lessThan(.constant(.number(5)), .constant(.number(10)))
        )
    }
    
    func testConditionLessThanOrEqual() throws {
        let snippet = "5 <= 10"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            CVUValue.ExpressionNode.lessThanOrEqual(.constant(.number(5)), .constant(.number(10)))
        )
    }
    
    func testLookup() throws {
        let snippet = ".bar and bar.foo(10) and bar[foo = 10] or shouldNeverGetHere"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .and(.and(.lookup(nodes: [.default, Lookup(name: "bar", isArray: false, type: .lookup(subExpression: nil))]),
                      .lookup(nodes: [Lookup(name: "bar", isArray: false, type: .lookup(subExpression: nil)), Lookup(name: "foo", isArray: false, type: .function(args: [.constant(.number(10.0))]))])),
                 .or(.lookup(nodes: [Lookup(name: "bar", isArray: false, type: .lookup(subExpression: Optional(.areEqual(.lookup(nodes: [Lookup(name: "foo", isArray: false, type: .lookup(subExpression: nil))]),
                                                                                                                         .constant(.number(10.0))))))]),
                     .lookup(nodes: [Lookup(name: "shouldNeverGetHere", isArray: false, type: .lookup(subExpression: nil))])))
        )
    }
    
    func testDotLookup() throws {
        let snippet = "."
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .lookup(nodes: [.default])
        )
    }
    
    func testMinusPlusModifier() throws {
        let snippet = "-5 + -(5+10) - +'5'"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .subtraction(
                .addition(
                    .subtraction(
                        .constant(.number(0.0)),
                        .constant(.number(5.0))),
                    .subtraction(
                        .constant(.number(0.0)),
                        .addition(
                            .constant(.number(5.0)),
                            .constant(.number(10.0))))),
                .constant(.string("5")))
        )
    }
    
    func testNegation() throws {
        let snippet = "!true"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(result,
                       .negation(.constant(.bool(true))))
    }
    
    func testStringEscaping() throws {
        let snippet = "'asdadsasd\\'asdasd'"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(result,
                       .constant(.string("asdadsasd'asdasd")))
    }
    
    func testTypeConversionToNumber() throws {
        let snippet = "5 + '10.34' + true"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .addition(
                .addition(
                    .constant(.number(5.0)),
                    .constant(.string("10.34"))),
                .constant(.bool(true)))
        )
    }
    
    func testTypeConversionToBool() throws {
        let snippet = "0 ? -1 : 1 ? '' ? -1 : 'yes' : -1"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .conditional(condition:
                            .constant(.number(0.0)), trueExp:
                                .subtraction(
                                    .constant(.number(0.0)),
                                    .constant(.number(1.0))), falseExp:
                                        .conditional(condition:
                                                        .constant(.number(1.0)), trueExp:
                                                            .conditional(condition:
                                                                            .constant(.string("")), trueExp:
                                                                                .subtraction(
                                                                                    .constant(.number(0.0)),
                                                                                    .constant(.number(1.0))), falseExp:
                                                                                        .constant(.string("yes"))), falseExp:
                                                                                            .subtraction(
                                                                                                .constant(.number(0.0)),
                                                                                                .constant(.number(1.0)))))
        )
    }
    
    func testSelfUsageInSubExpression() throws {
        let snippet = ".relation[. = me].firstName"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .lookup(nodes: [.default, Lookup(name: "relation", isArray: false, type: .lookup(subExpression: Optional(.areEqual(.lookup(nodes: [.default]),
                                                                                                                               .lookup(nodes: [Lookup(name: "me", isArray: false, type: .lookup(subExpression: nil))]))))), Lookup(name: "firstName", isArray: false, type: .lookup(subExpression: nil))])
        )
    }
    
    func testLookupItem() throws {
        let snippet = ".sibling"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .lookup(nodes: [.default, Lookup(name: "sibling", isArray: false, type: .lookup(subExpression: nil))])
        )
    }
    
    func testLookupReverseEdgeItems() throws {
        let snippet = ".~sibling"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .lookup(nodes: [.default, Lookup(name: "~sibling", isArray: false, type: .lookup(subExpression: nil))])
        )
    }
    
    func testLookupEdges() throws {
        let snippet = "._sibling"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .lookup(nodes: [.default, Lookup(name: "_sibling", isArray: false, type: .lookup(subExpression: nil))])
        )
    }
    
    func testLookupReverseEdges() throws {
        let snippet = "._~sibling[]"
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .lookup(nodes: [.default, Lookup(name: "_~sibling", isArray: true, type: .lookup(subExpression: nil))])
        )
    }
    
    func testStringModeStartWithString() throws {
        let snippet = "Hello {fetchName()}!"
        
        let result = try parse(snippet, stringMode: true)
        
        XCTAssertEqual(
            result,
            .stringMode(nodes: [.constant(.string("Hello ")),
                                .lookup(nodes: [Lookup(name: "fetchName", isArray: false, type: .function(args: []))]),
                                .constant(.string("!"))])
        )
    }
    
    func testStringModeMultipleBlocks() throws {
        let snippet = "Hello {.firstName} {.lastName}"
        
        let result = try parse(snippet, stringMode: true)
        
        print(result.description)
        
        XCTAssertEqual(
            result,
            .stringMode(nodes: [.constant(.string("Hello ")),
                                .lookup(nodes: [.default, Lookup(name: "firstName", isArray: false, type: .lookup(subExpression: nil))]),
                                .constant(.string(" ")),
                                .lookup(nodes: [.default, Lookup(name: "lastName", isArray: false, type: .lookup(subExpression: nil))])])
        )
    }
    
    func testStringModeUsingOr() throws {
        let snippet = "{.title or \"test\"} — {.content.plainString()}"
        
        let result = try parse(snippet, stringMode: true)
        
        print(result.description)
        
        XCTAssertEqual(
            result,
            .stringMode(nodes: [.or(.lookup(nodes: [.default, Lookup(name: "title", isArray: false, type: .lookup(subExpression: nil))]),
                                    .constant(.string("test"))),
                                .constant(.string(" — ")),
                                .lookup(nodes: [.default, Lookup(name: "content", isArray: false, type: .lookup(subExpression: nil)), Lookup(name: "plainString", isArray: false, type: .function(args: []))])])
        )
    }
    
    func testStringModeStartWithExpression() throws {
        let snippet = "{fetchName()} Hello"
        
        let result = try parse(snippet, stringMode: true)
        
        print(result.description)
        
        XCTAssertEqual(
            result,
            .stringMode(nodes: [.lookup(nodes: [Lookup(name: "fetchName", isArray: false, type: .function(args: []))]),
                                .constant(.string(" Hello"))])
        )
    }
    
    func testStringModeWithQuote() throws {
        let snippet = "Photo AND ANY includes.uid = {.uid}"
        
        let result = try parse(snippet, stringMode: true)
        
        print(result.description)
        XCTAssertEqual(
            result,
            .stringMode(nodes: [.constant(.string("Photo AND ANY includes.uid = ")),
                                .lookup(nodes: [.default, Lookup(name: "uid", isArray: false, type: .lookup(subExpression: nil))])])
        )
    }
    
    func testExample() throws {
        let snippet = """
        !(test + -5.63537) or 4/3 ? variable.func() : me.address[primary = true].country ? ((4+5 * 10) + test[10]) : 'asdads\\'asdad' + ''
        """
        
        let result = try parse(snippet)
        
        XCTAssertEqual(
            result,
            .conditional(condition: .or(.negation(.addition(.lookup(nodes: [Lookup(name: "test", isArray: false, type: .lookup(subExpression: nil))]),
                                                            .subtraction(.constant(.number(0.0)),
                                                                         .constant(.number(5.63537))))),
                                        .division(.constant(.number(4.0)),
                                                  .constant(.number(3.0)))), trueExp: .lookup(nodes: [Lookup(name: "variable", isArray: false, type: .lookup(subExpression: nil)), Lookup(name: "func", isArray: false, type: .function(args: []))]), falseExp: .conditional(condition: .lookup(nodes: [Lookup(name: "me", isArray: false, type: .lookup(subExpression: nil)), Lookup(name: "address", isArray: false, type: .lookup(subExpression: Optional(.areEqual(.lookup(nodes: [Lookup(name: "primary", isArray: false, type: .lookup(subExpression: nil))]),
                                                                                                                                                                                                                                                                                                                                                                                                                                                                       .constant(.bool(true)))))), Lookup(name: "country", isArray: false, type: .lookup(subExpression: nil))]), trueExp: .addition(.addition(.constant(.number(4.0)),
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              .multiplication(.constant(.number(5.0)),
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              .constant(.number(10.0)))),
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                    .lookup(nodes: [Lookup(name: "test", isArray: false, type: .lookup(subExpression: Optional(.constant(.number(10.0)))))])), falseExp: .addition(.constant(.string("asdads\'asdad")),
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                   .constant(.string("")))))
        )
    }
    
    func testErrorIncompleteCondition() throws {
        let snippet = "true ? 'yes'"
        
        do {
            _ = try parse(snippet)
        }
        catch CVUExpressionParseErrors.ExpectedConditionElse {
            return
        }
        
        XCTFail()
    }
    
    func testErrorIncompleteBinaryOp() throws {
        let snippet = "5 +"
        
        do {
            _ = try parse(snippet)
        }
        catch let CVUExpressionParseErrors.ExpectedExpression(token) {
            XCTAssertEqual("\(token)", "\(CVUExpressionLexer.ExprToken.EOF)")
            return
        }
        
        XCTFail()
    }
    
    func testErrorUnsupportedBinaryOp() throws {
        let snippet = "5 @ 4"
        
        do {
            _ = try parse(snippet)
        }
        catch let CVUExpressionParseErrors.UnexpectedToken(token) {
            if case let CVUExpressionLexer.ExprToken.Identifier(keyword, _) = token {
                XCTAssertEqual(keyword, "@")
            }
            return
        }
        
        XCTFail()
    }
    
    func testErrorMissingParenClose() throws {
        let snippet = "(5 + 10"
        
        do {
            _ = try parse(snippet)
        }
        catch let CVUExpressionParseErrors.ExpectedCharacter(chr) {
            XCTAssertEqual(chr, ")")
            return
        }
        
        XCTFail()
    }
    
    func testErrorMissingCallParenClose() throws {
        let snippet = "foo("
        
        do {
            _ = try parse(snippet)
        }
        catch let CVUExpressionParseErrors.ExpectedExpression(token) {
            XCTAssertEqual("\(token)", "\(CVUExpressionLexer.ExprToken.EOF)")
            return
        }
        
        XCTFail()
    }
    
    func testErrorMissingBracketClose() throws {
        let snippet = "test[10"
        
        do {
            _ = try parse(snippet)
        }
        catch let CVUExpressionParseErrors.ExpectedCharacter(chr) {
            XCTAssertEqual(chr, "]")
            return
        }
        
        XCTFail()
    }
    
    func testErrorMissingQuoteClose() throws {
        let snippet = "'asdads"
        
        do {
            _ = try parse(snippet)
        }
        catch CVUExpressionParseErrors.MissingQuoteClose {
            return
        }
        
        XCTFail()
    }
    
    func testErrorUsingCurlyBracesNotInStringMode() throws {
        let snippet = "Hello {fetchName()}"
        
        do {
            _ = try parse(snippet)
        }
        catch let CVUExpressionParseErrors.UnexpectedToken(token) {
            XCTAssertEqual("\(token)", "\(CVUExpressionLexer.ExprToken.CurlyBracketOpen(6))")
            return
        }
        
        XCTFail()
    }
    
    func testErrorUsingCurlyBracesInWrongContext() throws {
        let snippet = "Hello {'{fetchName()}'}"
        
        do {
            let _ = try parse(snippet, stringMode: true)
        }
        catch let CVUExpressionParseErrors.ExpectedExpression(token) {
            XCTAssertEqual("\(token)", "\(CVUExpressionLexer.ExprToken.CurlyBracketClose(22))")
            return
        }
        
        XCTFail()
    }
    
    //    func testPerformanceExample() throws {
    //        // This is an example of a performance test case.
    //        measure {
    //            // Put the code you want to measure the time of here.
    //        }
    //    }
}
