//
//  MemriPodAPITests.swift
//  MemriPodAPITests
//
//  Created by T Brennan on 19/12/20.
//

import XCTest
import Combine
@testable import Memri

/// This test suite requires a running pod in order to execute successfully.
class MemriPodAPITests: XCTestCase {    
    /// This connection config is used to connect to the pod for the tests. You can change url scheme/path/port etc here
    let connectionConfig = PodAPIConnectionDetails(keys: AuthKeys(ownerKey: "TESTPOD", dbKey: "DBKEY"))
    /// These are used to create test items in the pod for use in later tests
    let noteRowId: Int64 = 100
    let importerRowId: Int64 = 101
    let indexerRowId: Int64 = 102
    let fileRowId: Int64 = 103
    
    let noteUID = StringUUID()
    let importerUID = StringUUID()
    let indexerUID = StringUUID()
    let fileUID = StringUUID()

    let schema = try! Schema.loadFromFile()
    
    private var testCancellables: Set<AnyCancellable> = []
    
    func setupPodForTesting() throws {
        let noteItem = ItemRecord(rowId:noteRowId, id: noteUID, type: "Note")
        let importerItem = ItemRecord(rowId:importerRowId, id: importerUID, type: "Importer")
        let indexerItem = ItemRecord(rowId:indexerRowId, id: indexerUID, type: "Indexer")
        let fileItem = ItemRecord(rowId:fileRowId, id: fileUID, type: "File")

        let testNoteItem: [String: AnyEncodable] = {
            let testItemProperties: [ItemPropertyRecord] = [
                StringRecord(item: noteRowId, name: "content", value: "Test note content"),
                StringRecord(item: noteRowId, name: "title", value: "Test note title")
            ]
            
            return noteItem._syncDict(properties: testItemProperties, schema: schema)
        }()
        
        let testImporterItem: [String: AnyEncodable] = {
            let testItemProperties: [ItemPropertyRecord] = [
                StringRecord(item: importerRowId, name: "repository", value: "importerRepo")
            ]
            return importerItem._syncDict(properties: testItemProperties, schema: schema)
        }()
        
        let testIndexerItem: [String: AnyEncodable] = {
            let testItemProperties: [ItemPropertyRecord] = [
                StringRecord(item: indexerRowId, name: "repository", value: "indexerRepo")
            ]
            return indexerItem._syncDict(properties: testItemProperties, schema: schema)
        }()
        
        let testFileItem: [String: AnyEncodable] = {
            let testItemProperties: [ItemPropertyRecord] = [
                StringRecord(item: fileRowId, name: "sha256", value: testFileSHA)
            ]
            return fileItem._syncDict(properties: testItemProperties, schema: schema)
        }()
        
        let testEdge: [String: AnyEncodable] = {
            let testEdge = try? EdgeRecord.createFor(source: noteRowId, type: "file", target: fileRowId)
            return testEdge?._syncDict()
        }()
        
        let bulkAction = PodAPIPayload.BulkAction(
            createItems: [testNoteItem, testImporterItem, testIndexerItem, testFileItem],
            updateItems: [],
            deleteItems: [],
            createEdges: [testEdge])
        
        let request = PodAPIStandardRequest.bulkAction(payload: bulkAction)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
    
    override func setUpWithError() throws {
        try setupPodForTesting()
    }

    override func tearDownWithError() throws {
        // Put teardown code here. This method is called after the invocation of each test method in the class.
    }
    
    
    /// This function is used to await a network request to the pod before continuing
    /// It accepts a throwing function handling the response
    /// **Note:** in its current form this function assumes only a single value is returned (ie. one-off publishers)
    func await<T: Publisher>(publisher: T, handleValue: @escaping (T.Output) throws -> Void) {
        let e = expectation(description: "PodResponse")
    
        publisher.sink { (completion) in
            switch completion {
            case .failure(let error):
                XCTFail(error.localizedDescription)
            case .finished:
                break
            }
        } receiveValue: { (value) in
            do {
                try handleValue(value)
            }
            catch {
                XCTFail(error.localizedDescription)
            }
            e.fulfill()
        }.store(in: &testCancellables)

        waitForExpectations(timeout: 10.0, handler: nil)
    }
        
    func testRunIndexer() throws {
        let request = PodAPIStandardRequest.runIndexer(indexerUID: indexerUID, connectionConfig: connectionConfig)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
    
    func testGetItemWithEdges() throws {
        let request = PodAPIStandardRequest.getItemWithEdges(id: noteUID)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
            let items = try response.result.get() as! [[String: Any]]
            XCTAssert(items.count == 1)
        }
    }
    
    
    func testGetItemsWithEdges() throws {
        let request = PodAPIStandardRequest.getItemsWithEdges(itemIDs: [noteUID, indexerUID, importerUID])
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
            let items = try response.result.get() as! [[String: Any]]
            XCTAssert(!items.isEmpty)
        }
    }
    
    func testSearchByFields() throws {
        let request = PodAPIStandardRequest.searchByFields(itemType: "Note")
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
            let items = try response.result.get() as! [[String: Any]]
            XCTAssert(!items.isEmpty)
        }
    }
    
    func testGetAllItems() throws {
        let request = PodAPIStandardRequest.getAllItems()
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
            let items = try response.result.get() as! [[String: Any]]
            XCTAssert(!items.isEmpty)
        }
    }
    
    func testDownloadFile() throws {
        let request = PodAPIDownloadRequest.downloadFile(fileSHAHash: "2d05f86fec5b76dc7adcb19657885dc4844b7620edb0b5832b857a544a8eb08a", fileUID: "testFile")
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            let localFileURL = try response.result.get()
            let _ = localFileURL
        }
    }
    
    func testCreateItem() throws {
        let itemRowId: Int64 = 201
        let itemId = StringUUID()
        let item = ItemRecord(rowId: itemRowId, id: itemId, type: "Note")
        let properties = [
            StringRecord(item: itemRowId, name: "content", value: "Test note content"),
            StringRecord(item: itemRowId, name: "title", value: "Test note title")
        ]
        let syncDict = item._syncDict(properties: properties, schema: schema)
        let request = PodAPIStandardRequest.createItem(syncDict: syncDict)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
    
    let testFileSHA = "4156ae4d5c719f5adba859be68e5c4541058863dde6e76c93a6a0a3f0f9a361b"
    func testUploadFile() throws {
        // Make an upload request
        let url = Bundle.main.url(forResource: "schema", withExtension: "json")!
        let request = try PodAPIUploadRequest.uploadFile(fileURL: url, connectionConfig: connectionConfig)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            do {
                try response.throwIfError()
            }
            catch {
                // If file already exists, ignore this error -> otherwise throw
                if error.localizedDescription != "POD RESPONDED WITH ERROR: File already exists" {
                 throw error
                }
            }
        }
    }
    
    
    func testBulkAction() throws {
        let testCreateItemRowId: Int64 = 201
        let testCreateItemId = StringUUID()
        let testCreateItem = ItemRecord(rowId: testCreateItemRowId, id: testCreateItemId, type: "Note")

        let testCreateItemProperties = [
            StringRecord(item: testCreateItemRowId, name: "content", value: "Test note content"),
            StringRecord(item: testCreateItemRowId, name: "title", value: "Test note title")
        ]
        let testCreateItemSyncDict = testCreateItem._syncDict(properties: testCreateItemProperties, schema: schema)
        
        let testUpdateItemRowId: Int64 = 201
        let testUpdateItemId = StringUUID()
        let testUpdateItem = ItemRecord(rowId: testUpdateItemRowId, id: testUpdateItemId, type: "Note")
        let testUpdateItemProperties = [
            StringRecord(item: testUpdateItemRowId, name: "content", value: "Updated note content"),
            StringRecord(item: testUpdateItemRowId, name: "title", value: "Updated note title")
        ]
        let testUpdateItemSyncDict = testUpdateItem._syncDict(properties: testUpdateItemProperties, schema: schema)
        
        let bulkAction = PodAPIPayload.BulkAction(
            createItems: [testCreateItemSyncDict],
            updateItems: [testUpdateItemSyncDict],
            deleteItems: [],
            createEdges: [])
        
        let request = PodAPIStandardRequest.bulkAction(payload: bulkAction)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
    
    func testSync() throws {
        let database = DatabaseController(databaseName: "podAPITestDB")
        try database.runMigrations()
        try DemoData.importDemoData(databaseController: database, throwIfAgainstSchema: false)
        
        let syncController = SyncController(databaseController: database)
        let bulkAction = try syncController.makeSyncUploadData(maxItems: 10000)
        
        
        let request = PodAPIStandardRequest.bulkAction(payload: bulkAction)
        let networkCall = try request.execute(connectionConfig: connectionConfig)
        await(publisher: networkCall) { (response) in
            try response.throwIfError()
        }
    }
}

import Alamofire
extension DataResponse {
    func throwIfError() throws {
        switch result {
        case .success:
            break
        case let .failure(error):
            if let data = data,
               let responseString = String(data: data, encoding: .utf8) {
                throw StringError(description: "POD RESPONDED WITH ERROR: \(responseString)")
            } else {
                throw error
            }
        }
    }
}
